package Abstract;

import java.lang.*;
import java.util.*;
import java.net.*;
import java.io.*;

/**
* Abstract AgentAction class. 
* You will need to subclass this class to act according to the
* message received. Since JATLite assumes to use MessageBuffer,
* the ReceiverThread will receive message and add it to MessageBuffer.
* AgentAction will check the received message and retrieve it 
* when available.<BR>
* @see BaseLayer.BAgentAction
* @see KQMLLayer.KQMLAgentAction
* @see RouterLayer.Router.RouterAction
* @see RouterLayer.AgentClient.RouterClientAction
*/

/*
* Copyright (c) 1996,1997,1998 <A HREF="http://cdr.stanford.edu/">Center for Design Research</A>, Stanford University.
* All rights reserved.<p>
*
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You can obtain the GNU General Public License at
* http://www.gnu.org/copyleft/gpl.html or by writing to the Free
* Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
* 02111-1307, USA.
* 
* This software is bound by the terms and conditions listed in the
* attached <a href="../LICENSE">LICENSE file</A>.
*  @author <A HREF="http://cdr.stanford.edu/html/people/jeon-bio.html>Heecheol Jeon</A>
*  @version Java(tm) 1.1
*/ 

public abstract class AgentAction extends Thread {
	
	/**
	* Message Buffer data member
	* @see Abstract.MessageBuffer
	*/
	protected MessageBuffer _queue;
	
	/** 
	* Address table
	* @see Abstract.AddressTable
	*/
	protected AddressTable _addresses;
	
	/**
	* Connection table, which contains active ReceiverThread. 
	* The key is receiver name and the field is the ReceiverThread object.
	* @see Abstract.ConnectionTable
	* @see Abstract.ReceiverThread
	*/
	protected ConnectionTable _connections;
	
	
	/**
	* Regisger, unregister protocol will be taken by this security object. 
	* For any communication, this protocol should be specified.
	* @see Abstract.Connection.Security
	*/
	protected Security _security;
	
	/**
	* EndWith character to create ServerThread and ReceiverThead.
	* If your messages are finished with other than '\004', you must
	* set endWith by setEndWith(char) method. For example, if you
	* want to set endWith character for ServerThread,<BR>
	* AgentAction action = new AgentAction();<BR>
	* action.setEndWith('\004');<BR>
	* ServerThread server = action.createServerThread("Server",Thread.NORM_PRIORITY);<BR>
	* ....<BR>
	* If you set the endWith character on the ServerThread, the generated ReceiverThread will
	* have the same endWith character.
	* @see Abstract.ServerThread
	* @see Abstract.ReceiverThread
	*/
	protected char _endWith;
	
	/**
	* Set maximum idle time for the message receiving threads. Unit is minute.
	* Maximum idle time is not absolute value. If a message is received or sent,
	* timer will reset timer.
	* If this is set to -1, no time out check. Default is -1.
	* @see Abstract.ServerThread
	* @see Abstract.ReceiverThread
	*/
	protected int _durationTime;
	
	/**
	* Default constructor
	*/
	public AgentAction()  {
		super();
		_queue = null;
		_addresses = null;
		_connections = null;
		_security = null;
		_endWith = '\004';
		_durationTime = -1;
	}
	
	/**
	* Set the thread name identical to the agent name
	* @param id Agent Name
	*/
	public AgentAction(String id)  {
		this();
		setName(id);
	}

	/**
	* Set the thread name identical to the agent name. Default endWithChar and duration time 
	* will be assigned.
	* @param tg Thread group to which this thread will be owned
	* @param id Agent Name
	*/
	public AgentAction(ThreadGroup tg,String id)  {
		super(tg,id);
		_queue = null;
		_addresses = null;
		_connections = null;
		_security = null;
		_endWith = '\004';
		_durationTime = -1;
		
	}

	/**
	* Constructor
	* @param id Agent Name which is identical to the thread name
	* @param queue MessageBuffer
	* @param addresses AddressTable
	* @param connections ConnectionTable
	* @param security Security object
	*/
	public AgentAction(String id,
										 MessageBuffer queue,
										 AddressTable addresses,
										 ConnectionTable connections,
										 Security security)  {
		super(id);										 
		_queue = queue;
		_addresses = addresses;
		_connections = connections;
		_security = security;
		_endWith = '\004';
		_durationTime = -1;
	}
	
	
	/**
	* Constructor
	* @param id Agent Name which is identical to the thread name
	* @param queue MessageBuffer
	* @param addresses AddressTable
	* @param connections ConnectionTable
	* @param security Security object
	* @param tg Thread group for the initiated receiver thread
	*/
	public AgentAction(String id,
										 MessageBuffer queue,
										 AddressTable addresses,
										 ConnectionTable connections,
										 Security security,
										 ThreadGroup tg)  {
		super(tg,id);
 		_queue = queue;
		_addresses = addresses;
		_connections = connections;
		_security = security;
		_endWith = '\004';
		_durationTime = -1;
	}
	
	/**
	* Set MessageBuffer
	* @param queue MessageBuffer to restore the message from the socket buffer
	*/
	public void setMessageQueue(MessageBuffer queue)  {
		_queue = queue;
	}
	
	/**
	* Set AddressTable
	* @param addresses AddressTable
	*/
	public void setAddressTable(AddressTable addresses)  {
		_addresses = addresses;
	}
	
	/**
	* Set ConnectionTable
	* @param connections ConnectionTable
	*/
	public void setConnectionTable(ConnectionTable connections)  {
		_connections = connections;
	}
	
	
	/**
	* Get MessageBuffer
	* @return MessageBuffer, the AgentAction has
	*/
	public MessageBuffer getMessageQueue()  {
		return _queue;
	}
	
	/**
	* Get AddressTable
	* @return Agent action's AddressTable
	*/
	public AddressTable getAddressTable()  {
		return _addresses;
	}
	
	/**
	* Get ConnectionTable
	* @return Agent action's ConnectionTable
	*/
	public ConnectionTable getConnectionTable()  {
		return _connections;
	}
	
	/**
	* Set Security object
	* @param security Security object
	*/
	public void setSecurity(Security security)  {
		_security = security;
	}
	
	/**
	* Accessor to get Security object
	* @return _security data member
	*/
	public Security getSecurity()  {
		return _security;
	}
	
	/**
	* Set endWith character for all the ServerThread and ReceiverThread
	* @param c End with character to be set
	*/
	public void setEndWith(char c)  {
		_endWith = c;
	}
	
	/**
	* Get endWith character
	* @return End with character
	*/
	public char getEndWith()  {
		return _endWith;
	}		
		
	/**
	* Set ReceiverThread maximum idle time. All the ReceiverThread's idle time will 
	* be set if you pass time other than -1.
	* @param time Maximum idle time in minutes
	*/	
	public void setDurationTime(int time)  {
		_durationTime = time;
	}
	
	/**
	* Get current duration time set
	* @return Maximum idle time for the ReceiverThread
	*/
	public int getDurationTime()  {
		return _durationTime;
	}
	
					
	/**
	* Abstract class to create ServerThread. 
	* Create server thread using name of the agent. 
	* Override this method if your connection type is different.
	* @param id ServerThread name
	* @param priority Server thread priority
	* @return ServerThread, generated
	*/
	public abstract ServerThread createServerThread(String id,int priority);


	/**
	* Abstract class to create ReceiverThread. 
	* Create client receiver thread using name of the agent. 
	* Override this method if your connection type is different.
	* @param id agent name. This name is not my name. The name of the other
	* agent you want to connect
	* @param priority Receiver thread priority
	* @exception Abstract.ConnectionException if error occurs while creating and checking
	*  security
	* @return ReceiverThread, generated
	*/
  public abstract ReceiverThread createReceiverThread(String serverid,int priority)
    throws ConnectionException;

		
	/**
	* Send Message by specifying a receiver and a message.
	* @param receiver Receiver Agent
	* @param msg Message to be sent
	* @exception Abstract.ConnectionException if a message can not be sent
	*/
	public abstract void sendMessage(String receiver,String msg)
	  throws ConnectionException;
	
	/**
	* Broad cast to all the running connecitons. If connection is not alive, sending 
	* to the agent will be ignored.
	* @param msg Message to be sent all the living connections
	* @excetion Abstract.ConnectionException if a message can not be sent
	*/
	public abstract void broadCast(String msg)
	  throws ConnectionException;
	
	/**
	* Abstract method to process message given any other objects. 
	* This method can be used for data exchange between threads or 
	* between application and GUI. Alpha version processMessage(String,String[]) should 
	* be replaced by this method. If you already used processMessage(String,String[]) 
	* method, modify the processMessage(String,String[]) to processMessage(String,Object) 
	* and cast Object as String[].<BR>
	* For example :<BR>
	* processMessage(String c,String[] s) should be replaced as <BR>
	* processMessage(String c,Object obj)  {<BR>
	* String[] s = (String[])obj;<BR>
	* .....<BR>
	* }<BR>
	* @param command Command to be executed
	* @param obj Reference object
	*/
	public abstract void processMessage(String command,Object obj);

	/**
	* Actual Action method. Note that the parameter is object, which should
	* be instantiated by casting according to the parameter type. This method will be 
	* called whenever there's a received message from Receiver Thread object.
	*/
	protected abstract boolean Act(Object o);
	
	/**
	* Override run() method as you need. You can check the MessageBuffer
	* and act according to the message received.
	* Example implementation for run() method is<BR>
	*	public void run()  {<BR>
	*	while(true)  {<BR>
	*		<UL>if(!_queue.isEmpty())  {<BR>				
	*			<UL>Act(_queue.getMessage());</UL>
	*		}<BR>
	*		else  {<BR>
	*			<UL>yield();</UL>
	*		}<BR>
	*	}</UL>
	* }<BR>				
	*/
	public void run()  {}
	
	/**
	* You can stop gracefully to clean up
	*/
	public abstract void endAction(); 
}		
		
		
																	 	