package Abstract;

import java.lang.*;

/**
* Abstract class for Connection. The connection includes ServerThread and ReceiverThread.
* Although the JATLite only provide the TCP/IP based connection, you may extend UDP or
* other protocols by modifing this class.
* @see Abstract.ReceiverThread
* @see Abstract.ServerThread
*/

/*
* Copyright (c) 1996,1997,1998 <A HREF="http://cdr.stanford.edu/">Center for Design Research</A>, Stanford University.
* All rights reserved.<p>
*
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You can obtain the GNU General Public License at
* http://www.gnu.org/copyleft/gpl.html or by writing to the Free
* Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
* 02111-1307, USA.
*
* This software is bound by the terms and conditions listed in the
* attached <a href="../LICENSE">LICENSE file</A>.
*  @author <A HREF="http://cdr.stanford.edu/html/people/jeon-bio.html>Heecheol Jeon</A>
*  @version Java(tm) 1.1
*/

public abstract class Connection  extends Thread  {

	/**
	* ConnectionTable, which contains the running receiver thread information
	* @see Abstract.ConnectionTable
	*/
	protected ConnectionTable _connections;

	/**
	* Address which represent my(agent) address
	* @see Abstract.Address
	*/
	protected Address _myaddress;

	/**
	* Data member to manage register and connection
	* @see Abstract.Security
	*/
	protected Security _security;

	/**
	* endWith character, may be dependent upon the implementation
	*/
	protected char _endWith;

	/**
	* Maximum idle time. Default is no setting. You can set duration time
	* using setDurationTime(long). Unit is minute.
	*/
	protected int _durationTime;

	/**
	* Default Constructor
	*/
	public Connection()  {
	}

	/**
	* Set Connection name(which is identical to the agent name)
	*/
	public Connection(String name)  {
		super(name);
	}

	/**
	* Constructor with thread group
	*/
	public Connection(ThreadGroup g,String name)  {
		super(g,name);
	}

	/**
	* Set EndWith character. Note that the endWith character is not
	* passed to the constructor. You can set endWith character at any time
	* if your endWith character is different from '\004'
	* @param c EndWith character to be set to send/receive message
	*/
	public void setEndWith(char c)  {
		_endWith = c;
	}

  /**
  * Accessor to get endWith character
  */
  public char getEndWith()  {
  	return _endWith;
  }


	/**
	* Set my address
	* @param myaddress My address object
	*/
	public void setAddress(Address myaddress)  {
		_myaddress = myaddress;
	}

	/**
	* Get Connection address
	* @return My address
	*/
	public Address getAddress()  {
		return _myaddress;
	}

	/**
	* Set ConnectionTable for Connection object
	* @param ctable ConnectionTable
	*/
	public void setConnectionTable(ConnectionTable ctable)  {
		_connections = ctable;
	}

	/**
	* Get ConnectionTable
	* @return Connection's ConnectionTable
	*/
	public ConnectionTable getConnectionTable()  {
		return _connections;
	}

	/**
	* Set Security object
	* @param security Security object
	*/
	public synchronized void setSecurity(Security security) throws ConnectionException {
		_security = security;
	}

	/**
	* Get Security object
	* @return Security object, the connection object are using
	*/
	public Security getSecurity()  {
		return _security;
	}

  /**
  * Set duration time as a measure of minutes.
  * The implementation of this method is dependent upon the ReceiverThread.
  * If you set DurationTime for ServerThread, the generated ReceiverThread
  * will have the same duration time set to ServerThread.
  * @param minutes Idle duration time in minutes
  */
	public abstract void setDurationTime(int minutes);

  /**
  * Disconnect connection
  */
  public abstract /*synchronized*/ void endConn();

}
