package Abstract;

import java.net.*;
import java.lang.*;


/**
* Security class to check the connection protocol and update AddressTable, ConnectionTable.
* For any connection, there should be an assumed connection protocol between two sides.
* The Security will implement the connection protocol. If the protocol is satisfied, the Security
* will add the connection to the connection. 
* @see BaseLayer.BSecurity
* @see KQMLLayer.KQMLSecurity
* @see RouterLayer.Router.RouterSecurity
*/

/*
* Copyright (c) 1996,1997,1998 <A HREF="http://cdr.stanford.edu/">Center for Design Research</A>, Stanford University.
* All rights reserved.<p>
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You can obtain the GNU General Public License at
* http://www.gnu.org/copyleft/gpl.html or by writing to the Free
* Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
* 02111-1307, USA.
* 
* This software is bound by the terms and conditions listed in the
* attached <a href="../LICENSE">LICENSE file</A>.
*  @author <A HREF="http://cdr.stanford.edu/html/people/jeon-bio.html>Heecheol Jeon</A>
*  @version Java(tm) 1.1
*/ 

public abstract class Security   {
	
	/**
	* AddressTable, which will be maintained by the Security
	* @see Abstract.AddressTable
	*/
	protected AddressTable _addresses;
	
	/**
	* Password file path.
	*/
	protected String _passwordFilePath;
	
	/**
	* Set true if password check will be performed. If true, it will perform password checking.
	*/
	protected boolean _passwordCheck;
	
	/**
	* Default constructor. By default, no password checking is assumed
	*/
  public Security()  {
		_passwordCheck = false;
	}
	
	/**
	* Constructor with AddressTable
	* @param addresses AddressTable to be maintained
	*/
	public Security(AddressTable addresses)  {
		_addresses = addresses;
	}
		
	/**
	* Constructor with AddressTable and password file path
	* @param addresses AddressTable to be maintained
	* @param passwordFilePath Password file path 
	*/	
	public Security(AddressTable addresses,String passwordFilePath)  {
		_addresses = addresses;
		_passwordFilePath = passwordFilePath;
		if(passwordFilePath == null)  {
			_passwordCheck = false;
		}
		else  {
			_passwordCheck = true;
		}		
	}
	
	/**
	* Get AddressTable from the Security object
	* @return AddressTable
	*/
	public AddressTable getAddressTable()  {
		return _addresses;
	}
	
	/**
	* Get password file path, assigned to the Security object
	* @return Password file path
	*/
	public String getPasswordFilePath()  {
		return _passwordFilePath;
	}
	
	/**
	* Set address table to the Security object
	* @param addresses AddressTable will be maintained by the Security object
	*/
	public void setAddressTable(AddressTable addresses)  {
		_addresses = addresses;
	}
	
	/**
	* Set the password file path
	* @param passwordFilePath
	*/
	public void setPasswordFilePath(String passwordFilePath)  {
		_passwordFilePath = passwordFilePath;
		_passwordCheck = true;
	}
	
	/**
	* Set the password checking bit. If true, password checking will be performed
	* @param value Boolean value to set password checking
	*/
	public void setPasswordCheck(boolean value)  {
		_passwordCheck = value;
	}
		
	/**
	* Connection is requested from the client. 
	* This will be override if new register protocol is added.
	* @param c ReceiverThread will be started
	*/
	public abstract void processServerLogin(ReceiverThread c); 
	
  /**
  * Opposite to the processServerLogin. Connection should be added to the
  * connection table, by caller.
  * @param c ReceiverThread, which will connect to the other server
  * @param serveraddress Other agent address where I want to connect
  */
  public abstract void processClientLogin(ReceiverThread c,Address serveraddress)
  	throws ConnectionException;

	/**
	* Send disconnect message to the server. The ReceiverThread is initiated
	* by me.
	* @param c ReceiverThread, which will stop.
	* @param myid My name, may be passed to the server to disconnect gracefully
	*/
  public abstract void processClientLogout(ReceiverThread c); 
  
  /**
  * Check whether the given agent is valid to connect to the Server
  * @param obj Object which dependent upon the security implementation
  * @return true if the agent is valid else, retur false
  */
  public abstract boolean isValidAgent(Object o);
  
}		
			
	
				