package BaseLayer;

import java.lang.*;
import java.util.*;
import java.io.*;

import Abstract.*;

/**
* Implementation of AddressTable in Abstract Layer.
* AddressTable class for store and retain Address.
* The static addresses can be stored to the AddressTable from file reading and
* dynamic addresses can be stored by addAddress method.
* You can extend this class to use 'description' field of Address more structured
* way. 
*/

/*
* Copyright (c) 1996,1997,1998 <A HREF="http://cdr.stanford.edu/">Center for Design Research</A>, Stanford University.
* All rights reserved.<p>
*
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You can obtain the GNU General Public License at
* http://www.gnu.org/copyleft/gpl.html or by writing to the Free
* Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
* 02111-1307, USA.
* 
* This software is bound by the terms and conditions listed in the
* attached <a href="../LICENSE">LICENSE file</A>.
*  @author <A HREF="http://cdr.stanford.edu/html/people/jeon-bio.html>Heecheol Jeon</A>
*  @version Java(tm) 1.1
*/ 

public class BAddressTable extends AddressTable   {

	/**
	* Default constructor
	*/	
	public BAddressTable()  {
	}
	
	/**
	* Construct AddressTable from file reading.
	* If you are going to use different file format other than provided here,
	* subclass AddressTable and override this constructor.
	* The Address file format is a collection of Addresses and each Address 
	* should be finished with '\n' and items are separated by ','. Therefore
	* you should not add '\n' and ',' unintentionally.<BR>
	* Each line's format is :<BR>
	* [Agent Name],[Internet Host Address],[Port No],[Message method for Connection],[String Description]\n
	* Example of AddressTable file format : <BR>
	* AgentNameA,java.stanford.edu,1234,ServerThread,(AServer Connection)\n
	* AgentNameB,aaa.bb.com,3456,ReceiverThread,(Sort Service)\n
	* AgentNameC,ccc.aol.com,21,FTPClient,MyFTPClientService\n
	* ..<BR>
  * @param in BufferedReader for reading Addresses
  * @exception IOException, ConnectionException
  */
	public BAddressTable(BufferedReader in) 
	throws IOException,ConnectionException  {
		String aline;
		while((aline = in.readLine()) != null)  {
			// comment line, starts with '#'
			if((!aline.startsWith("#")) && (!aline.equals("")))  {
				Address addr = new Address(aline);
				put(addr.getID(),addr);
			}	
		}	
	}
	
	/**
	* If the address table for the id is already exist, The address will be 
	* replaced with new one.
	* @param id : Agent Name
	* @param addr : Address object for the agent
	*/
	public void addAddress(Address addr) {
		String newid = addr.getID();
		if(containsKey(newid))  {
			remove(newid);
		}
		
		put(newid,addr);	
	}
		
	/**
	* Remove address from the table. If the geven agent name is not in the 
	* address table, nothing will happen
	* @param id Agent name want to be removed from the AddressTable
	*/
	public void removeAddress(String id)  {
		remove(id);
	}
			
	/**
	* Get an Address object using agent name
	* @param id Agent name
	* @return Address object, the agent has. Null if the address is not defined
	*/
	public Address getAddress(String id)  {
		Enumeration e = elements();
		while(e.hasMoreElements())  {
			Address already = (Address)e.nextElement();
			if(already.getID().equals(id))  {
				return already;
			}
		}		
		return null;
	}
	
	/**
	* Get an AddressTable of the same Type address
	* @param type Type definition(e.g. Message method) to retrieve
	* @return AddressTable
	*/
	public AddressTable getAddressTableWithType(String type)   {
		AddressTable atable = new BAddressTable();
		Enumeration e = this.elements();
		while(e.hasMoreElements())  {
			Address addr = (Address)e.nextElement();
			if(type.equals(addr.getType()))  {
				atable.put(addr.getID(),addr);
			}
		}
		return atable;
	}
}	
					
		
			