package BaseLayer;

import java.net.*;
import java.io.*;
import java.util.*;

import Abstract.*;

/**
*  BRecvThread is a subclass of the Receiver thread abstract
*  class. BRecvThread will take a core role to send/receive String
*  message. The received message will be saved to the MessageBuffer
*  with the type String. Typically,AgentAction or other class will
*  parse the String message and construct Message object.<BR>
*  BRecvThread is based on the design under the assumption :<BR>
*  <UL>
*	 <LI>Use message queue which might be centralized to AgentAction or distributed
*      to each BRecvThread
*  <LI>Message will be stored to MessageQueue with type String. Appropriate
*      object will be constructed lately. This approach separates Message class
* 		 from BRecvThread
*  <LI>Use ConnectionTable to control running threads
*  <LI>Use timer. The timer will be set if you invoke setDurationTime() method.
*	     Default timer is no setting(so that this thread will running until it is killed
* 		 by other object). Duration time unit is a minute.
*  <LI>ConnectionException will be thrown if error occur.
*	 <LI>The ReceiverThread name will be set as the agent name
*  <LI>Default Maximum length of a message is 16K
*  <LI>End message character will be set by setEndWith(char). Default is '\004'.
*  </UL>
* The constructor with socket parameter will be invoked by ServerThread. This
* means that the ReceiverThread will be initiated by other agent to this<BR>
* The constructor with host name and port parameter will be invoked if you
* initiate socket connection to the host and specified port.<BR>
* Use accessors to change default data member values such as setEndWith(char) or
* setMaxMessageSize(int) if necessary.<BR>
* From beta version, writeMsg(String) and setSecurity(Security) will throw
* Abstract.ConnectionException if IOException or Security fails.
* @see BaseLayer.BConnectionTable
*/

/*
* Copyright (c) 1996,1997,1998 <A HREF="http://cdr.stanford.edu/">Center for Design Research</A>, Stanford University.
* All rights reserved.<p>
*
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You can obtain the GNU General Public License at
* http://www.gnu.org/copyleft/gpl.html or by writing to the Free
* Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
* 02111-1307, USA.
*
* This software is bound by the terms and conditions listed in the
* attached <a href="../LICENSE">LICENSE file</A>.
*  @author <A HREF="http://cdr.stanford.edu/html/people/jeon-bio.html>Heecheol Jeon</A>
*  @version Java(tm) 1.1
*/

public class BRecvThread extends ReceiverThread {

  /**
   * ServerAddress, to which I want to connect to
   * @see Abstract.Address
   */
  protected Address _serverAddress;

  /**
  * Socket object from which the Receiver Thread send/receive message
  */
  protected Socket _client;


  /**
  * Input stream with type BufferedReader, derived from the Socket to receive message
  */
  protected BufferedReader _in;

  /**
  * Output stream with type BufferedWriter, derived from the socket to send message
  */
  protected BufferedWriter _out;
//  protected PrintWriter _out;


  /**
   * Timeout check when idle. Record the last sending/receiving time
   * If there's no communication during setting time, this object will
   * be stop by itself. Timer is not set until setDurationTime() method
   * is called.
   */
  protected long _lastTime;


  /**
   * true if ServerSocket acctept(server) false if I want to connect to other
   * server(client). Since this data is automatically set, you do not need
   * to set this member.
   */
  protected boolean _role = false;

  /**
   * Message buffer size. Default is 16K
   */
  protected int MAX_MSG_SIZE = 0x4000;

	/**
	  * Message header reserved for read bit and time stamp, not included in MAX_MSG_LEN
	  */
  protected int HEADER_LEN = 15;  // message header length, not included in message length

  /**
   * Default Constructor. You need member-wise initialization for all data members
   * using accessors - setSocket(Socket), setConnectionTable(ConnectionTable),
   * setMessageBuffer(MessageBuffer), setSecurity(Security). Optionally you can use
   * setEndWith(char) and setDurationTime(int minutes).
   */
  public BRecvThread()  {
    _client = null;
    _endWith = '\004';
    _connections = null;
    _queue = null;
    _in = null;
    _out = null;
    _security = null;
    _durationTime = -1;
  }

  /**
   * Create ReceiverThread with specific thread name. You need member-wise initialization for all data members
   * using accessors - setSocket(Socket), setConnectionTable(ConnectionTable),
   * setMessageBuffer(MessageBuffer), setSecurity(Security). Optionally you can use
   * setEndWith(char) and setDurationTime(int minutes).
   * @param name Agent name
   */
  public BRecvThread(String name)  {
    super(name);
    _client = null;
    _endWith = '\004';
    _connections = null;
    _queue = null;
    _in = null;
    _out = null;
    _security = null;
    _durationTime = -1;

  }

  /**
   * Create ReceiverThread with specific thread name and Thread group.
   * Passing thread group is essential when you will create the receiver thread for an applet
   * because of security issue. You need member-wise initialization for all data members
   * using accessors - setSocket(Socket), setConnectionTable(ConnectionTable),
   * setMessageBuffer(MessageBuffer), setSecurity(Security). Optionally you can use
   * setEndWith(char) and setDurationTime(int minutes).
   * @param tg Thread group to which this thread will be owned to
   * @param name Agent name
   */
  public BRecvThread(ThreadGroup tg,String name)  {
    super(tg,name);
    _client = null;
    _endWith = '\004';
    _connections = null;
    _queue = null;
    _in = null;
    _out = null;
    _security = null;
    _durationTime = -1;
  }


  /**
   *  Constructor, to be called from the serverthread after accepting
   *  a connection. Note default endWith is '\004'. You can change endWith
   *  character by setEndWith(char) method. Note that the timer will not
   *  be set. To set timer, use setDurationTime(int minutes) method.
   *  <I>Do not use this thread when you want to initiate(conect) to
   *  other server. If you use this constructor the role will be set
   *  as a server.</I> If there is a special reason you should use
   *  this constructor even if your agent role is a server, use setRole(false)
   *  method to set the role as a client.
   * @param client_scoket The socket for the remote client;
   * @param priority the priority of the thread.
   * @param address My address.
   * @param connections the connection table is passed as an argument.
   * @param queue the Message Queue.
   * @param security Security for checking this connection security
   * @exception ConnectionException if error occurs
   */
  public BRecvThread(Socket client_socket,
		     int priority,
		     Address address,
		     ConnectionTable connections,
		     MessageBuffer queue,
		     Security security)
    throws ConnectionException    {
      _myaddress = address;
      _role = true;  // server
      setPriority(priority);
      setEndWith('\004');
      _durationTime = -1;

      // setConnection method will kill the same name connection
      setConnectionTable(connections);
      setMessageQueue(queue);
      setSocket(client_socket);
      setSecurity(security);
  }


  /**
   *  This constructor is called when an agent want to initialize a connection
   *  to the other agents, using host name and port. This constructor is
   *  will be useful if you do not want to use any security check._security
   *  data member will be set as null. Do not use this constructor if you
   *  want security check.
   *  @param serverid Agent name to shich the this receiver thread will connect
   *  @param host the name of the remote host.
   *  @param priority the priority of the thread.
   *  @param connections the connection table is passed as an argument.
   *  @param queue the Message Queue.
   *  @param port the port of the remote host.
   *  @param tg Thread group to which this thread will be owned to. Necessary to evade applet security.
   * @exception ConnectionException if error occurs
   */
  public BRecvThread(String serverid,
		     String host,
		     int priority,
		     Address myaddress,
		     ConnectionTable connections,
		     MessageBuffer queue,
		     int port,
		     ThreadGroup tg)
    throws ConnectionException   {
      super(tg,serverid);
      try  {
	_myaddress = myaddress;
	_role = false; // client
	setPriority(priority);

	setEndWith('\004');
	_durationTime = -1;

	setConnectionTable(connections);

	setMessageQueue(queue);
	setSocket(new Socket(host,port));

	_security = null;
      } catch (Exception e)  {
	throw new ConnectionException(ConnectionException.RECEIVERINIT + " " + host + ":" + port);
      }
  }

  /**
   *  This constructor is called when an agent want to initialize a connection
   *  to the other agents, using Address. Use this constructor if you want
   *  security check.
   *  @param serveraddress The address of the remote host.
   *  @param priority The priority of the thread.
   *  @param myaddress My addres
   *  @param connections the connection table is passed as an argument.
   *  @param queue the Message Queue.
   *  @param security The security object
   * @exception ConnectionException if error occurs
   */
  public BRecvThread(Address serveraddress,
		     int priority,
		     Address myaddress,
		     ConnectionTable connections,
		     MessageBuffer queue,
		     Security security,
		     ThreadGroup tg)
    throws ConnectionException   {
      this(serveraddress.getID(),serveraddress.getHost(),priority,myaddress,connections,queue,serveraddress.getPort(),tg);
      _serverAddress = serveraddress;

      setSecurity(security);
  }


  /**
   * Set client socket. This method will generate input(BufferedReader), output(BufferedWriter) stream
   * @param socket client socket
   */
  public void setSocket(Socket socket)
    throws ConnectionException    {
      _client = socket;
      try  {
    	_in = new BufferedReader(new InputStreamReader(_client.getInputStream()));
//	_out = new PrintWriter(new BufferedWriter(new OutputStreamWriter(_client.getOutputStream())));
			_out = new BufferedWriter(new OutputStreamWriter(_client.getOutputStream()));

      }	catch (IOException e)  {
    	throw new ConnectionException(ConnectionException.INPUTOPEN + " " + toString());
      }
  }

  /**
   * Set maximum message size in character length. Default is 16K
   * @param newMaxLen Maximum unit message length
   */
  public void setMaxMessageSize(int newMaxLen)  {
    MAX_MSG_SIZE = newMaxLen;
  }

  /**
   * Get maximum allowable message length. Default is 16K
   * @return Maximum unit message size
   */
  public int getMaxMessageSize()  {
    return MAX_MSG_SIZE;
  }

  /**
   * Set input stream
   * @param input BufferedReader
   */
  public void setInputStream(BufferedReader input)  {
    if(_in != null)  {
      try  {
	_in.close();
      } catch (IOException e)  {}
    }
    _in = input;
  }

  /**
   * Set output stream
   * @param output DataOutputStream
   */
//  public void setOutputStream(PrintWriter output)  {
  public void setOutputStream(BufferedWriter output)  {

    if(_out != null)  {
      try  {
	_out.close();
      } catch (Exception e)  {}
    }
    _out = output;
  }

  /**
   * Set duration time. Note that the duration time is not passed to the
   * constructor. You can use setDurationTime(int) after you construct
   * this object. Default is no time set.
   * @param minutes Maximum duration time to be set(unit is minute)
   */
  public void setDurationTime(int minutes)  {
//    if(minutes == -1)  return;
    _lastTime = getCurrentTimeMillis();
    _durationTime = minutes;
  }

  /**
   * Accessor to get client socket
   * @return Receiving socket
   */
  public Socket getSocket()  {
    return _client;
  }


  /**
   * Accessor to get input stream
   * @return input stream
   */
  public BufferedReader getInputStream()  {
    return _in;
  }

  /**
   * Accessor to get output stream
   * @return output stream
   */
//  public PrintWriter getOutputStream()  {
  public BufferedWriter getOutputStream()  {

    return _out;
  }

  /**
   * Accessor to get duration time
   * @return Maximum idle time.
   */
  public long getDurationTime()  {
    return _durationTime;
  }

  /**
   * Accessor to get last communication time
   * @return Most recent time, sending/receiving messages
   */
  public long getlastTime()  {
    return _lastTime;
  }

  /**
   * Set role. If true is passed, its role will be a server. The default is that if you
   * initiate the connection to the other server, its role is a client(false).
   * @param role Boolean value to set role. If true, role is a server
   */
  public void setRole(boolean role)  {
    _role = role;
  }

  /**
   * Get Receiver's role. Return true if the role is a server
   * @return True if the role is a server
   */
  public boolean getRole()  {
    return _role;
  }


  /**
   * Because of synchronization problem, the client side connect, registration
   * will not be working as a separate thread. Therefore if the receiver thread
   * is created as a client side, security check will be performed at this method.
   * Fail on security check will close socket and its input/output streams.
   * @param security. Security object
   * @exception Abstract.ConnectionException If security check fails, Connection
   * exception will be thrown
   */
  public synchronized void setSecurity(Security security) throws ConnectionException {
    super.setSecurity(security);

    if(_security != null)  {
      // client
      if(_role == false)  {
	try  {
	  _security.processClientLogin(this,_serverAddress);
	}  catch (ConnectionException e)  {
		closeSocket();
	  throw new ConnectionException(e.toString());
	}
      }
    }
  }

  /**
   * Override Thread run() method
   * It will first perform the Security check. If passed, it will started and
   * receive messages. If fails, the Security will kill this thread.
   * Receive a message and store it to MessageBuffer
   * Rather than read a message character by character(this need read() call
   * repeatedly and conversion from int to char or byte to char), use
   * buffer to read full available input. Since the full read input may
   * have several messages, it will be divided into several messages
   * and saved to the message queue as String object.
   * The assumption here is that the maximum message length is less than
   * 16K.
   */
  public void run()   {
    if(_security != null)  {
      // from server side, connection will be performed as asynchronously
      if(_role == true)  {
	_security.processServerLogin(this);
      }
    }


    System.out.println("Client " + getName() + " running");
    char[] buffer = new char[MAX_MSG_SIZE];
    int readStart = 0;

    while(true) {
      try {

	if(_in.ready()) {
		if(readStart >= (MAX_MSG_SIZE - 1))  {
			// Message size is too big. End condition
			endConn();
		}

	  // read full unblocked messages
	  int readLen = readBuffer(buffer,readStart,MAX_MSG_SIZE - readStart);

	  // if something is read
	  if(readLen != -1)  {
	    int startPosition = 0;
	    while(true)  {
	      int unitLen = findEndLocation(buffer,startPosition,readStart + readLen);
	      if(unitLen != -1)  {
		_queue.addElement(new String(buffer,startPosition,unitLen));
		startPosition = startPosition + unitLen + 1;
	      }
	      else  {
		readStart = readStart + readLen - startPosition;
		// shift to the left
		for(int i=0; i< readStart; i++)  {
		  buffer[i] = buffer[i + startPosition];
		}
		break;
	      }
	    }
	  }
	  resetTime();
	}
	else  {
	  sleep(1000);
	  checkTime();
	}
      } catch (Exception e)  {
	System.out.println("IOException " + getName());
	endConn();
      }
    }
  }


  /**
   * Read unblocked stream from the Socket and save it to the buffer.
   * @param buffer Socket buffer
   * @param start Starting point to read
   * @param max Maximum length to read
   * @return Message length. -1 if the current buffer does not end with endWith character
   */
  protected synchronized int readBuffer(char[] buffer,int start, int max)  {
    int readLen = -1;
    try  {
      readLen = _in.read(buffer,start,max);
    } catch (IOException e)  {
      endConn();
    }
    return readLen;
  }

  /**
   * Get one message. This method is useful to setup initial connection.
   * But we are suggesting that do not use this method actual, heavy loaded
   * message passing since the method is not efficient.
   * Instead to add messages to message queue, it will return the received
   * message. This method is used only in the Security check stage.
   * @return Received message
   */
  public String getOneMessage()   {
    String msg = "";
    try  {
      int i = 0;
      while(true)  {
				if(_in.ready())  {
	  			char c = (char)_in.read();
	  			i++;

	  			// error. not ended with.
	  			if(i > MAX_MSG_SIZE) {
	    			return "";
	  			}

					if(c == _endWith)  {
						return msg;
					}

    			msg = msg + c;
				}
				else  {
	  			// yield may cause CPU leackage
	  			try {
	    			sleep(100);
	  			}  catch (InterruptedException ee) {
	  				System.out.println("Interrupted");
	  			}
				}
      }
    } catch (IOException e)  {
    	System.out.println(e.toString());
    }

    return msg;
  }

  /**
   * Find endWith location and return the unit message length
   * startPosition will be updated.
   * @param buffer message buffer, read from the socket
   * @param startPosition start position to search end character
   * @param maxLen maximum allowable message length. Full length of the stream
   * @return unit message length. If the message is not ended with _endWith, it
   * will return -1
   */
   protected int findEndLocation(char[] buffer,
                                int startPosition,
                                int maxLen)  {
    int unitLen = 0;

    while(startPosition < maxLen)  {
      if(buffer[startPosition] != _endWith)  {
        startPosition++;
        unitLen++;
      }
      else  {
        return unitLen;
      }
    }
    return -1;
  }

  /**
   * The thread just writes on the connection for the remote side. Close socket and
   * kill this thread not performed here. After catching the exception, you need
   * to invoke endConn() method to clean up.
   * @param writebuffer: the message as a normal ASCII string.
   * @exception Abstract.ConnectionException if IO error occurs.
   */
  public synchronized void writeMsg(String writebuffer)
  throws ConnectionException   {

		if((writebuffer.length() + 1) >= (MAX_MSG_SIZE - HEADER_LEN))  {
			throw new ConnectionException(ConnectionException.MSGLENGTH);
		}

    try  {

  		// why flushing twice?
  		// java does not catch IOException for the first flushing,but
  		// the second time, it will throw IOException. The reason is we want to
  		// know IOException, not to lose any message.
  		_out.write(writebuffer);
      _out.flush();
      _out.write(_endWith);
      _out.flush();
      resetTime();
    } catch (Exception e)  {
      throw new ConnectionException(ConnectionException.OUTPUTCLOSE);
    }
  }

  /**
   * Check time out. If the timer is not set(no timer setting is default), no action,
   * If idle time is longer than the setting time, this thread will be stop
   */
  protected void checkTime()   {
    if(_durationTime == -1) return;
    long ctimevalue = getCurrentTimeMillis();
    if(((ctimevalue - _lastTime)/60000) > _durationTime)  {
      System.out.println("Timeout");
      endConn();
    }
  }

  /**
   * Reset the timer to current time.
   */
  protected void resetTime()  {
    if(_durationTime != -1)  {
      _lastTime = getCurrentTimeMillis();
    }
  }

  /**
   * Set the server address to connect
   *	@param addr Server address(other agent address)
   */
  public void setServerAddress(Address addr)  {
    _serverAddress = addr;
  }

  /**
   * Get current Server address
   * @return Server addnress to which this receiver thread is connected
   */
  public Address getServerAddress()  {
    return _serverAddress;
  }

  /**
   * Close the connection. Automatically update the connectionTable. Processing the
   * logout message to the other end connection. Remove myself from the ConnectionTable
   * Close socket,input stream and output stream. Stop this thread.
   */
  public synchronized void endConn()   {
    cleanUp();
    processLogoutMessage();
    closeSocket();
  }

  /**
   * Remove this thread from the connection table
   */
  protected void cleanUp()  {
    if(_connections != null)  {
      _connections.remove(getName());
      _connections = null;
    }
  }

  /**
   * Close all the connections and stop this thread
   */
  protected void closeSocket()  {
    try  {
      if(_in != null)  {
System.out.println("Input close");
         _in.close();
	 _in = null;
      }
System.out.println("Output close");

      if(_out != null)  {
	_out.close();
	_out = null;
      }
System.out.println("Socket close");

      if(_client != null)  {
	_client.close();
	_client = null;
      }

    }  catch (IOException e)  {
      System.out.println(e.toString());
    }

    if(isAlive())  {
System.out.println("Receiver Thread <" + getName() + "> stop");

      stop();
    	try  {
      	join();
    	} catch (InterruptedException e3)  {}
    }

  }

  /**
   * Process logout message. Send disconnect message to the other end connection
   * The disconnect message will be sent only once.
   */
  public void processLogoutMessage()   {
    if(_security != null)  {
      _security.processClientLogout(this);
    }
  }

//  public void checkAccess() {}

  public long getCurrentTimeMillis()  {
  	Date time = new Date();
  	return time.getTime();
  }
}
