package KQMLLayer;

import BaseLayer.*;
import Abstract.*;

/**
* Secure the connection. Check whether the register,unregister protocol
* is satisfied. No password check will be performed. You can extend Security
* class to perform more complex security check include password check.
* Typical process for the KQML agents' protocol is :<BR>
* At first, upon request, ServerSocket(BServerThread) will accept()
* and introduce itself. Then, the client will receive the introduction and
* send the client's introduction. If the message is legal, the Security
* will send accepted message.<BR>
* <UL>
* <LI>Server send message to the client<BR>
* (connected :sender server :receiver client :host java.stanford.edu 
* :port 1234 :type BServerThread :description (Example test))<BR>
* <LI>Client will response<BR>
* (register :sender client :receiver server :host aaa.ggg.com 
* :port 4321 :type BRecvThread :description (Client test))<BR>
* <LI>Server will send message<BR>
* <LI>(accepted :sender server :receiver client)<BR>
* </UL>
* The address table for this class should be defined. When you construct
* your agent, the address table should be created before construct
* the KQMLSecurity object.To pass address table for the KQMLSecurity
* constructor is preferable not to forget set address table to the
* KQMLSecurity object.
*/

/*
* Copyright (c) 1996,1997,1998 <A HREF="http://cdr.stanford.edu/">Center for Design Research</A>, Stanford University.
* All rights reserved.<p>
*
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You can obtain the GNU General Public License at
* http://www.gnu.org/copyleft/gpl.html or by writing to the Free
* Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
* 02111-1307, USA.
* 
* This software is bound by the terms and conditions listed in the
* attached <a href="../LICENSE">LICENSE file</A>.
*  @author <A HREF="http://cdr.stanford.edu/html/people/jeon-bio.html>Heecheol Jeon</A>
*  @version Java(tm) 1.1
*/ 

public class KQMLSecurity extends Security  {

	/**
	* Return value from secureAgent method. Client satisfies the protocol
	*/
	static final int CLIENTOK = 0;
	
	/**
	* Return value from secureAgent method. Client misses its name in the connection request
	*/
	static final int CLIENTNAMEMISSING = 1;
	
	/**
	* Return value from secureAgent method. Client's name is the same as mine.
	*/
	static final int CLIENTSAMENAME = 2;

	/**
	* Default constructor
	*/
	public KQMLSecurity()  {
		super();
	}
	
	/**
	* Constructor with AddressTable
	* @param addresses AddressTable , which will be maintained by this.
	*/
	public KQMLSecurity(AddressTable addresses)  {
		super(addresses);
	}

	/**
	* Override processServerLogin since KMQL use different register process
	* By request, server will kindly introduce itself. Then wait for the
	* client introduction. If it matches the protocol, receiver thread
	* will be started. If the first message from client has not 'register'
	* performative, accession will be denied. If the first message's content
	* does not satisfy the security check, this method will give a chance
	* to the client to correct. If the corrected response is still wrong,
	* access will be denied. No password cross-check will be performed.
	* @param con BRecvThread, will be started
	*/
	public void processServerLogin(ReceiverThread con)   {
	
		// cross check will be performed. By sending host address and port
		// the client side will check. If correct the client will send 
		// register message to server
		Address address = con.getAddress();
		
		try  {
			// introduce me without specifying description field
			String loginString = "(connected :host ";
			loginString = loginString + address.getHost() + " :port " + Integer.toString(address.getPort()) + " :sender " + address.getID() + " :type BRecvThread ";
			String description = address.getDescription();
			if(description != null)  {
				loginString = loginString + " :description " + description;
			}
			loginString = loginString + ")";	

			con.writeMsg(loginString);
		
			String response = con.getOneMessage();	
			KQMLmessage kqml = new KQMLmessage(response);
			String perf = kqml.getValue("performative");
			perf = perf.toLowerCase();
		
			// response will be 
			// (register :sender agent :receiver xx :host kk :port *** :description...)
			if(perf.equals("register"))  {
				String agentid = kqml.getValue("sender");
				String myid = kqml.getValue("receiver");
				String host = kqml.getValue("host");
				String port = kqml.getValue("port");
				String type = kqml.getValue("type");
				description = kqml.getValue("description");
		
				int checkresult = secureAgent(myid,agentid);
				if((checkresult == CLIENTOK) ||
					 (processErrorLogin(checkresult,con,myid))) {
					_addresses.addAddress(new Address(agentid,host,(Integer.valueOf(port)).intValue(),type,description));
					startConnection(con,agentid,myid);
		    }		
		    else  {
		    		denyConnection(con,myid);
	    	}
		  }
		}  catch (Exception e)  {
			System.out.println(e.toString());
			con.endConn();
		}
	}					

	/**
	* Process error state. If the error is CLIENTNAMEMISSING or CLIENTSAMENAME, it will
	* give one more chance to get the response from the end connection. If still fails
	* it will close the connection and kill the ReceiverThread
	* @param errorCode error code
	* @param con ReceiverThread, which is checked
	* @param myid My name in agent society
	*/
	protected boolean processErrorLogin(int errorCode,ReceiverThread con,String myid)   {		  
		try  {
			String error = "(ask-one :sender ";
			error = error + myid + " :content ";
		    
			switch(errorCode)   {
   		case CLIENTNAMEMISSING:
	   		error = error + " name :reply-with name)";
   			break;
   		case CLIENTSAMENAME:
	   		error = error + " dupname :reply-with newname)";
    	}		
    	// send error
    	con.writeMsg(error);
	
			String msg = con.getOneMessage();
			KQMLmessage kqml = new KQMLmessage(msg);
			String perf = kqml.getValue("performative");
			if(perf.equals("reply"))  {
				String agentid = "";
				String replywith = kqml.getValue("reply-with");
				if((replywith.equals("name")) || (replywith.equals("newname")))  {
					agentid = kqml.getValue("content");
					if(agentid.equals("") || (agentid == null))  {
						return false;
					}	
				}
					
				if(secureAgent(agentid,myid) == CLIENTOK)   {
					return true;
				}
				else  {
					return false;
				}
			}
		}		catch (Exception e)  {
			System.out.println(e.toString());
			con.endConn();
		}
		return false;
	}				

	/**
	* Start connection. All protocol is satisfied. Add the ReceiverThread to the
	* ConnectionTable
	* @param con ReceiverThread, to be checked
	* @param agentid Other agent name
	* @param myid My name
  * @exception Abstract.ConnectionException If write a message to the connection fails
	*/
	protected void startConnection(ReceiverThread con,String agentid,String myid)
	throws ConnectionException	  {
		System.out.println("Connection Request from " + agentid);				
		String access = "(accepted :sender ";
		access = access + myid + " :receiver " + agentid + ")";
		
		con.writeMsg(access);	
   	con.getConnectionTable().addConnection(agentid,con,true);
  } 	
  
  /**
  * Deny connection since error. Send error message and disconnect
  * @param con ReceiverThread, to be killed
  * @param id Other agent name
  * @exception Abstract.ConnectionException If write a message to the connection fails
  */
  protected void denyConnection(ReceiverThread con,String id) 
  throws ConnectionException {
		String errormsg = "(error :sender ";
		errormsg = errormsg + id + " :content register-expected)";
  
		con.writeMsg(errormsg);
		con.endConn();
  }
	
	/**
	* Check the client send its name and the name is different from mine
	* No password check.
	* @param me My name, sent back from the client
	* @param client Client agent name
	*/
	protected int secureAgent(String me,String client)  {

		if(client == null) return CLIENTNAMEMISSING;
		// client can not have the same name with me
		if(client.equals(me)) return CLIENTSAMENAME;
		return CLIENTOK;
	}	
	
  /**
  * Opposite to the processServerLogin. Connection should be added to the
  * connection table, by caller. The KQMLTCPRecvThread generator should
  * specify the Server name and Password. The description 
  * @param c BRecvThread, which will connect to the other server
  * @param serveraddress Server address to whom I want to connect
  */
  public void processClientLogin(ReceiverThread c,Address serveraddress) 
  	throws ConnectionException {
  	if(_addresses == null)  {
 			System.out.println("No AddressTable is defined. Client connection failed");
			throw new ConnectionException("No AddressTable is defined. Client connection failed");
  	}
  					
System.out.println("Me:" + c.getAddress().getID() + " try to connect to " + serveraddress.getID());  
  	// My address should be specified
  	Address myaddress = c.getAddress();
  	if(myaddress == null)   {
  		System.out.println("Address for " + c.getAddress().getID() + " is not specified. Fail to client connect");
  		throw new ConnectionException("Address for " + c.getAddress().getID() + " is not specified. Fail to client connect");
  	}
  	
  	if(serveraddress == null)   {
  		System.out.println("Address for " + c.getAddress().getID() + " is not specified. Fail to client connect");
  		throw new ConnectionException("Address for " + c.getAddress().getID() + " is not specified. Fail to client connect");
		}  		
  			
  	String initial = c.getOneMessage();
System.out.println(initial);  	
  	try  {
  		KQMLmessage kqml = new KQMLmessage(initial);
  	
	  	String perf = kqml.getValue("performative");
  		if(perf.equals("connected"))  {
	  		String loginString = "(register :host ";
				loginString = loginString + myaddress.getHost() + " :port " + Integer.toString(myaddress.getPort()) + " :sender " + myaddress.getID() + " :type BRecvThread ";
				loginString = loginString + " :receiver " + serveraddress.getID();
			
				// if you need password check, your password for accessing to the server
				// and server password you will give to the server should be specified
				String description = serveraddress.getDescription();
				if(description != null)  {
					loginString = loginString + " :description " + description;
				}
				loginString = loginString + ")";	
			
				c.writeMsg(loginString);  		
				String response = c.getOneMessage();
System.out.println(response);
				KQMLmessage respkqml = new KQMLmessage(response);
				String respperf = respkqml.getValue("performative");
				if(respperf.equals("accepted"))  {		
System.out.println(serveraddress.getID() + " receiver is started");	  			
					// success
					c.getConnectionTable().addConnection(serveraddress.getID(),c,false);	
	  		}
	  		else  {
	  			throw new ConnectionException(respkqml.getReadString());
	  		}		
  		}		
  		else
  		if(perf.equals("error"))  {
  			System.out.println("Connection failed for " + c.getAddress().getID() +  ". Check Your Agent Name is correct");
  			throw new ConnectionException("Connection failed for " + c.getAddress().getID() +  ". Check Your Agent Name is correct");
  		}	
  	} catch (Exception e)  {
  		System.out.println(e.toString());
  		throw new ConnectionException("Security throw error " + e.toString());
  	}		
	}
	
	/**
	* Send disconnect message to the BRecvThread for some reasons such as
	* timeout, or reboot the Server,etc. This method will be invoked by 
	* BRecvThread or AgentAction. The processClientLogin does not kill the ReceiverThread
	* @param c BRecvThread to be disconnected
	*/
	public void processClientLogout(ReceiverThread c)  {
		String logout = "(disconnected :sender ";
		logout = logout + c.getAddress().getID() + " :receiver " + c.getName() + ")";
		
		try  {
			c.writeMsg(logout);
		} catch (Exception e) {}	
	}	
		
  /**
  * No checking on the validity, user name check
  */
  public boolean isValidAgent(Object o) {
  	return true;
  }
		
}		