package KQMLLayer;

import java.util.*;
import java.lang.*;
import java.io.*;

/**
* KQMLmessage class.
* Basically KQMLmessage is a Hashtable with key field as slot name and field value
* as slot value. The peformative key is assumed key as performative. You can retrieve
* slot value using getValue(String slotname). If you want to retrieve performative,
* use String perf = kqmlmessage.getValue("performative") where kqmlmessage is instantce
* of KQMLmessage. The KQML parse has limitations:<BR>
* <UL>
* <LI>It does not allow unpaired opened,closed parenthesis.<BR>
* Therefore, string (tell :sender xx :receiver yy :content ((error) will throw ParseException
* since the message is not closed with paired parenthesis<BR>
* <LI>It does not distinguish the quoted string><BR>
* Therefore, in (tell :sender "me"), getValue("sender") is "me", not me.
* </UL>
*/

/*
* Copyright (c) 1996,1997,1998 <A HREF="http://cdr.stanford.edu/">Center for Design Research</A>, Stanford University.
* All rights reserved.<p>
*
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You can obtain the GNU General Public License at
* http://www.gnu.org/copyleft/gpl.html or by writing to the Free
* Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
* 02111-1307, USA.
* 
* This software is bound by the terms and conditions listed in the
* attached <a href="../LICENSE">LICENSE file</A>.
*  @author <mailto:raquet@cdr.stanford.edu>Andreas Raquet</A>
*  @version Java(tm) 1.1
*/ 
public class KQMLmessage extends Hashtable {

	/**
	* Default constructor
	*/ 
  public String performative;
  public KQMLmessage () {
  }

	/**
	* Given string, create KQMLmessage object
	* @param message String which has legal KQML syntax
	* @exception KQMLLayer.ParseException if syntax does not match.
	*/
  public KQMLmessage(String message) throws ParseException    {
    super();	
    parseMessage(message);	
  }

  public KQMLmessage(KQMLStreamTokenizer message) throws ParseException    {
    super();	
    parseMessage(message);	
  }

  /**
   * Parse message according to the KQML syntax
   * @param str KQML syntax string
   * @exception KQMLLayer.ParseException if syntax does not match
   */

  public void parseMessage(String str) throws ParseException  {
    try {
    KQMLStreamTokenizer parser = new KQMLStreamTokenizer(new BufferedReader(new StringReader(str)));
    parseMessage(parser);
    } catch (Exception e) {
      e.printStackTrace(); };
};
  public void parseMessage(KQMLStreamTokenizer parser) throws ParseException  {
    try {
      parser.consume('(');
      performative = parser.getNextToken();
      addFieldValuePair("performative",performative);
      while (!parser.viewToken().equals(")")) {
	String keyword = parser.getNextToken();
	KQMLExpression expression = KQMLExpression.parse(parser);
	addFieldValuePair(keyword.substring(1,keyword.length()),expression.toString());};
      parser.consume(')');}
    catch (Exception e) { 
      e.printStackTrace();
      throw new ParseException(); }; 
  } 
  
  /**
   * add field, value pair to construct a KMQLmessage
   * @param field slot name
   * @value value slot value
   */
  public void addFieldValuePair(String field, String value) {
  	String f = field.toLowerCase();
    put(field, value);
  }

  public void addFieldExpressionPair(String field, KQMLExpression value) {
    put(field, value);
  }

  /**
   * Returns the value of the "field" field. If it does not exit, returns null.
   * @return String representing field value.
   */
  public String getValue(String field) {
  	String f = field.toLowerCase();
	Object ret = get(field);
	if (ret!=null) 
	  return ret.toString();
	else return null;
  }	

  public KQMLExpression getExpression(String field) {
    Object o = get(field);
    if (o instanceof KQMLExpression) return (KQMLExpression)o;
    else return null;
  }	
  
  /**
   * Returns the KQML message as a single human readable string.
   * @return Message in string format.
   */
  public String getReadString() {
    String answer = "";
    Enumeration fields = keys();

    while(fields.hasMoreElements()){
      String field = (String)fields.nextElement();
      answer = answer.concat(field + ": " + getValue(field) + "\r\n");
    }
    return answer;
  }
  
  /**
   * Returns the KQML message as a single machine readable string.
   * This is the format which is sent to other agents.
   * @return Message in string format.
   */
  public String getSendString () {
    String answer = "(" + getValue("performative");
    Enumeration fields = keys();

    while(fields.hasMoreElements()){
      String field = (String)fields.nextElement();
      if(!field.equals("performative")){
  	  	answer = answer.concat(" :" + field + " " + getValue(field));
      }
    }
    answer = answer.concat(")");
    return answer;
  }
  
}