package ProtocolLayer.Example.CAD;


import java.awt.*;
import java.io.*;

/*
* Copyright (c) 1996,1997,1998 <A HREF="http://cdr.stanford.edu/">Center for Design Research</A>, Stanford University.
* All rights reserved.<p>
*
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You can obtain the GNU General Public License at
* http://www.gnu.org/copyleft/gpl.html or by writing to the Free
* Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
* 02111-1307, USA.
*
* This software is bound by the terms and conditions listed in the
* attached <a href="../LICENSE">LICENSE file</A>.
*  @author <A HREF="http://cdr.stanford.edu/html/people/jeon-bio.html>Heecheol Jeon</A>
*  @version Java(tm) 1.1
*/

public abstract class CADObject implements Cloneable   {

	// orientation, origin
	Matrix3D _mat;
	Matrix3D _initConfig;

	Point3d[] _vertex;
	IPoint3d[] _point;
	float _dimX,_dimY,_dimZ;

	// vertex number
	int _totalNo;

	// transfer matrix
	Matrix3D _tmat;
	Color _color;

	// bounding box
	Rectangle _bbox;

	static final int BOX = 1;
	static final int CYLINDER = 2;
	static final int CONE = 3;
	static final int SPHERE = 4;
	static final int EXTRUSION = 5;
	static final int AXISSYM = 6;

	int _type = 0;

	boolean _isSelected = false;

	public CADObject()  {
		_mat = new Matrix3D();
		_tmat = new Matrix3D();
		_color = new Color(0,0,0);
		_initConfig = (Matrix3D)_mat.clone();
		_dimX = 0;
		_dimY = 0;
		_dimZ = 0;
		_totalNo = 0;
	}

	// default origin(0,0,0) and orientation
	public CADObject(float width, float height, float depth)  {
		this();
		_dimX = width;
		_dimY = height;
		_dimZ = depth;
		setVertex();
	}


	public CADObject(float width, float height, float depth, Point3d origin, Point3d rotation)  {
		this(width,height,depth);
		rotate(rotation);
		translate(origin);
		_mat.copyTo(_initConfig);
	}

	public CADObject(float width, float height, float depth, float ox,float oy,float oz, float theatx,float thetay, float thetaz)  {
		this(width,height,depth);
		rotate(theatx,thetay,thetaz);
		translate(ox,oy,oz);
		_mat.copyTo(_initConfig);
	}

	public CADObject(float[] matelm, float width, float height, float depth)  {
		this(width,height,depth);
		_mat.setElement(matelm);
		_mat.copyTo(_initConfig);
	}

	public CADObject(Matrix3D mat,float width, float height, float depth)  {
		this(width,height,depth);
		mat.copyTo(_mat);
		mat.copyTo(_initConfig);
	}

	// returns original configuration. Width,depth,height is current one
	public Object clone()  {
		Box b = new Box(_initConfig,_dimX,_dimY, _dimZ);
		b.setColor(_color);
		b.setType(_type);
		return b;
	}

	public void setType(int type)  {
		_type = type;
	}

	public int getType()  {
		return _type;
	}

	public Color getColor()  {
		return _color;
	}

	public Matrix3D getMatrix()  {
		return _mat;
	}

	public void setColor(Color c)  {
		_color = c;
	}

	// translation
	public void translate(Point3d to)  {
		_mat.translate(to._x,to._y,to._z);
	}

	public void translate(float x, float y, float z)  {
		_mat.translate(x,y,z);
	}

	public void rotate(Point3d p)  {
		rotate(p._x,p._y,p._z);
	}

	// given angle x,y,z axiz, rotate
	public void rotate(float xtheta,float ytheta, float ztheta)  {
		_tmat.unit();
		_tmat.xrot(xtheta);
		_tmat.yrot(ytheta);
		_tmat.zrot(ztheta);
		_mat.mult(_tmat);
	}

	public void scale(float factor)  {
		_mat.scale(factor);
	}

	public void scale(float x,float y,float z)  {
		_mat.scale(x,y,z);
	}

	public boolean contains(Point p)  {
		Rectangle rect = getBoundingBox();
		return rect.contains(p);
	}

	public void setSelected(boolean selected)  {
		_isSelected = selected;
	}

	public void setOrientation(float xtheta,float ytheta,float ztheta)  {
		_tmat.unit();
		_initConfig.unit();
		_tmat.xrot(xtheta);
		_tmat.yrot(ytheta);
		_tmat.zrot(ztheta);
		_initConfig.mult(_tmat);
	}

	public void setOrigin(float ox,float oy, float oz)  {
		_initConfig.xo = ox;
		_initConfig.yo = oy;
		_initConfig.zo = oz;
	}

	public Point3d getOrientation()  {
		Point3d p = new Point3d();

		float roty = _initConfig.zx;
		double thetay = Math.asin(-roty);  // x-rotation
		double thetax=0,thetaz = 0;

		if(Math.abs(thetay - Math.PI / 2) < 0.000000001)  {
			// thetay is 90 degree. Only relative difference between thetax, thetaz is important.
			thetaz = Math.acos(_initConfig.yy);
		}
		else  {
			double rotxc = Math.cos(thetay);
			thetaz = Math.asin(_initConfig.yx / rotxc);
			thetax = Math.asin(- _initConfig.zy / rotxc);
		}

		p._x = (float)(thetax * 180 / Math.PI);
		p._y = (float)(thetay * 180 / Math.PI);
		p._z = (float)(thetaz * 180 / Math.PI);

		return p;
	}

	public Point3d getOrigin()  {
		Point3d p = new Point3d();
		p._x = _initConfig.xo;
		p._y = _initConfig.yo;
		p._z = _initConfig.zo;
		return p;
	}

	public Rectangle getBoundingBox()  {
		int minx = 1000000;
		int maxx = -1000000;
		int miny = 1000000;
		int maxy = -1000000;

		for(int i=0; i < _totalNo; i++)  {
			if(minx >= _point[i]._x)  {
				minx = _point[i]._x;
			}

			if(miny >= _point[i]._y)  {
				miny = _point[i]._y;
			}

			if(maxx <= _point[i]._x)  {
				maxx = _point[i]._x;
			}

			if(maxy <= _point[i]._y)  {
				maxy = _point[i]._y;
			}
		}

		return new Rectangle(minx,miny,maxx-minx,maxy-miny);
	}

	public abstract void draw(Graphics g);
	protected abstract void setVertex();

	public String toString()  {
		String myself = "";
		myself = myself + _color.getRed() + "," + _color.getGreen() + "," + _color.getBlue() + "\n";
		myself = myself + _initConfig.toString() + "\n";
		myself = myself + _dimX + "," + _dimY + "," + _dimZ + "\n";
		return myself;
	}

	public void writeToFile(/*DataOutputStream out*/ObjectOutputStream out) throws IOException {
		out.writeInt(_color.getRed()); out.writeInt(_color.getGreen()); out.writeInt(_color.getBlue());
		out.writeFloat(_initConfig.xo);	out.writeFloat(_initConfig.xx);	out.writeFloat(_initConfig.xy);	out.writeFloat(_initConfig.xz);
		out.writeFloat(_initConfig.yo);	out.writeFloat(_initConfig.yx);	out.writeFloat(_initConfig.yy);	out.writeFloat(_initConfig.yz);
		out.writeFloat(_initConfig.zo);	out.writeFloat(_initConfig.zx);	out.writeFloat(_initConfig.zy);	out.writeFloat(_initConfig.zz);
		out.writeFloat(_dimX);out.writeFloat(_dimY); out.writeFloat(_dimZ);
 	}

	public void writeToTxtFile(PrintWriter out)   {
		out.print(toString());
	}

	// Box is already read from the main process
	public void readFromFile(DataInputStream in) throws IOException {
		_color = new Color(in.readInt(),in.readInt(),in.readInt());

		float[] matelm = new float[12];
		for(int i=0; i < 12; i++)  {
			matelm[i] = in.readFloat();
		}

		if(_mat != null)  {
			_mat.setElement(matelm);
		}
		else  {
			_mat = new Matrix3D(matelm);
		}

		if(_initConfig != null)  {
			_initConfig.setElement(matelm);
		}
		else  {
			_initConfig = new Matrix3D(matelm);
		}

		_dimX = in.readFloat();
		_dimY = in.readFloat();
		_dimZ = in.readFloat();
		setVertex();
	}


	public void copyTo(CADObject obj)  {
		// copy my configuration to obj configuration. Additional info should be copied
		// by overriding this method
		_initConfig.copyTo(obj._initConfig);
		obj._color = _color;
		obj._type = _type;
		obj._dimX = _dimX;
		obj._dimY = _dimY;
		obj._dimZ = _dimZ;

	}
	public float getDX()  {
		return _dimX;
	}

	public float getDY()  {
		return _dimY;
	}

	public float getDZ()  {
		return _dimZ;
	}

	public void setDX(float x)  {
		_dimX = x;
	}

	public void setDY(float y)  {
		_dimY = y;
	}

	public void setDZ(float z)  {
		_dimZ = z;
	}

	public void reset()   {
		_initConfig.copyTo(_mat);
		setVertex();
	}
}
