package ProtocolLayer;

import Abstract.*;
import BaseLayer.*;

import java.net.*;
import java.io.*;

/**
 * Take a role to make a server socket for applet data transfer. File transfer will
 * be performed asynchronously. 
 */
 /*
* Copyright (c) 1996,1997,1998 <A HREF="http://cdr.stanford.edu/">Center for Design Research</A>, Stanford University.
* All rights reserved.<p>
*
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You can obtain the GNU General Public License at
* http://www.gnu.org/copyleft/gpl.html or by writing to the Free
* Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
* 02111-1307, USA.
* 
* This software is bound by the terms and conditions listed in the
* attached <a href="../LICENSE">LICENSE file</A>.
*  @author <A HREF="http://cdr.stanford.edu/html/people/jeon-bio.html>Heecheol Jeon</A>
*  @version Java(tm) 1.1
*/ 
 
public class FTPPortConnectorThread extends ConnectorThread   {
	
	/**
	 * True if put
	 */
	boolean _getOrPut;
	
	/** 
	  * 4K buffer
	  */
	protected int MAX_MSG_SIZE = 0x1000; 

	/**
	 * Input stream to read data
	 */
	BufferedReader _in = null;
	
	/**
	  * Output stream to write data
	  */
	BufferedWriter _out = null;
	
	/**
	 * FTP server side server socket
	 */
	ServerSocket _serverSideSS;
	
	/**
	  * Parent connector thread. After data transfer, this thread will resume the
	  * parent thread if it is suspended.
	  */
	FTPConnectorThread _parent;
	
	/**
	 * Constructor
	 * @param parent Parent FTPConnectorThread
	 * @param serverSideSS Server side server socket,where FTP server will connect.
	 * @param clientSideSocket Client side socket.
	 * @param priority Thread priority
	 * @param connectionWaitTime Maximum idle time to wait input from data provider
	 * @exception Abstract.ConnectionException
	 */
 	public FTPPortConnectorThread(FTPConnectorThread parent,
 																ServerSocket serverSideSS,
 																Socket clientSideSocket,
												 				int priority,
												 				int connectionWaitTime)
    throws ConnectionException {
    
    if(clientSideSocket != null)   {	
    	_client = new BRecvThread(clientSideSocket,priority,null,null,null,null);
    }
    	
		_maxWaitTime = connectionWaitTime;
		_getOrPut = true;
		_timer = null;
		_clientServerSocket = null;
		_server = null;
		_serverSideSS = serverSideSS;
		_parent = parent;
	}	

	/**
	  * Set server side socket. Server side BRecvThread will be created
	  * @param s Socket which accepts server side connection
	  * @param priority Priority of BRecvThread
	  * @exception Abstract.ConnectionException
	  */
	public void setServerSideSocket(Socket s,int priority) throws ConnectionException {
		if(_server == null)  {
    	_server = new BRecvThread(s,priority,null,null,null,null);
		}
	}
		
	/**
	  * Set client side socket. Client side BRecvThread will be created
	  * @param s Socket which accepts server side connection
	  * @param priority Priority of BRecvThread
	  * @exception Abstract.ConnectionException
	  */
	public void setClientSideSocket(Socket s,int priority) throws ConnectionException {
		if(_server == null)  {
    	_client = new BRecvThread(s,priority,null,null,null,null);
		}
	}
				
	/**
	  * Set get or put mode. True for put mode
	  * @param putIsTrue True if put
	  */				
	public void setGetOrPut(boolean putIsTrue)  {
		_getOrPut = putIsTrue;
	}
		
	/**
	  * Create server BRecvThread after accepting socket connection from FTP server.
	  * Then, data transfer will be performed. Finally it will resume parent FTPConnectorThraed
	  * and stop thread.
	  */
	public void run()  {
		resetTimer();
		try  {
			if(_server == null)   {
				connectToServerPort();
			}
						
			transfer();
		} catch (Exception e)  {
			System.out.println(e.toString());
		}	finally  {
			// parent will clean up this thread if necessary.
			// parent will be suspended only if "226" transfer complete message is received
			// from FTP server side
			if(_parent != null)  {
System.out.println("Parent wake");				
				_parent.resume();
			}	
			
			endConn();
		}				
	}
			
	/**
	  * Accept server side socket connection and create a BRecvThread for server side.
	  * @exception Abstract.ConnectionException 
	  */
 protected void connectToServerPort() throws ConnectionException  {
				
		Socket socket = null;
		try  {
System.out.println("Wait for FTP server connection....");		
 			// wait for connection
			socket = _serverSideSS.accept();
System.out.println("Port server socket accepted");			
			_serverSideSS.close();
			_serverSideSS = null;
    	_server = new BRecvThread(socket,getPriority(),null,null,null,null);

		} catch (Exception ce)  {
System.out.println("Port server socket accept exception");	
			throw new ConnectionException(ce.toString());
		}		
 }		
			
	/**
	  * Data transfer. Continuous read-->write operation until EOF meets from the
	  * input side.
	  * @exception Abstract.ConnectionException May include IOException.
	  */			
  public void transfer()  throws ConnectionException  {
  	try  {
			resetTimer();  
    	char[] buffer = new char[MAX_MSG_SIZE];

			if(_getOrPut == true)  {
				_in = _client.getInputStream();
				_out = _server.getOutputStream();
			}
			else  {
				_in = _server.getInputStream();
				_out = _client.getOutputStream();
			}

			int readlen;	
	
			while((readlen = _in.read(buffer,0,MAX_MSG_SIZE)) > 0)  {
				resetTimer();
				_out.write(buffer,0,readlen);
				_out.flush();		
			}		
System.out.println("End transmission");			
		}  catch (Exception e)  {
				System.out.println("Error while data transfer::" + e.toString());
				throw new ConnectionException(e.toString());
		}
 }	

 /**
   * End this thread. Cleanup and stop thread
   */	
 public void endConn()   {
System.out.println("FTPPort Connector thread stop" + getName()); 	
 	try  {
 		if(_serverSideSS != null)  {
 			_serverSideSS.close();
 			_serverSideSS = null;
 		}	
 		
 	  if(_in != null)  {
 	  	_in.close();
 	  	_in = null;
 	  }
 	  
 	  if(_out != null)  {
 	  	_out.close();
 	  	_out = null;
 	  }
 	 }  catch (Exception e)  {
 	 	 System.out.println("Close socket stream error in FTPPortConnector");
 	 }	   
 	 super.endConn();
 }
}