package ProtocolLayer;

import Abstract.*;
import BaseLayer.*;
import KQMLLayer.*;

import java.util.*;
import java.text.*;
import java.net.*;
import java.io.*;

/**
* Example class of extending IPThread(Internet protocol thread). FTPThread will be created
* for data transfer using File Transfer Protocol(FTP).<BR>
* Current implementation performs a minamal set of command procedure - connect to 
* FTP server, change remote site directory, set transfer mode(ASCII or binary) and 
* GET/PUT/APPEND. List command is not implemented.
*/

/*
* Copyright (c) 1996,1997,1998 <A HREF="http://cdr.stanford.edu/">Center for Design Research</A>, Stanford University.
* All rights reserved.<p>
*
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You can obtain the GNU General Public License at
* http://www.gnu.org/copyleft/gpl.html or by writing to the Free
* Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
* 02111-1307, USA.
* 
* This software is bound by the terms and conditions listed in the
* attached <a href="../LICENSE">LICENSE file</A>.
*  @author <A HREF="http://cdr.stanford.edu/html/people/jeon-bio.html>Heecheol Jeon</A>
*  @version Java(tm) 1.1
*/ 

public class FTPThread extends IPThread   {
	
	/**
	  * Socket for applet port. Will be used only for applet data transfer
	  */
	Socket _appletPortSocket = null;
	
	/**
	  * FTP port thread to perform get/put/append operation
	  */
	FTPPortThread _portThread = null;	
	
	/**
	  * Server socket for port. Will be used only for stand alone agent
	  */ 
	ServerSocket _portServerSocket = null;
	
	/**
	  * FTPCon for current data transfer operation
	  */
	FTPCon _ftpCon = null;	
	
	/**
	* Default Constructor.
	*/
	public FTPThread()  {
		super();
		_endWith = '\n';
	}
		
	/**
	* Create ReceiverThread with specific thread name
	* @param name Agent name
	*/	
	public FTPThread(String name)  {
		super(name);
		_endWith = '\n';
	}
		
	/**
	* Create FTPThread with specific thread name and Thread group.
	* Passing thread group is essential when you will create the receiver thread for an applet
	* because of security issue.
	* @param tg Thread group to which this thread will be owned to
	* @param name Agent name
	*/	
	public FTPThread(ThreadGroup tg,String name)  {
		super(tg,name);
		_endWith = '\n';
	}
		
	/**
	* Can be used both for applets and stand-alone agents, but for efficiency, normally
	* used for applets. Router will create a connector which connects an applet and SMTP
	* server. Through the connector, email will be sent as if the applet is a stand-aloone
	* agent.
	* @param con FTPCon object 
	* @see ProtocolLayer.FTPCon
	* @param priority Thread priority. Default is NORM_PRIORITY
	* @param tg Thread group, the thread will be grouped.
	* @exception Throws ConnectionException if connection error occurs.
	*/	
	public FTPThread(FTPCon con,
									 int priority,
									 ThreadGroup tg) throws ConnectionException {
		this(con.getID(),
				 con.getRemotePath(),
				 con.getHost(),
				 con.getPort(),
				 con.getUserName(),
				 con.getPassword(),
				 con.getLocalPath(),
				 con.getFiles(),
				 con.getMode(),
				 priority,
				 tg);
	}			 	

	/**
	* For stand-alone agent file transfer.
	* @parem conID FTPCon id
	* @param remotePath Remote site initial path
  * @param remotePath FTP site initial directory
  * @param host FTP server host name
  * @param port 21. Otherwiser, you need to set
  * @param userName FTP server login user name
  * @param password FTP server login password
  * @param localPath Local machine initial directory
  * @param files Files to put or get
  * @param mode Mode definition
  * @param priority Thread priority
  * @param tg Thread group
	* @exception Throws ConnectionException if connection error occurs.
	*/
	public FTPThread(String conID,
									 String remotePath,
									 String host,
									 int port,
									 String userName,
									 String password,
									 String localPath,
									 String[] files,
									 int mode,
									 int priority, 
									 ThreadGroup tg) 
		throws ConnectionException {
		super(tg,"FTP");
		// set server address for connection
		_serverAddress = new Address("FTP",host,port,"FTP","");
		
		// security check will be performed by userName , password and remotePath accessibility
		boolean createdir = false;
		if((mode & FTPCon.CREATEDIR) != 0)  {
			createdir = true;
		}	
		_security = new FTPSecurity(remotePath,userName,password,createdir);

		String nlocalPath = null;	
		// get mode? then set local path. For put mode, this parameter is meaningless.
		if((mode & FTPCon.ISGETMODE) != 0)  {
			nlocalPath = localPath;
		}

		int nmode = mode;

		_ftpCon = new FTPCon(conID,remotePath,host,port,userName,password,nlocalPath,files,nmode);
		
		setPriority(priority);				
			
		// local path is valid only for GET operation.
		// files should have full file paths.
		boolean get = false;
		if((nmode & FTPCon.ISGETMODE) != 0)  {
			get = true;
		}	
		
		makeSimpleCommandList(files,get);
		setPriority(priority);	
	}			 	

  /** 
   * For applet data transfer using FTP. agentAction(class IPRouterClientAction) should
   * be passed for receiving callback. By creating a FTPThread from an applet,
   * FTPThread will prepare FTPPortThread. FTPPortThread will create
   * InputStream/OutputStream for the agentAction. 
   * Then FTPPortThread will execute FTPStore/FTPAppend/FTPRetrieve. From the applet side,
   * it should implement above three abstract methods for actual behavior from given
   * InputStream/OutputStream which connects FTPServer through router. 
   * @param agentAction IPRouterClientAction, which implements FTPStore/FTPAppend/FTPRetrieve.
   * @param serveraddress Router FTP proxy address
   * @parem con FTPCon
   * @param priority The priority of the thread.
   * @param tg Thread group
   * @exception ConnectionException if error occurs
   */
  public FTPThread(IPRouterClientAction agentAction,
  								 Address serveraddress,
  								 FTPCon con,
  								 int priority, 
							     ThreadGroup tg)
    throws ConnectionException   {
      this(con,priority,tg);  
      _ipAgentAction = agentAction;
      _ftpCon.setMode(_ftpCon.getMode() | FTPCon.ISAPPLET);
			_proxyAddress = serveraddress;	
			_useProxy = true;
  }
  
  /** 
   * For applet data transfer using FTP. agentAction(class IPRouterClientAction) should
   * be passed for receiving callback. By creating a FTPThread from an applet,
   * FTPThread will prepare FTPPortThread. FTPPortThread will create
   * InputStream/OutputStream for the agentAction. 
   * Then FTPPortThread will execute FTPStore/FTPAppend/FTPRetrieve. From the applet side,
   * it should implement above three abstract methods for actual behavior from given
   * InputStream/OutputStream which connects FTPServer through router. 
   * @param agentAction IPRouterClientAction, which implements FTPStore/FTPAppend/FTPRetrieve.
   * @param serveraddress Router FTP proxy address
	* @parem conID FTPCon id
	* @param remotePath Remote site initial path
  * @param remotePath FTP site initial directory
  * @param host FTP server host name
  * @param port 21. Otherwiser, you need to set
  * @param userName FTP server login user name
  * @param password FTP server login password
  * @param localPath Local machine initial directory
  * @param files Files to put or get
  * @param mode Mode definition
   * @param priority The priority of the thread.
   * @param tg Thread group
   * @exception ConnectionException if error occurs
   */
  
	public FTPThread(IPRouterClientAction agentAction,
									 Address serveraddress,
									 String conid,
									 String remotePath,
									 String host,
									 int port,
									 String userName,
									 String password,
									 String localPath,
									 String[] files,
									 int mode,
									 int priority, 
									 ThreadGroup tg) 
		throws ConnectionException {
  	this(conid,remotePath,host,port,userName,password,localPath,files,mode,priority,tg);
  	if(serveraddress == null)  {
  		throw new ConnectionException("Server address not specified for FPT Thread");
  	}
    _ipAgentAction = agentAction;
    _ftpCon.setMode(_ftpCon.getMode() | FTPCon.ISAPPLET);
		_proxyAddress = serveraddress;	
		_useProxy = true;
	}  	
	
	

	/**
	* Make the simplest FTP command list from given data transfer information.
	* Typical command list is :<P>
	* type a<BR>
	* port<BR>
	* appe file1<BR>
	* appe file2<BR>
	* stor file3<P>
	* for appending file1, file2 and storing file3 with ASCII mode.	
	* @param files String list of files to put/get/append
	* @param getOrPut Get mode is true.
	*/
	public void makeSimpleCommandList(String[] files,
																		boolean getOrPut)   {
		_cmdList = new Vector();
		
		int trMode = _ftpCon.getMode() & FTPCon.ASCIITYPE;  // ascii type -> true
		
		if(trMode != 0)  {
			_cmdList.addElement("type a");
		}
		else  {
			_cmdList.addElement("type i");
		}
		
		String cmd;
		// get mode true
		if(getOrPut)  {
			
			for(int i=0; i < files.length; i++)  {
				cmd = "port";  // do necessary thing according to Applet or Stand-alone
				               // note that port informations are needed for stand-alone
				               // but not necessary for applets
				_cmdList.addElement(cmd);               
				_cmdList.addElement("retr " + files[i]);
			}		
		}
		// put mode
		else  {
			
			// for put case, all files has full file paths
			for(int i=0; i < files.length; i++)  {
				_cmdList.addElement("port");
				if((_ftpCon.getMode() & FTPCon.APPEND) != 0)  {
					_cmdList.addElement("appe " + files[i]);		
				}
				else  {
					_cmdList.addElement("stor " + files[i]);
				}
			}			
		
		}
	}
	
	/**
	* Set file transfer mode if different from default mode.
	*	Default mode is binary transfer, no directory creation, no append.
	* Modes are defined :<BR>
	* 	public static final int NORMALMODE = 0;  // default mode<BR>
	* 	public static final int APPEND = 1;  // file append mode<BR>
	*   public static final int CREATEDIR = 2;  // create directory<BR>
	*   public static final int ASCIITYPE = 3; // true for ascii. <BR>
	* @param mode FTP mode.
	*
	*/
	public void setMode(int mode)  {
		_ftpCon.setMode(mode);
		
		if(_security != null)  {
			boolean createdir = false;
			if((mode & FTPCon.CREATEDIR) != 0) {
				createdir = true;
			}	
			((FTPSecurity)_security).setMode(createdir);
		}	
	}
						
	/**
	* processCommands implementation. Send predefined command to mail server and manage received
	* response.
	*/
	public void processCommands()   {
		boolean isApplet = false;
		boolean isError = false;
		
		if((_ftpCon.getMode() & FTPCon.ISAPPLET) != 0)  {
			isApplet = true;
		}	
				
		try  {
			if(_cmdList != null)   {
				Enumeration e = _cmdList.elements();
				while(e.hasMoreElements())  {
					String cmd = (String)e.nextElement();
					
System.out.println(cmd);					
					if(cmd.startsWith("port"))  {
						if(isApplet) {
							appletPort();
						}
						else  {
							port();
						}	
					}
					else		
					if(cmd.startsWith("stor"))  {
						String filefullpath = cmd.substring(5);
						int delloc = filefullpath.lastIndexOf(File.separatorChar);
						if(delloc != -1)  {
							String filename = filefullpath.substring(delloc + 1);
							writeMsg("stor " + filename);
						}
						else  {	 
							writeMsg(cmd);
						}	
						
						if(isApplet) {
							appletDataTransfer(cmd);
						}
						else {
							store(cmd);
						}
					}
					else
					if(cmd.startsWith("retr"))  {
						writeMsg(cmd);
					
						if(isApplet)  {
							appletDataTransfer(cmd);
						}
						else {
							retrieve(cmd);
						}
					}
					else
					if(cmd.startsWith("appe"))  {
						String filefullpath = cmd.substring(5);
						int delloc = filefullpath.lastIndexOf(File.separatorChar);
						if(delloc != -1)  {
							String filename = filefullpath.substring(delloc + 1);
							writeMsg("appe " + filename);
						}
						else  {	 
							writeMsg(cmd);
						}	
					
						if(isApplet)  {
							appletDataTransfer(cmd);
						}
						else  {
							append(cmd);
						}
					}		
					else  {
						writeMsg(cmd);
					}	
					String returnmsg = getOneMessage();
					
					// if AgentAction want to get control, give it a control
					if(_ipAgentAction != null)  {
						_ipAgentAction.processMessage("FTPMessage",returnmsg);
					}
						
					// port ready. need to receive '226'
					if(returnmsg.startsWith("150"))  {
						returnmsg = getOneMessage();
												
						if(returnmsg.startsWith("226"))  {
							if((_portThread != null) && (_portThread.isAlive()))  {
								suspend();
								_portThread.endConn();
								_portThread = null;
							}
						}
						else  {
							break;
						}	
					}	
								
					String errormsg = processErrorMsg(returnmsg);
					if(errormsg != null)   {
						System.out.println(errormsg);
						// will be disconnected by run() method
						
						notifyError(errormsg);
						isError = true;					
						writeMsg("quit");
						break;
					}	
				}	// while loop
				
				if(!isError)  {
					// if ipAgentAction is not null, notify message
				  notifyMessage();
				}
			}	// if cmdlist is not null
		}	 catch (Exception e)  {
			System.out.println(e.toString());

			// if ipAgentAction is not null, notify error
			notifyError(e.toString());
		}  finally  {
System.out.println("End FTPProcess");			
			
			endConn();
		}	
	}	
	
	/**
	  * Execute 'port' command
	  * @exception Abstract.ConnectionException If server socket can not be opened,Connection
	  * exception will be thrown
	  */
	public void port() throws ConnectionException  {
		try  {
			if(_portServerSocket != null) {
				_portServerSocket.close();
			}
				
 			// make a new server socket
 			_portServerSocket = new ServerSocket(0);

 			// send server socket information to FTP server
			int localPort = _portServerSocket.getLocalPort();
			InetAddress localIp = InetAddress.getLocalHost();
			byte[] addrbytes = localIp.getAddress();
			short addrshorts[] = new short[4];

			for(int i=0; i<=3; i++){
				addrshorts[i] = addrbytes[i];
				if(addrshorts[i] < 0) addrshorts[i] += 256;
			}

			String portstr = "port ";
			portstr = portstr + addrshorts[0] + "," + addrshorts[1] + "," + 
												  addrshorts[2] + "," + addrshorts[3] + "," + 
						  						((localPort & 0xff00) >> 8) + "," + (localPort & 0x00ff);
			writeMsg(portstr);
		} catch (Exception e)  {
			throw new ConnectionException("FTPPort error ::" + e.toString());
		}		
  						
	}

	/**
	  * Retrieve file from a remote FTP server. Create FTPPortThread and start it.
	  * @param cmd 'retr fileName'
	  * @exception Abstract.ConnectionException If FTPPort can not be created or 
	  * receiving file can not be created.
	  */
	public void retrieve(String cmd)  throws ConnectionException   {
	
			String filename = "NotSpecified";
		try  {
			if((_portThread != null) && (_portThread.isAlive()))  {
				_portThread.endConn();
			}
				
			filename = cmd.substring(5);
			String fullFilePath = _ftpCon.getLocalPath() + File.separatorChar + filename;

			BufferedWriter fout = new BufferedWriter(new FileWriter(fullFilePath,false),1024); 		
			_portThread = new FTPPortThread(this,_portServerSocket,fout,_ftpCon.getMode());
			_portThread.start();
		}  catch (Exception e)  {
			throw new ConnectionException("Error while retrieving " + filename + "::" + e.toString());
		}		
	}

	/**
	  * Save file to a  remote FTP server. Create FTPPortThread and start it.
	  * @param cmd 'stor fileName'
	  * @exception Abstract.ConnectionException If FTPPort can not be created or 
	  * receiving file can not be created.
	  */
	public void store(String cmd) throws ConnectionException {
		String filepath = "NotSpecified";
		try  {
			if((_portThread != null) && (_portThread.isAlive()))  {
				_portThread.endConn();
			}

			filepath = cmd.substring(5);

			// open the file
			BufferedReader in = new BufferedReader(new FileReader(filepath));
			_portThread = new FTPPortThread(this,_portServerSocket,in,_ftpCon.getMode());
			_portThread.start();
		}  catch (Exception e)  {
			throw new ConnectionException("Error while retrieving " + filepath + "::" + e.toString());
		}		
	}

	/**
	  * Append file to a  remote FTP server. Create FTPPortThread and start it.
	  * @param cmd 'appe fileName'
	  * @exception Abstract.ConnectionException If FTPPort can not be created or 
	  * receiving file can not be created.
	  */
	public void append(String cmd) throws ConnectionException {
			store(cmd);
	}
	
	/**
	  * Execute 'port' command for applets. The router will response 'port' message
	  * and provide a proxy port with '299 port-number'.
	  * @exception Abstract.ConnectionException If can not connect to the proxy socket	  */
	public void appletPort()	throws ConnectionException {
		writeMsg("port");
		String retstr = getOneMessage();
System.out.println("port info from router " + retstr);		
		// JATLite specific protocol
		if(!retstr.startsWith("299"))  {
			throw new ConnectionException(retstr);
		}
		else  {
			String portNo = retstr.substring(4);
			
			try  {
				if((_portThread != null) && (_portThread.isAlive()))  {
					_portThread.endConn();
				}
				
				if(_appletPortSocket != null)  {
					_appletPortSocket.close();
					_appletPortSocket = null;
				}	
				
				_appletPortSocket = new Socket(_proxyAddress.getHost(),(Integer.valueOf(portNo)).intValue());
				
			} catch (Exception e)  {
				throw new ConnectionException("Applet Port Exception::" + e.toString());
			}		
		}		
	}
	
	/**
	  * Close port socket if not already closed
	  */
	protected void closePortSocket()  {

		try  {
			if(_appletPortSocket != null)  {
				_appletPortSocket.close();
				_appletPortSocket = null;
			}						
			
			if(_portServerSocket != null)  {
				_portServerSocket.close();
				_portServerSocket = null;
			}	
		} catch (Exception ee)  {
			System.out.println(ee.toString());
		}		
		
		try  {
			if((_portThread != null) && (_portThread.isAlive()))  {
				_portThread.endConn();
				_portThread = null;
			}
		}  catch (Exception eee) {
			System.out.println(eee.toString());
		}			
	}
		
	/**
	* File names will be given from command and the file is already opened and waiting for
	* data. You should implement FTPStore(String remoteFilePath,PrintWriter) callback in 
	* IPRouterClientAction. 
	* @param command Simple command list created by default.
	* @exception Abstract.ConnectionException 
	*/
	public void appletDataTransfer(String command) throws ConnectionException {
		if(_ipAgentAction == null)  {
			throw new ConnectionException("IPRouterClientAction is not defined");
		}
		
		if((_portThread != null) && (_portThread.isAlive()))  {
			_portThread.endConn();
		}

		if(_appletPortSocket == null)  {
			throw new ConnectionException("FTP Port socket for APPLET fails");
		}
	
		try  {				
			_portThread = new FTPPortThread(this,_appletPortSocket,_ftpCon.getMode(),command.substring(5));
			_portThread.start();
			
		} catch (ConnectionException e)  {
			throw e;
		}		
	}
	
 /**
	* According to the response from mail server, return false if the error message
	* can not be managed. Return true if the response from server is success or 
	* the error can be fully managed.<P>
	*
  *  202 Command not implemented, superfluous at this site.<BR>
  *  221 Service closing control connection. Logged out if appropriate.<BR>
  *  226 Closing data connection.<BR>
  *  350 Requested file action pending further information.<BR>
  *        
  *  421 Service not available, closing control connection.<BR>
  *  425 Can't open data connection.<BR>
  *  426 Connection closed; transfer aborted.<BR>
  *  450 Requested file action not taken.File unavailable (e.g., file busy).<BR>
  *  451 Requested action aborted: local error in processing.<BR>
  *  452 Requested action not taken. Insufficient storage space in system<BR>
  *	 500 Syntax error, command unrecognized.<BR>
  *  501 Syntax error in parameters or arguments.<BR>
  *  502 Command not implemented.<BR>
  *  503 Bad sequence of commands.<BR>
  *  504 Command not implemented for that parameter.<BR>
  *  530 Not logged in.<BR>
  *  532 Need account for storing files.<BR>
  *  550 Requested action not taken. File unavailable (e.g., file not found, no access).<BR>
  *  551 Requested action aborted: page type unknown.<BR>
  *  552 Requested file action aborted.Exceeded storage allocation (for current directory or dataset).<BR>
  *  553 Requested action not taken. File name not allowed.<BR>
  *
	* @param msg Returned message from email server
	*/
	protected String processErrorMsg(String msg)   {
		String retstr = null;
		String retcode = msg.substring(0,3);
		int code = (Integer.valueOf(retcode)).intValue();
		switch(code)   {
			case 202: retstr = "Command not implemented, superfluous at this site."; break;
			case 221: retstr = "Service closing control connection. Logged out if appropriate."; break;
			case 350: retstr = "Requested file action pending further information."; break;
			case 421: retstr = "Service not available, closing control connection."; break;
			case 425: retstr = "Can't open data connection."; break;
			case 426: retstr = "Connection closed; transfer aborted."; break;
			case 450: retstr = "Requested file action not taken.File unavailable (e.g., file busy)."; break;
			case 451: retstr = "Requested action aborted: error in processing"; break;
			case 452: retstr = "Requested action not taken: insufficient system storage"; break;
			case 500: retstr = "Syntax error, command unrecognized."; break;
			case 501: retstr = "Syntax error in parameters or arguments."; break;
			case 503: retstr = "Bad sequence of commands."; break;
			case 504: retstr = "Command not implemented for that parameter."; break;
			case 530: retstr = "Not logged in."; break;
			case 532: retstr = "Need account for storing files."; break;
			case 550: retstr = "Requested action not taken. File unavailable (e.g., file not found, no access)."; break;
			case 551: retstr = "Requested action aborted: page type unknown."; break;
			case 552: retstr = "Requested FTP action aborted: exceeded storage allocation"; break;
			case 553: retstr = "Requested action not taken: File name not allowed."; break;
		}
		
		return retstr;	
	}
	
	/**
	  * Close port socket and stop. 
	  */
	public void endConn()  {
		try  {			
			// to be safe and make less burden to FTP server.
			writeMsg("quit");
		} catch (Exception e)  {}	
			
		closePortSocket();
		super.endConn();
 }
	
	/**
	  * Notify error if _ipAgentAction(class IPRouterClientAction) is not null
	  * @param msg Message to be notified to _ipAgentAction
	  */
	public void notifyError(String msg)  {
		if(_ipAgentAction != null)   {
			if(_ftpCon != null)  {
				_ftpCon.setErrorMessage(msg);
			}	
			_ipAgentAction.processMessage("FTP",_ftpCon);
		}
	} 
	
	/**
	  * Notify message(FTPCon itself) if _ipAgentAction(class IPRouterClientAction) is not null
	  * _ipAgentAction will invoke processMessage(String,Object) where String is "FTP" and
	  * Object is _ftpCon.
	  */
	public void notifyMessage()  {
		if(_ipAgentAction != null)   {
			_ipAgentAction.processMessage("FTP",_ftpCon);
		}
	}
	
	
}