package ProtocolLayer;

import Abstract.*;
import BaseLayer.*;
import KQMLLayer.*;
import RouterLayer.Router.*;

import java.util.*;
import java.net.*;

/**
  * Extended from RouterRecvThread, this class handles SMTP and FTP proxy. This class
  * is only for router. IPRouterClientAction will not use this class.
  */
/*
* Copyright (c) 1996,1997,1998 <A HREF="http://cdr.stanford.edu/">Center for Design Research</A>, Stanford University.
* All rights reserved.<p>
*
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You can obtain the GNU General Public License at
* http://www.gnu.org/copyleft/gpl.html or by writing to the Free
* Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
* 02111-1307, USA.
* 
* This software is bound by the terms and conditions listed in the
* attached <a href="../LICENSE">LICENSE file</A>.
*  @author <A HREF="http://cdr.stanford.edu/html/people/jeon-bio.html>Heecheol Jeon</A>
*  @version Java(tm) 1.1
*/ 
  
public class IPRouterRecvThread extends RouterRecvThread  {

  /**
   * Error codes : Server address(SMTP or FTP server) is not found
   */
  protected final static int SERVERADDRESSMISSING = 5000;
  
  /**
   * Error codes : Proxy server socket can not be created
   */
  protected final static int INTERNALSOCKETERROR = 5001;
  
  /**
   * Error codes : The protocol is not supported
   */
  protected final static int NOTSUPPORTEDPROTOCOL = 5002;
  
  /**
   * Error codes : The content of KQML message is wrong
   */
  protected final static int MESSAGECONTENTERROR = 5003;
  
  /**
   * Error codes : The internet service is closed
   */
  protected final static int SERVICECLOSED = 5004;
  
  /**
   * Error codes : The protocol is not specified
   */
  protected final static int PROTOCOLNOTSPECIFIED = 5006;
  
  /**
   * Default Constructor.
   */
  public IPRouterRecvThread()  {
    super();
  }
  
  /**
   * Constructor with the Agent name. If you use this constructor, you should
   * use set methods to set data members
   * @param name Agent name
   */	
  public IPRouterRecvThread(String name)  {
    super(name);
  }
  
  /**
   *  Constructor, to be called from the serverthread after accepting
   *  a connection. 
   * @param client_scoket The socket for the remote client;
   * @param priority The priority of the thread.
   * @param myaddress My address to be sent to the Router to check security
   * @param connections The connection table is passed as an argument.
   * @param queue The Message Queue. Normally null. It will be created by this if null
   * @param security The RouterSecurity
   * @param messageboxpath The Message incoming box directory
   * @param reservedBook The reservation table, passed from the RouterAction
   * @param offlinelist The off line agent name list.
   * @param disconnectlist The disconnected agent name list.
   */
  public IPRouterRecvThread(Socket client_socket,
			    int priority, 
			    Address myaddress,
			    ConnectionTable connections,	
			    MessageBuffer queue,
			    Security security,
			    String messageboxpath,
			    Hashtable reservedBook,
			    OffLineAgentTable offlinelist,
			    Vector disconnectlist) 
    throws ConnectionException    {
      super(client_socket,
	    priority,
	    myaddress,
	    connections,
	    queue,
	    security,
	    messageboxpath,
	    reservedBook,
	    offlinelist,
	    disconnectlist);
  }
  
  
  
  /**
   *  This constructor is called when an agent want to initialize a connection
   *  to the other agents, using Address
   * @param serveraddress The Router address
   * @param priority The priority of the thread.
   * @param myaddress My address to be sent to the Router to check security
   * @param connections The connection table is passed as an argument.
   * @param queue The Message Queue. Normally null. It will be created by this if null
   * @param security The RouterSecurity
   * @param messageboxpath The Message incoming box directory
   * @param reservedBook The reservation table, passed from the RouterAction
   * @param offlinelist The off line agent name list.
   * @param disconnectlist The disconnected agent name list.
   * @param tg The thread group to which this receiver thread will be owned to
   */
  public IPRouterRecvThread(Address serveraddress,
			    int priority, 
			    Address myaddress,
			    ConnectionTable connections,
			    MessageBuffer queue,
			    Security security,
			    String messageboxpath,
			    Hashtable reservedBook,
			    OffLineAgentTable offlinelist,
			    Vector disconnectlist,
			    ThreadGroup tg)  
    throws ConnectionException   {
      super(serveraddress,
	    priority,
	    myaddress,
	    connections,
	    queue,
	    security,
	    messageboxpath,
	    reservedBook,
	    offlinelist,
	    disconnectlist,
	    tg);
  }
  
  
  /**
   * Router receiver thread to send off-line agent's messages back to the senders with error.
   *	@param serveraddress Router address
   * @param myaddress Router address to be sent to the Router to check security
   * @param connections The connection table is passed as an argument.
   * @param security The RouterSecurity
   * @param messageboxpath The Message incoming box directory
   * @param offlinelist The off line agent name list.
   */
  public IPRouterRecvThread(Address serveraddress,
			    Address myaddress,
			    ConnectionTable connections,
			    String messageboxpath,
			    Security security,
			    OffLineAgentTable offlinelist)  {
    super(serveraddress,
	  myaddress,
	  connections,
	  messageboxpath,
	  security,
	  offlinelist);
    
  }
  
  
  /**
   * Processing default protocols : disconnect-agent, reserve-message, list-agent, request-address,
   * and unregister-agent. You need to extend this method if your router has extended capability
   * other than message routing(e.g. to support another internet protocol such as SMTP, FTP,etc.)
   * @param kqml Received KQML message
   */
  protected synchronized void processDefaultProtocol(KQMLmessage kqml)   {
    super.processDefaultProtocol(kqml);
    
    String perf = kqml.getValue("performative");
    if(perf.equals("data-transfer"))  {
      String content = kqml.getValue("content");
      String protocol = kqml.getValue("protocol");
      String sender = kqml.getValue("receiver");
      String receiver = kqml.getValue("sender");
      String replywith = kqml.getValue("reply-with");
      String language = kqml.getValue("language");
      if(language == null)  {
				language = "KQML";
      }
      
      KQMLmessage ckqml;
      try  {			
				ckqml = new KQMLmessage(content);
      }  catch (ParseException e)   {
				sendProtocolErrorMessage(MESSAGECONTENTERROR,kqml);
				return;
      }
      
      String server = ckqml.getValue("server");
      if(server == null)  {
				sendProtocolErrorMessage(SERVERADDRESSMISSING,kqml);
				return;
      }
      
      String serverport = kqml.getValue("port");
      
      String retmsg = "(port-info :sender " + sender + " :receiver " + receiver;
      
      retmsg = retmsg + " :language " + language;
      if(replywith != null)  {
				retmsg = retmsg + " :in-reply-to " + replywith;
      }	
      
      if(protocol == null)  {
	sendProtocolErrorMessage(PROTOCOLNOTSPECIFIED,kqml);
	return;
      }
      
      protocol = protocol.toLowerCase();
      
      retmsg = retmsg + " :protocol " + protocol;
      
      if(protocol.equals("smtp"))  {
	if(serverport == null)  {
	  serverport = "25";
	}
	
	Address addr = new Address("SMTP",server,Integer.valueOf(serverport).intValue(),"SMTP","");
	String ereceiver = ckqml.getValue("receiver");
	Address recaddr = (_security.getAddressTable()).getAddress(ereceiver);
	if(ereceiver == null)  {
	  sendErrorMessage(NOTREGISTEREDUSER,ereceiver);
	  return;
	}			
	
	String email = getEmailAddress(recaddr);
	if((email == null) || (email.equals("")))  {
		sendErrorMessage(EMAILNOTSPECIFIED,ereceiver);
		return;
	}
			
	try  {
	  // ten minutes max connection time
	  ConnectorThread t = new ConnectorThread(addr,Thread.NORM_PRIORITY,10);
	  System.out.println("Connector created");				
					// start connector thread and send listening port information to the receiver
	  int clientPort = t.getClientServerSocketNumber();
	  if(clientPort > 0)  {
	    t.start();
	    retmsg = retmsg + " :content (smtp-port :port " + clientPort;
	    if(!email.equals(""))  {
	      retmsg = retmsg + " :receiver-email " + email;
	    }
	    
	    retmsg = retmsg	 + "))";
	    String reply = getMsgHeader();
	    reply = reply + retmsg + _endWith;						
	    _queue.addElement(reply);
	  }	
	} catch (ConnectionException c)  {
	  System.out.println(c.toString());
	  sendProtocolErrorMessage(SERVICECLOSED,kqml);
	}						
      }
      else
	if(protocol.equals("ftp"))  {
	  if(serverport == null)  {
	    serverport = "21";
				}
	  Address addr = new Address("FTP",server,Integer.valueOf(serverport).intValue(),"FTP","");
	  try  {
	    // ten minutes max connection time
	    FTPConnectorThread t = new FTPConnectorThread(addr,Thread.NORM_PRIORITY,10);
	    
	    // start connector thread and send listening port information to the receiver
	    int clientPort = t.getClientServerSocketNumber();
	    if(clientPort > 0)  {
	      t.start();
	      retmsg = retmsg + " :content (ftp-port :port " + clientPort + "))";
	      String reply = getMsgHeader();
	      reply = reply + retmsg + _endWith;						
	      _queue.addElement(reply);
	    }	
	  }  catch (Exception cee)  {
	    System.out.println(cee.toString());
	    sendProtocolErrorMessage(SERVICECLOSED,kqml);
	  }	
	  
	}
    }	
    
  }
  
	/**
	  * Get email address from Address object
	  * @param addr Address
	  * @return null String("") if no email is specified. Email address(e.g. jhc@cdr.stanford.edu)
	  * if addr contains email address information.
	  */		
  public String getEmailAddress(Address addr)   {
    if(addr == null)  {
      return "";
    }
    
    String descript = addr.getDescription();
    if(descript == null)  {
      return "";
    }
    
    try  {
      KQMLmessage kqml = new KQMLmessage(descript);
      return kqml.getValue("email");
    }  catch (Exception e)  {
    }
    
    return "";
  }	
  
  /**
    * Send protocol error message occurs dusing transaction<P>
    * SERVERADDRESSMISSING<BR>
    * INTERNALSOCKETERROR<BR>
    * MESSAGECONTENTERROR<BR>
    * SERVICECLOSED<BR>
    * PROTOCOLNOTSPECIFIED<P>
    * @param code Code described above
    * @param kqml KQML message modified to 'error' protocol
    */
  public void sendProtocolErrorMessage(int code,KQMLmessage kqml)   {
    String sender = kqml.getValue("receiver");
    String receiver = kqml.getValue("sender");
    String language = kqml.getValue("language");
    
    String errormsg = getMsgHeader();
    errormsg = errormsg + "(error ";
    errormsg = errormsg + " :sender " + sender + " :receiver " + receiver;
    errormsg = errormsg + " :language " + language + " :content (";
    
    switch(code)   {
    case SERVERADDRESSMISSING: 
      errormsg = errormsg + "Server address is missing"; break;
    case INTERNALSOCKETERROR:
      errormsg = errormsg + "Internal error"; break;
    case NOTSUPPORTEDPROTOCOL:
      errormsg = errormsg + "Protocol not supported from Router"; break;
    case MESSAGECONTENTERROR:
      errormsg = errormsg + "Message content syntax error"; break;
    case SERVICECLOSED:
      errormsg = errormsg + "Can not connect to server"; break;
    case PROTOCOLNOTSPECIFIED:
      errormsg = errormsg + "Protocol is not specified"; break;	
    }
    
    errormsg = errormsg + "))" + _endWith;
    _queue.addElement(errormsg);
    
  }	
}