package ProtocolLayer;

import Abstract.*;
import BaseLayer.*;

import java.util.*;
import java.net.*;


/**
 *  Abstract class for diverse internet protocol implementations such as
 *  FTP, SMTP,HTTP,POP3, etc. <BR>
 *  Design concept :<BR>
 *  For data transfer through internet, Server information, client information and actual
 *  data will be necessary. Those three different types of information are stored to
 *  _serverAddress, _myAddress and _cmdList data members.<BR>
 *  Server information(service id,host name, port no) will be abstracted in _serverAddress
 *  data member(Address object).<BR>
 *  Client information(user name, password, email address, etc) will be abstract in _myAddress
 *  data member(Address object). <BR>
 *  Data transfer based on one time connection, not a consistent socket connection.
 *  This means that when IPThread is created, it will be disconnected after it finishes
 *  data transfer to or from the server. <BR>
 *  Separate login and logout process from IPThread object. '_security' data member will take a
 *  role to connect and to disconnect from the server.<BR>
 *  Passing a set of command to perform data transfer automatically. Run() method will
 *  perform the command list. Command will be store or retrieve file, email, etc.<BR>
 *  This class is a subclass of BRecvThread. You can use getOneMessage() and writeMsg(String)
 *  methods to receive/send string from/to the server.<BR>
 *  What should do to implement new protocol?<BR>
 *  Subclass IPThread and fill in abstract methods - processCommands(). Override connect() and
 *  disconnect() method if necessary.<BR>
 *  Subclass Security and fill in abstract methods - processClientLogin(),processClientLogout()
 *  processServerLogin() and isValidAgent() methods. Normally you will not need to implement
 *  processServerLogin() and isValidAgent() methods.<BR>
 *  
 *  
 */ 
 
/*
* Copyright (c) 1996,1997,1998 <A HREF="http://cdr.stanford.edu/">Center for Design Research</A>, Stanford University.
* All rights reserved.<p>
*
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You can obtain the GNU General Public License at
* http://www.gnu.org/copyleft/gpl.html or by writing to the Free
* Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
* 02111-1307, USA.
* 
* This software is bound by the terms and conditions listed in the
* attached <a href="../LICENSE">LICENSE file</A>.
*  @author <A HREF="http://cdr.stanford.edu/html/people/jeon-bio.html>Heecheol Jeon</A>
*  @version Java(tm) 1.1
*/ 
 
public abstract class IPThread extends BRecvThread  {
	
	/**
	* Since we are thinking software agent, a set of command needs to be 
	* provided. For FTP, command list would be :
	* cwd pub/KJATLite,stor c:\FileToBePut 
	* for file 'PUT' operation.
	*/
	Vector _cmdList;
	
	/**
	  * Proxy address from router. Host should be the same as router but port number
	  * will be different for every transaction
	  */
	Address _proxyAddress;

  /**
  * Use router as proxy or stand alone application
  * if true, router will be a proxy. Default is false.
  */
  boolean _useProxy;
  
	/**
	 * If you want to get notification from the FTP thread whether it is 
	 * successfully terminated, use this method. If _ipAgentAction data member is not
	 * zero, FTPCon object will be passed with success or error code.
	 */
	IPRouterClientAction _ipAgentAction = null;
	
 
	/**
	* Default Constructor.
	*/
	public IPThread()  {
		super();
		_endWith = '\n';
		_proxyAddress = null;
		_useProxy = false;
	}
		
	/**
	* Create ReceiverThread with specific thread name
	* @param name Agent name
	*/	
	public IPThread(String name)  {
		super(name);
		_endWith = '\n';
		_proxyAddress = null;
		_useProxy = false;
	}
		
	/**
	* Create ReceiverThread with specific thread name and Thread group.
	* Passing thread group is essential when you will create the receiver thread for an applet
	* because of security issue.
	* @param tg Thread group to which this thread will be owned to
	* @param name Agent name
	*/	
	public IPThread(ThreadGroup tg,String name)  {
		super(tg,name);
		_endWith = '\n';
		_proxyAddress = null;
		_useProxy = false;
	}
		

  /**
   *  This constructor is called when an agent want to initialize a connection
   *  to the other agents, using Address. 'security' object will be used to 
   *  satisfy connection protocol to server. Specific client information such as
   *  user name,password,email address,etc can be added to myaddress object as
   *  necessary.
   *  @param serveraddress The address of the remote host.
   *  @param priority The priority of the thread.
   *  @param myaddress My addres
   *  @param connections the connection table is passed as an argument.
   *  @param queue the Message Queue.
   *  @param security The security object 
   * @exception ConnectionException if error occurs
   */
  public IPThread(Address serveraddress,
		     int priority, 
		     Address myaddress,
		     Security security,
		     ThreadGroup tg)
    throws ConnectionException   {
      super(serveraddress,priority,myaddress,null,null,security,tg);    
  	  setEndWith('\n');
      _serverAddress = serveraddress;
      setSecurity(security);
			_useProxy = false;
  }
	
	/**
	  * Set agent action
	  * @param action IPRouterClientAction object
	  */
	public void setAgentAction(IPRouterClientAction action)  {
		_ipAgentAction = action;
	}
	
	/**
	  * Get agent action
	  * @return IPRouterClientAction object
	  */
	public IPRouterClientAction getAgentAction()  {
		return _ipAgentAction;
	}		

	/**
	  * Set command list otherwise default command list provided
	  * @param cmdlist Command list vector
	  */
	public void setCommandList(Vector cmdlist)   {
		_cmdList = cmdlist;
	}
	
	/**
	  * Get command list 
	  * @return Command list
	  */
	public Vector getCommandList()   {
		return _cmdList;
	}
	
	/**
	  * Set use proxy if an applet case
	  * @param yes True if use proxy
	  */
	public void useProxy(boolean yes)   {
		_useProxy = yes;
	}
		
	/**
	* Override run() method to provide virtual abstract method for processing commands
	* You needs not to override this method.
	*/
	public void run()   {
		processCommands();
		disconnect();
	}
	
	/**
	* Process command list. This is actual, main data transfer processing.
	* You should implement this method for a specific protocol.
	*/	
	protected abstract void processCommands();
	
	/**
  * You should use security object to satisfy connection protocol
  */  		
	public void connect()  throws ConnectionException   {
		if(_serverAddress == null)   {
			throw new ConnectionException("Server address is not specified. Can not connect");
		}
			
		// close input,output stream and socket first
    try  {
      if(_in != null)  {
				_in.close();
      }
      
      if(_out != null)  {
				_out.close();
      }
      
      if(_client != null)  {	
				_client.close();
      }	
      
      String host;
      int port;
      if(_useProxy)  {
      	host = _proxyAddress.getHost();
      	port = _proxyAddress.getPort();
      }
      else  {	
        host = _serverAddress.getHost();
        port = _serverAddress.getPort();
      }
System.out.println(host + "::" + port);      
      
      setSocket(new Socket(host,port));
      
    }  catch (Exception e)  {
    	throw new ConnectionException("Connection error " + e.toString());
    }
	
		
	  if(_security != null)   {
System.out.println("Start to connect");	  
	     try  {
          _security.processClientLogin(this,_serverAddress);
	     } catch (Exception e)  {
	     		endConn();	
	        throw new ConnectionException(e.toString());
	     }
	  }
	  else  {
	  	endConn();
	  	throw new ConnectionException("Security not specified. Connection failed");
	  }	
  }
	
		
	/**
  * You should use security object to satisfy connection protocol
  */  		
	public void disconnect()  {
		endConn();
	}	
		
		
}