package RouterLayer.AgentClient.Example.CalcServer;

import java.io.*;
import java.util.*;

import Abstract.*;
import KQMLLayer.*;
import RouterLayer.AgentClient.*;

/**
* A very simple example of a router client.
* The CalcServer will receive a message from other agents for integer arithmetics,
* addition, subtraction, multiplication and division. The content field expression
* is consistent with KIF.<BR>
* For addition, (+ x y), <BR>
*     subtraction, (- x y),<BR>
*	    multiplication, (* x y),<BR>
*     division, (/ x y).<BR>
* You can see how a router client will be setup and how to implement Act(Object) method.
*/

/*
* Copyright (c) 1996,1997,1998 <A HREF="http://cdr.stanford.edu/">Center for Design Research</A>, Stanford University.
* All rights reserved.<p>
*
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You can obtain the GNU General Public License at
* http://www.gnu.org/copyleft/gpl.html or by writing to the Free
* Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
* 02111-1307, USA.
* 
* This software is bound by the terms and conditions listed in the
* attached <a href="../LICENSE">LICENSE file</A>.
*  @author <A HREF="http://cdr.stanford.edu/html/people/jeon-bio.html>Heecheol Jeon</A>
*  @version Java(tm) 1.1
*/ 

public class CalcServer extends RouterClientAction  {

	/**
	* The return value, as a result of arithmetic. Example of a data member
	*/	
	int _returnValue;
	
	/**
	* Will be used by dynamic class downloading. The RCAddressHelper will use 
	* this constructor. After construction, setMyAddress(Address) and 
	* setRouterAddress(Address) should be invoked before invoke connect method.
	*/
	public CalcServer()  {
		super();
	}
		
	/**
	* Router client constructor. Since the agent is stand alone, it will 
	* pass its address and router address. Maximum idle time can be set, 
	* by passing the durationtime parameter.If the registerrequest is true, 
	* it will try to register to the router using registraraddress.
	* @param myaddress My address
	* @param routeraddress Router address
	* @param registraraddress Registrar address
	* @param durationtime Maximum idle time for the receiver thread
	* @param registerrequest If true, registration will be performed
	*/
	public CalcServer(Address myaddress,
										Address routeraddress,
										Address registraraddress,
										int durationtime)     {
		// use RouterLayer.AgentClient.RouterClientAction constructor								
		super(myaddress,routeraddress,registraraddress,durationtime);
	}
	
	/**
	* You should implement Act(Object) and processMessage(String,String[]) methods. 
	* In RouterClientAction class, those two methods are declared as abstract.
	* In this example, agent action is to add/subtract/multiply/divide two 
	* integer numbers, passed by a message. Since you should delete message 
	* after execution, you need to add the received message to the _mailQueue 
	* and send delete-message to the router(addToDeleteBuffer).
	* The content field of the received KQML message should be one of<BR>
	* (+ x y),(- x y),(/ x y),(* x y) where x and y are integers.<BR>
	* The KQML message performative should be ask-one and reply will be sent 
	* by this agent.
	* @param o Received KQMLmail string
	* @return false if error occurs.
	*/
 	public boolean Act(Object o)  {
 		String stampmsg = (String)o;

 		try  {
	 		//KQMLmail wrap the KQMLmessage
 			KQMLmail mail = new KQMLmail(stampmsg,0);
 			
 			// add the mail to the _mailQueue.
 			// this step is essential to delete message after execution
 			_mailQueue.addElement(mail);
 			
 			// use KQMLmail getKQMLmessage() to get KQMLmessage object
			KQMLmessage kqml = mail.getKQMLmessage();
			
			// this is a typical way to find out the performative
			String perf = kqml.getValue("performative");
			
			// this is a typical way to find out the field value
			String content = kqml.getValue("content");
			String receiver = kqml.getValue("sender");
			
			// should be an error condition
			if(content == null)  {
				sendErrorMessage(kqml);
				return false;
			}	
			if(perf.equals("ask-one"))  {
				// the content field will look like (+ 5 6) and takeOffList will constructor a
				// vector which contains as +,5,6 as its first,second and third element.
				// takeOffList is eliminating left, right parenthesis and construct a vector.
				// see KQMLLayer.takeOffList class.
				Vector v = new takeOffList(content);
				
				// if its size is not three, error(Note that this is not strictly consistent
				// with KIF). I want to be remained simple for user understanding.
				if(v.size() != 3)  {
					sendErrorMessage(kqml);
					return false;
				}
					
				// command should be one of +.-./,*	
				String command = (String)v.elementAt(0);
				
				// left number
				String left = (String)v.elementAt(1);
				// convert string to integer
				int ileft = (Integer.valueOf(left)).intValue();
				
				// right number
				String right = (String)v.elementAt(2);
				int iright = (Integer.valueOf(right)).intValue();
				
				if(command.equals("+"))  {
					_returnValue = ileft + iright;				
				}	
				else
				if(command.equals("-"))  {
					_returnValue = ileft - iright;
				}
				else
				if(command.equals("/"))  {
					if(iright == 0)  {
						// divide by zero
						sendErrorMessage(kqml);
						return false;
					}	
					else  {
						_returnValue = ileft / iright;	
					}	
				}
				else
				if(command.equals("*"))  {
					_returnValue = ileft * iright;
				}	
				
				// send result back to the receiver
				sendResult(receiver);
				
 				// send delete message to the Router
				addToDeleteBuffer(0);
				
			}
			else  {
				// error message
				sendErrorMessage(kqml);
				return false;
			}		
 		} catch (Exception e)  {
 			return false;
 		}
		
 		return true;
 }
	
	/**
	* Reply the result to the agent
	* @param receiver The message receiver. Not the router
	*/
	protected void sendResult(String receiver)  {
		String sendmsg = "(reply :sender ";
		sendmsg = sendmsg + getName() + " :receiver " + receiver;
		sendmsg = sendmsg + " :language KQML :content " + Integer.toString(_returnValue) + ")";
		
		// use sendMessage method of RouterClientAction
		try  {
			sendMessage(sendmsg);
		} catch (ConnectionException	c)  {
			System.out.println(c.toString());
		} catch (ParseException p)  {
			System.out.println(p.toString());
		}		
	}
	
	/**
	* Send error message of the received message
	* @param kqml KQMLmessage received
	*/
	protected void sendErrorMessage(KQMLmessage kqml)  {
		String receiver = kqml.getValue("sender");
		
		String sendmsg = "(error :sender ";
		sendmsg = sendmsg + getName() + " :receiver " + receiver;
		
		//use getSendString() method of the KQMLmessage object
		sendmsg = sendmsg + " :language KMQL :content (" + kqml.getSendString() + "))";	
		
		try  {
			sendMessage(sendmsg);
		} catch (ConnectionException c)  {
			System.out.println(c.toString());
		}  catch (ParseException p) {
			System.out.println(p.toString());
		}
					
		
		// should be deleted after execution. Send delete-message message to the Router
		addToDeleteBuffer(0);
	}
		
	// does not implement processMessage
	public void processMessage(String s,Object obj) {}

	/**
	* Main program
	*/
	public static void main(String args[])  {
		// if RCAddressHelper will be used,
		if(args.length == 0)  {
			// CalcServer class will be loaded dynamically. Its class name should be passed to the 
			// RCAddressHelper.
			RCAddressHelper ah = new RCAddressHelper("RouterLayer.AgentClient.Example.CalcServer.CalcServer");
			ah.show();
		}
		// address file path is passed as an argument
		else  {
			
			Address myAddress = null;
			Address routerAddress = null;
			Address registrarAddress = null;
			boolean registerRequest = false;
			int idletime = 1000;
			try  {
				BufferedReader in = 
					new BufferedReader(new FileReader(new File(args[0])));
			
				while(true)  {
					String line = in.readLine();
					if(line == null)  break;
				
					String next;
				
					if(line.startsWith("MyAddress"))  {
						next = in.readLine();
						myAddress = new Address(next);
					}
					else
					if(line.startsWith("RouterAddress"))  {
						next = in.readLine();
						routerAddress = new Address(next);
					}
					else
					if(line.startsWith("RouterRegistrarAddress"))  {
						next = in.readLine();
						registrarAddress = new Address(next);
					}
					else
					if(line.startsWith("RegisterRequest"))  {
						next = in.readLine();
						if(next.startsWith("y"))  {
							registerRequest = true;
						}
						else  {
							registerRequest = false;
						}		
					}
					else
					if(line.startsWith("MaxIdleTime"))  {
						next = in.readLine();
						idletime = (Integer.valueOf(next)).intValue();
					}	
				}
				
				in.close();
				
				CalcServer server = new CalcServer(myAddress,routerAddress,
							registrarAddress,idletime);				

				// Since stand alone, create server thread using my address
				server.createServerThread(myAddress.getID(),Thread.NORM_PRIORITY);
				
				if(registerRequest == true)  server.register(registrarAddress,myAddress);
		
				// connect to the router by sending my address
				server.connect(myAddress);
				
				// start to receive messages				
				server.start();				
			} catch (Exception e)  {
				System.out.println(e.toString());
				System.exit(0);
			}					
		}	
	}	
}