package RouterLayer.AgentClient;

import Abstract.*;
import BaseLayer.*;
import KQMLLayer.*;
import RouterLayer.Router.*;
import RouterLayer.AgentClient.*;
import RouterLayer.util.*;


import java.net.*;
import java.lang.*;
import java.io.*;
import java.util.*;

/**
* Template to support RouterClientAction. You should fill in processAction
* and Act method.
* The RouterClient is a KQML agent which just follow the RouterSecurity
* protocol for connection. However 'sendMessage' is different - the messages will be
* sent to the Router whoever the receiver is. And the Router will deliver the 
* message to the actual receiver. To maintain the Router more efficiently, it 
* would be better to send the delete-message message frequently.(It is just like
* the same as using e-mail. If your mail box has a lot of mails, the mail server's
* process speed will be very slow).
* @see RouterLayer.AgentClient.AgentAction.RCActionExample
*/

/*
* Copyright (c) 1996,1997,1998 <A HREF="http://cdr.stanford.edu/">Center for Design Research</A>, Stanford University.
* All rights reserved.<p>
*
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You can obtain the GNU General Public License at
* http://www.gnu.org/copyleft/gpl.html or by writing to the Free
* Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
* 02111-1307, USA.
* 
* This software is bound by the terms and conditions listed in the
* attached <a href="../LICENSE">LICENSE file</A>.
*  @author <A HREF="http://cdr.stanford.edu/html/people/jeon-bio.html>Heecheol Jeon</A>
*  @version Java(tm) 1.1
*/ 

public abstract class RouterClientAction extends KQMLAgentAction {
	
  /**
   * The Router name
   */
  protected String _routerName;
  
  /**
   * The registrar name
   */
  protected String _registrarName;
  
  /**
   * My address
   */
  protected Address _myAddress;
  
  /**
   * number of messages you have received from the Router 
   */
  protected int _messageNumber;	
  
  /**
   * Delete message number array for buffering the delete messages
   */
  protected int[] _deleteFPT;
  
  /**
   * Current buffered delete message size
   */
  protected int _deleteFPTSize;
  
  /**
   * Message repository
   */
  protected Vector _mailQueue;
  
  /**
   * Deleted message numbers will be piled up to the maximum size.
   * For more delete, actual delete operation will be happen on the
   * Router and AgentClient also should update the message number
   * Default is 1.
   */
  protected int _maxDeleteSize;
  
  /**
    * Default constructor. All data members will be initialized except
    * addresses for me, route and registrar. If you use this constructor,
    * you shoud use setMyAddress(Address),setRouterAddress(Address) and setRegistrarAddress(Address)
    * to initialize essential addresses.
    */	
  public RouterClientAction() {
  	initDataMembers();
  }
  
	/**
	  * Initialize all data members
	  */
  protected void initDataMembers()  {	
    try  {				
      _addresses = new BAddressTable();
      _security = new RouterSecurity(_addresses,null);
      
      // ConnectionTable 
      _connections = new BConnectionTable();
      // no message queue for the router
      _queue = new BMessageBuffer();				
      _endWith = '\004';
      _durationTime = -1;
      _messageNumber = 0;
      
      // delete operation for every delete command. No buffering(very inefficient)
      // if you want to increase the performance, increase this number(say 100)
      // AND set the RouterRecvThread _maxDeleteSize as the same as you have increased
      _maxDeleteSize = 1;
      
      // set the deletion related data members
      _deleteFPT = new int[_maxDeleteSize];
      _deleteFPTSize = 0;
      
      // Router client will retain its own mail queue for buffered deletion operation
      _mailQueue = new Vector();	
      System.out.println("Initialization success");			
      
    } catch (Exception ee)  {
      System.out.println(ee.toString());
      // critical error, can't start
      String args[] = new String[1];
      args[0] = ee.toString();
      processMessage("InitializeFailed",args);
    }		
  }

  
  /**
   * Constructor. Note that your address is not specified yet. You should use setMyAddress()
   * method before invoke connect() method.
   * @param routeraddress The Router address to connect
   * @param router registrar address 
   * @param durationTime Maximum idle time for the connection
   */	
  public RouterClientAction(Address routeraddress,
			    Address registraraddress,
			    int durationTime)  {
    this();
    try  {				
      setRouterAddress(routeraddress);
      setRegistrarAddress(registraraddress);
      setDurationTime(durationTime);
    } catch (Exception ee)  {
      System.out.println(ee.toString());
      // critical error, can't start
      String args[] = new String[1];
      args[0] = ee.toString();
      processMessage("InitializeFailed",args);
    }		
  }
  
  /**
   * Router AgentAction constructor with given my address,rotuer address and router registrar
   * address. 
   * @param myAddress My address
   * @param routerAddress The Router address
   * @param registrarAddress The Registrar address
   * @param durationTime Maximum idle time
   */
  public RouterClientAction(Address myAddress,
			    Address routerAddress,
			    Address registrarAddress,
			    int durationTime)  {
    this(routerAddress,registrarAddress,durationTime);
    setMyAddress(myAddress);
  }
  
  /**
   * Constructor using Address strings. Addresses string is basically a line of string :<BR>
	 * The address string should have the format :<BR>
   * (agent-address :name [agent-name] :host [host-inetaddress] :port [port no]<BR>
   *  :password [password] :KQML-extensions [kqml extensions] :email [email address]<BR>
   *  :description [agent description])<BR>
   * KQML-extensions, email and description is optional.<BR>
   * Example<BR>
   * (agent-address :name agentA :host java.stanford.edu :port 1234 :password xxxxxxxx<BR>
   * :KQML-extensions http://java.stanford.edu/newprotocol :email jhc@cdr.stanford.edu<BR>
   * :description (Network service))<BR> 
   * @param routeraddressstr The Router address string to connect
   * @param registraraddressstr The Router registrar address string
   * @param durationTime Maximum idle time for the connection
   */	
  public RouterClientAction(String routeraddressstr,
			    String registraraddressstr,
			    int durationTime)  {
    try  {				
    	Address routeraddress = ConvertStringToAddress(routeraddressstr);
    	Address registraraddress = ConvertStringToAddress(registraraddressstr);
    	if((routeraddress == null) || (registraraddress == null))  {
    		return;
    	}
    		
      setRouterAddress(routeraddress);
      setRegistrarAddress(registraraddress);
      setDurationTime(durationTime);
    } catch (Exception ee)  {
      System.out.println(ee.toString());
      // critical error, can't start
      String args[] = new String[1];
      args[0] = ee.toString();
      processMessage("InitializeFailed",args);
    }		
  }
  
  /**
   * Router AgentAction constructor with given my address,rotuer address and router registrar
   * address.   
   * The address string should have the format :<BR>
   * (agent-address :name [agent-name] :host [host-inetaddress] :port [port no]<BR>
   *  :password [password] :KQML-extensions [kqml extensions] :email [email address]<BR>
   *  :description [agent description])<BR>
   * KQML-extensions, email and description is optional.<BR>
   * Example<BR>
   * (agent-address :name agentA :host java.stanford.edu :port 1234 :password xxxxxxxx<BR>
   * :KQML-extensions http://java.stanford.edu/newprotocol :email jhc@cdr.stanford.edu<BR>
   * :description (Network service))<BR> 

   * @param myAddressstr My address string
   * @param routerAddressstr The Router address string
   * @param durationTime Maximum idle time
   */
  public RouterClientAction(String myAddressstr,
			    String routerAddressstr,
			    String registrarAddressstr,
			    int durationTime)  {
    this(routerAddressstr,registrarAddressstr,durationTime);
    setMyAddress(ConvertStringToAddress(myAddressstr));
  }
  
  /**
   * Router AgentAction constructor with given my address,rotuer address ,router registrar
   * address ,maximum idle time, and thread group.
   * @param myAddress My address
   * @param routerAddress The Router address
   * @param routerAddress The Router address
   * @param durationTime Maximum idle time
   * @param t Thread group
   */
  public RouterClientAction(Address myAddress,
			    Address routerAddress,
			    Address registrarAddress,
			    int durationTime,
			    ThreadGroup t)  {
		super(t,myAddress.getID() + "AgentAction");	
		initDataMembers();
    try  {				
    	setMyAddress(myAddress);
      setRouterAddress(routerAddress);
      setRegistrarAddress(registrarAddress);
      setDurationTime(durationTime);
    } catch (Exception ee)  {
      System.out.println(ee.toString());
      // critical error, can't start
      String args[] = new String[1];
      args[0] = ee.toString();
      processMessage("InitializeFailed",args);
    }		
  }
  
  /**
   * Set my address
   * @param address My address to be set
   */						 		 		
  public void setMyAddress(Address address)  {
    if(address != null)  {
      _addresses.addAddress(address);
      _myAddress = address;
      setName(address.getID());
    }
  }		
  
  /**
  * Set router address
  * @param addr The Router address
  */
  public void setRouterAddress(Address addr)  {
    if(addr != null)  {
      _addresses.addAddress(addr);
      _routerName = addr.getID();
    }	
  }
  
  /**
  * Set router registrar address
  * @param addr The rotuer registrar address
  */
  public void setRegistrarAddress(Address addr)  {
    if(addr != null)  {
      _addresses.addAddress(addr);
      _registrarName = addr.getID();
    }	
  }
  
  /**
   * Set my address using address string
   * @param addressstr My address string to be set
   */						 		 		
  public void setMyAddress(String addressstr)  {
  	this.setMyAddress(ConvertStringToAddress(addressstr));
  }		
  
  /**
  * Set router address using address string
  * @param addr The Router address
  */
  public void setRouterAddress(String addrstr)  {
  	this.setRouterAddress(ConvertStringToAddress(addrstr));
  }
  
  /**
  * Set router registrar address using address string
  * @param addr The rotuer registrar address
  */
  public void setRegistrarAddress(String addrstr)  {
  	this.setRegistrarAddress(ConvertStringToAddress(addrstr));
  }
  
  
  /**
   * Get my address
   * @return My address
   */
  public Address getMyAddress()  {
    return _addresses.getAddress(_myAddress.getID());
  }
  
  /**
   * Connect to the Router. Addrss should be passed. If the connection is already exist,
   * it first disconnect the connection and try to reconnect. 
   * @param addr My address to be sent to the Router
   * @exception ConnectionException for any error occurs during processing connect. If addr
   * is null, it will also throws ConnectionException
   */	
  public synchronized void connect(Address addr)  
    throws ConnectionException  {
      if(addr == null)  {
	throw new ConnectionException("Address not specified");
      }
      
      if(!_connections.isLiving(_routerName))  {
	setMyAddress(addr);
	ReceiverThread thread = createReceiverThread(_routerName,addr.getID(),Thread.NORM_PRIORITY);
	if(thread == null)  {
	  throw new ConnectionException("Connection to the Router failed");
	}	
      }	
  }
  
  /**
   * Connect to the Router, using the agent name
   * @param id My name
   */	
  public synchronized void connect(String id)
    throws ConnectionException  {
      this.connect(_addresses.getAddress(id));
  }
  
  /**
   * Connect to the Router, using current agent name
   */	
  public synchronized void connect()
    throws ConnectionException  {
    	if(_myAddress == null)  {
    		throw new ConnectionException("My address not specified");
    	}
    		
      this.connect(_myAddress.getID());
  }
  
  /**
   * Disconnect from the Router.
   */
  public void disconnect()  {
    ReceiverThread thread = _connections.getConnection(_routerName);
    if(thread != null)  {
      _mailQueue.removeAllElements();
      _deleteFPTSize = 0;
      _messageNumber = 0;
      _queue.removeAllElements();
      _connections.remove(_routerName);
      thread.endConn();
    }
  }
  
  /**
   * Router utility to register. Declared to be as a critical section, to finish
   * registration before connect.
   * @param registrar The Router Registrar address
   * @param myaddress My address to be registered
   * @exception ConnectionException for any error occurs during processing this service
   */	
  public synchronized void register(Address registrar,Address myaddress)  
    throws ConnectionException  {
	_addresses.addAddress(myaddress);
	_addresses.addAddress(registrar);
	this.register();
  }	
  
  /**
   * Router utility to register. Declared to be as a critical section, to finish
   * registration before connect. Before you invoke this method, my address and the
   * registrar address should be set(either by constructor or setRegistrarAddress,
   * setMyAddress methods.
   * @exception ConnectionException for any error occurs during processing this service
   */	
  public synchronized void register()  
    throws ConnectionException  {
      try  {
      	if((_myAddress == null) || (_registrarName == null))  {
	  throw new ConnectionException("Addresses not specified");
      	}	
	RegistrarRecvThread rth = new RegistrarRecvThread(_addresses.getAddress(_registrarName),
							  Thread.NORM_PRIORITY,_myAddress,null,null,new RegistrarSecurity((RouterSecurity)_security),getThreadGroup());
      } catch (Exception e)  {
	throw new ConnectionException(e.toString());
      }		
  }	
  
  
  /**
   * Router utility to unregister
   * @param myaddress My address to be unregistered
   * @exception ConnectionException for any error during processing this service
   */
  public synchronized void unregister(Address myaddress) 
    throws ConnectionException   {
    _myAddress = myaddress;
    unregister();
  }	
  
  /**
   * Router utility to unregister
   * @exception ConnectionException for any error during processing this service
   */
  public synchronized void unregister() 
    throws ConnectionException   {
      try  {
	String description = _myAddress.getDescription();
	String password = null;
	if(description != null)  {
	  KQMLmessage passkqml = new KQMLmessage(description);
	  password = passkqml.getValue("password");
	}	
	
	String sendstr = "(unregister-agent :sender ";
	sendstr = sendstr + _myAddress.getID() + " :receiver " + _routerName;
	if(password != null)  {
	  sendstr = sendstr + " :password " + password;
	}
	sendstr = sendstr	 + ")";
	sendKQMLMessage(sendstr);
      } catch (Exception e)  {
	throw new ConnectionException(e.toString());
      }		
  }	
  
  
  /**
   * Router utility for agent name service. Request the other agent address by passing
   * its name.<BR>
   * Message format:<BR>
   * (request-address :sender xx :receiver Router :content Xname)<BR>
   * Returned message format from the Router:<BR>
   * (address :sender Router :receiver xx :name Xname :host cdr.stanford.edu :port 1234
   * :message-method MessageRouter :description (agent-info :service sort-service))<BR>
   * @param name Agent name you want to get the address information
   * @exception ConnectionException for any error during the processing this service
   */	
  public void requestAddress(String name)
    throws ConnectionException  {
      try  {
	String sendstr = "(request-address :sender ";
	sendstr = sendstr + _myAddress.getID() + " :receiver " + _routerName;
	sendstr = sendstr + " :content " + name + ")";
	sendKQMLMessage(sendstr);
      } catch (Exception e)  {
	throw new ConnectionException(e.toString());
      }		
  }
  
    /**
   * Router utility for agent name service. Request the other agent address by passing
   * its name.<BR>
   * Message format:<BR>
   * (request-address :sender xx :receiver Router :content Xname)<BR>
   * Returned message format from the Router:<BR>
   * (address :sender Router :receiver xx :name Xname :host cdr.stanford.edu :port 1234
   * :message-method MessageRouter :description (agent-info :service sort-service))<BR>
   * @param name Agent name you want to get the address information
   * @exception ConnectionException for any error during the processing this service
   */	
  public void requestAddress(String name,String replyWith)
    throws ConnectionException  {
      try  {
	String sendstr = "(request-address :sender ";
	sendstr = sendstr + _myAddress.getID() + " :receiver " + _routerName;
	if((replyWith != null) && (!replyWith.equals("")))  {
	  sendstr = sendstr + " :reply-with " + replyWith;
	}	
	sendstr = sendstr + " :content " + name + ")";
	sendKQMLMessage(sendstr);
      } catch (Exception e)  {
	throw new ConnectionException(e.toString());
      }		
  }
  
  /**
   * Router utility to request list the current registered agent to the Router<BR>
   * Message format:<BR>
   * (list-agent :sender xx :receiver Router)<BR>
   * Returned message format from the Router<BR>
   * (registered-agent :sender Router :receiver xx :content (aaa bbb ccc))<BR>
   * @exception ConnectionException for any error during processing the message
   */	
  public void listUsers()
    throws ConnectionException  {
      try  {
	String sendstr = "(list-agent :sender ";
	sendstr = sendstr + _myAddress.getID() + " :receiver " + _routerName + ")";
	sendKQMLMessage(sendstr);
      }	catch (Exception e)  {
	throw new ConnectionException(e.toString());
      }	
  }
  
  /**
   * Send 'reserve-message' message to the Router
   * The reserve-message will follow the message format:<BR>
   * (reserve-message :sender xx :receiver Router :content (1 3 5 7) :time 888837363524
   * :address (receive-address :name xx :host cdr.stanford.edu :port 4352 :message-method
   * MessageRouter :description (agent-info :password zzz)))<BR>
   * The 'receive-address' name slot represent that the intended receiver may be different
   * from the 'sender'. You can reserve message for other agent.<BR>
   * The 'address' slot is optional.
   * @param time Time stamp
   * @param mailNoList list of message numbers
   * @param address Intended receiving address. Null if the receiving location is not different
   * @exception ConnectionException if the given information is incomplete
   */
  public void reserveMessage(long time,String mailNoList,Address address)
    throws ConnectionException   {
      try  {
	String sendstr = "(reserve-message :sender ";
	sendstr = sendstr + _myAddress.getID() + " :receiver " + _routerName;
	sendstr = sendstr + " :content " + mailNoList;
	sendstr = sendstr + " :time " + Long.toString(time);
	if(address != null)  {
	  sendstr = sendstr + " :address (receiver-address :name " + address.getID();
	  sendstr = sendstr + " :host " + address.getHost();
	  sendstr = sendstr + " :port " + address.getPort();
	  sendstr = sendstr + " :message-method " + address.getType();
	  if(address.getDescription() != null)  {
	    sendstr = sendstr + " :description  " + address.getDescription();
	  }
	  sendstr = sendstr	 + ")";
	}	
	sendstr = sendstr + ")";
	sendKQMLMessage(sendstr);
      }	catch (Exception e)  {
	throw new ConnectionException(e.toString());
      }	
  }	
  
  /**
   * Make message reservation content from given time,message number array and reserved
   * address and send 'reserve-message' message to the Router
   * @param time Time stamp
   * @param a[] Reserved message numbers
   * @param address Reserved location if different from the current address
   */ 	
  public void reserveMessage(long time,int a[],Address address)
    throws ConnectionException  {
      String content = "(";
      for(int i=0; i< a.length; i++)  {
	content = content + Integer.toString(a[i]) + " ";
      }
      this.reserveMessage(time,content,address);
  }
  
  /**
   * @param receiver Whoever the receiver is, the message will be sent to the Router
   * @param msg Message to be sent
   * @exception Abstract.Connection
   */
  public void sendMessage(String receiver,String msg) 
  throws ConnectionException {
    if(_routerName == null) {
      throw new ConnectionException("Router name is not specified. Message sending aborted");
    }

    super.sendMessage(receiver,msg);
  }	
  
  /**
   * Send message to the router
   * @param msg Message to be sent to the Router
   * @exception Abstract.ConnectionExcaption 
   */
  public void sendMessage(String msg)  
  throws ConnectionException, ParseException   {
  	// pre-check for KQML parsing
  	KQMLmessage kqml = new KQMLmessage(msg);
    super.sendMessage(_routerName,msg);
  }	
  
  /**
   * Send KQMLmessage. 
   * @param kqml KQMLmessage object
   */
  public void sendMessage(KQMLmessage kqml) 
  throws ConnectionException   {
    sendMessage(_routerName,kqml.getSendString());
  }


  /**
   * Send message to the router. Message will be parsed and checked before sent to router. If KQML parse error occurs, ParseException will be thrown. This method is safer than sendMessage(String msg).
   * @param msg Message to be sent to the Router
   * @exception Abstract.ConnectionExcaption, KQMLLayer.ParseException.
   */
  public void sendKQMLMessage(String msg)  
  throws ConnectionException, ParseException   {
    if(_routerName == null) {
      throw new ConnectionException("Router name is not defined. Message sending aborted");
    }
    KQMLmessage kqml = new KQMLmessage(msg);
    super.sendMessage(_routerName,kqml.getSendString());
  }	
  
  /**
   * Get the Router name
   * @return Router name the agent is registered to and communicate witl
   */	
  public String getRouterName()  {
    return _routerName;
  }
  
  /**
   * Get mail queue
   * @return Mail queue
   */	
  public Vector getMailQueue()  {
    return _mailQueue;
  }
  
  /**
   * Get maximum delete buffer size.
   * @return Maximum delete buffer size
   */	
  public int getMaxDeleteSize()  {
    return _maxDeleteSize;
  }
  
  /**
   * You should implement this method.
   * @param command Command to be executed
   * @param obj Reference object
   */		
  public abstract void processMessage(String command,Object obj);
  
  /**
   * Since message format is different(reading tag and time stamp), this
   * method should be overriden. At least, you need to add the received
   * KQMLmail to the _mailQueue. 
   * @param o KQMLmail object.
   */
  public boolean Act(Object o)  {
    String stampmsg = (String)o;
    
    //KQMLmail wrap the KQMLmessage
    try  {
      KQMLmail mail = new KQMLmail(stampmsg,_mailQueue.size());
      _mailQueue.addElement(mail);
      ////////////////////////////////////////
      //
      // Fill in what you need to do here
      //
      ///////////////////////////////////////
    } catch (Exception e)  {
      System.out.println(e.toString());
    }	
    return true;
  }	
  
  /**
   * If a message is performed, you should request to delete the message
   * for the Router to maintain the incoming files. Since Router will
   * update the deleteMessage when the number of deletedMessage Size is
   * _maxDeleteSize, the message number after delete _maxDeleteSize messages will be updated
   * @param mailno Message number to be deleted
   */
  public void addToDeleteBuffer(int mailno)  {
    _deleteFPT[_deleteFPTSize] = mailno;
    _deleteFPTSize++;
    if(_deleteFPTSize == _maxDeleteSize)  {
      sendDeleteMessage();
    }	
    //System.out.print("MAILSIZE " + _mailQueue.size());		
  }				
  
  /**
   * Update the message number when the deleted message number is upto _maxDeleteSize.
   */
  public void sendDeleteMessage()  {
  	try  {
    	QSort qsort = new QSort();
    	// sort needs size - 1		
    	qsort.sort(_deleteFPT,0,_deleteFPTSize - 1);
    
    	int mailSize = _mailQueue.size();
    	// temporary queue to retain the undeleted vector
    	Vector tmpQueue = new Vector(mailSize - _maxDeleteSize);
    
    
    	String deletemsg = "(delete-message :sender ";
    	deletemsg = deletemsg + _myAddress.getID() + " :receiver " + _routerName;
    
    	int i = 0;
    	// global counter
    	int count = 0;	
    	for(int j = 0; j<_deleteFPTSize;)   {
      	// _deleteFPT[i] is greater than the i(current index). Save.
      	while((_deleteFPTSize > j) && (_deleteFPT[j] > i))	{
					// add to the tmp queue
					KQMLmail mail = (KQMLmail)_mailQueue.elementAt(i);
					mail.setMessageNumber(count);	
					count++;
					tmpQueue.addElement(mail);
					i++;
      	}
      
      	while((_deleteFPTSize > j) && (_deleteFPT[j] == i)) {
					//System.out.println("Removed" + i);			
					String msg = deletemsg;
					msg = msg + " :content " + i + ")";
					sendKQMLMessage(msg);				
					i++;
					j++;
      	}
      
      	// remove the same elements in _deleteFPT, may happen
      	while((j < _deleteFPTSize) && (i > _deleteFPT[j])) {
					//System.out.println("Greater " + _deleteFPT[j] + " " + i);				
					j++;
      	}	
      
      
    	}				
    	//System.out.println("mailsize " + mailSize);			
    	// if rest
    	for(; i < mailSize; i++)   {
      	KQMLmail mail = (KQMLmail)_mailQueue.elementAt(i);
      	mail.setMessageNumber(count);	
      	count++;
      	tmpQueue.addElement(mail);
    	}					
    
    	_mailQueue.removeAllElements();
    
    	for(i=0; i<tmpQueue.size(); i++)  {
      	_mailQueue.addElement(tmpQueue.elementAt(i));
    	}	
    	//System.out.println("remaining size " + _mailQueue.size());
    	_deleteFPTSize = 0;
    
    	tmpQueue.removeAllElements();
    }  catch (Exception e)  {
    	System.out.println(e.toString());
    }		
    
  }	
  
  /**
   * Passing integer array which contains the message number to be deleted.
   * Internally, if the deleted message size is over the _maxDeleteSize, the message
   * numbers will be updated.
   * @param a[] Current message numbers to be deleted
   */	
  public void addToDeleteBuffer(int a[])  {
    int total = a.length + _deleteFPTSize;
    if(total < _maxDeleteSize)  {
      for(int k=0; k<a.length; k++)  {
	addToDeleteBuffer(a[k]);
      }
    }
    else  {		  	
      int[] b = new int[total];
      int i;
      for(i=0; i< a.length; i++)  {
	b[i] = a[i];
	//System.out.print(b[i]);  			
      }
      
      int j = 0;
      for(; i < total; i++)  {	
	b[i] = _deleteFPT[j];
	//System.out.print(b[i]);  			
      }
      
      QSort qsort = new QSort();
      // sort needs size - 1		
      qsort.sort(b,0,b.length - 1);
      
      _deleteFPTSize = 0;
      
      int loopcount = 0;
      int count = 0;
      for(i=0; i < total; i++)  {
	//System.out.print("B[i]" + b[i] +  "loop" + loopcount);		
	KQMLmail mail = (KQMLmail)_mailQueue.elementAt(b[i] - loopcount);
	//System.out.print("Mno" + (mail.getMessageNumber()));			
	addToDeleteBuffer(mail.getMessageNumber());
	if(_maxDeleteSize == ++count)  {
	  loopcount += _maxDeleteSize;
	  count = 0;
	}
      }	
    }
  }			
  
  
  /**
   * Override endAction() to send buffered delete-messages
   */
  public void endAction()  {
    try  {
      if(_deleteFPTSize > 0)  {
	sendDeleteMessage();
      }	  
    } catch (Exception d)  {
      System.out.println(d.toString());
    }		
    super.endAction();			
  }		
  
  /**
  * Parse KQML address string and convert to Address object.   
  * The address string should have the format :<BR>
  * (agent-address :name [agent-name] :host [host-inetaddress] :port [port no]<BR>
  *  :password [password] :KQML-extensions [kqml extensions] :email [email address]<BR>
  *  :description [agent description])<BR>
  * KQML-extensions, email and description is optional.<BR>
  * Example<BR>
  * (agent-address :name agentA :host java.stanford.edu :port 1234 :password xxxxxxxx<BR>
  * :KQML-extensions http://java.stanford.edu/newprotocol :email jhc@cdr.stanford.edu<BR>
  * :description (Network service))<BR> 
  * @param addrstr Address string
  * @return Address object. If wrong format, return null
  */
  public Address ConvertStringToAddress(String kqmladdr)  {
  	Address retaddr = null;
  	try  {
  		KQMLmessage kqml = new KQMLmessage(kqmladdr);
  		String name = kqml.getValue("name");
  		String host = kqml.getValue("host");
  		String port = kqml.getValue("port");
  		String messagemethod = kqml.getValue("message-method");
  		String password = kqml.getValue("password");
  		String kqmlext = kqml.getValue("KQML-extensions");
  		String email = kqml.getValue("email");
  		String desc = kqml.getValue("description");
  		
			if((name == null) || (host == null) || (port == null) || (messagemethod == null))  {
				return null;
			}	  			 
  		
  		String newdesc = "(agent-info ";
  		if(password != null)  {
  			newdesc = newdesc + " :password " + password;
  		}
  		if(kqmlext != null)  {
  			newdesc = newdesc + " :KQML-extensions " + kqmlext;
  		}
  		if(email != null)  {
  			newdesc = newdesc + " :email " + email;
  		}
  		if(desc != null)  {
  			newdesc = newdesc + " :description " + desc;
  		}
  		newdesc = newdesc + ")";

			retaddr = new Address(name,host,(Integer.valueOf(port)).intValue(),messagemethod,newdesc);
		}  catch (Exception e)  {
			System.out.println(e.toString());
		}
		
		return retaddr;
	}			  		
} 		

 		
 