package RouterLayer.Router;

import java.net.*;
import java.io.*;
import java.util.*;

import Abstract.*;
import BaseLayer.*;
import RouterLayer.util.TimerThread;

/**
* Lightweight ReceiverThread, which is used only for the Router registration.
* After security check, it will stop. You can set timer by invoking setDurationTime(int minutes)
* method.
* @see BaseLayer.ReceiverThread
* @see BaseLayer.ConnectionTable
* @see RouterLayer.util.TimerThread
*/

/* 
* Copyright (c) 1996,1997,1998 <A HREF="http://cdr.stanford.edu/">Center for Design Research</A>, Stanford University.
* All rights reserved.<p>
*
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You can obtain the GNU General Public License at
* http://www.gnu.org/copyleft/gpl.html or by writing to the Free
* Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
* 02111-1307, USA.
* 
* This software is bound by the terms and conditions listed in the
* attached <a href="../LICENSE">LICENSE file</A>.
*  @author <A HREF="http://cdr.stanford.edu/html/people/jeon-bio.html>Heecheol Jeon</A>
*  @version Java(tm) 1.1
*/ 

public class RegistrarRecvThread extends BRecvThread {

	TimerThread _timer;

  /**
   * Default Constructor.
   */
  public RegistrarRecvThread()  {
    super();
  }
  
  /**
   * Conctructor with the Agent name
   * @param name Agent name
   */	
  public RegistrarRecvThread(String name)  {
    super(name);
  }
  
  /**
   * Constructor. 
   * This constructor will be invoked by RegistrarServerThread
   * @param client_scoket The socket for the remote client;
   * @param priority The priority of the thread.
   * @param address Registrar address
   * @param connections the connection table is passed as an argument.
   * @param queue the Message Queue. Normally null
   * @param security RegistrarSecurity
   */
  public RegistrarRecvThread(Socket client_socket,
			     int priority, 
			     Address address,
			     ConnectionTable connections,
			     MessageBuffer queue,  // will be null
			     Security security)
    throws ConnectionException    {
      super(client_socket,priority,address,connections,queue,security);
  }
  
  
  /**
   *  This constructor is called when an agent want to initialize a connection
   *  to the Router Registrar, using Address. 
   *  @param serveraddress The address of the Router Registrar.
   *  @param priority The priority of the thread.
   *  @param myaddress My addres
   *  @param connections the connection table is passed as an argument.
   *  @param queue the Message Queue. Normally, null
   *  @param security The RouerSecurity object 
   */
  public RegistrarRecvThread(Address serveraddress,
  			     int priority, 
			     Address myaddress,
			     ConnectionTable connections,
			     MessageBuffer queue, // will be null
			     Security security,
			     ThreadGroup tg)	
    throws ConnectionException   {
      super(serveraddress,priority,myaddress,connections,queue,security,tg);
  }
  
  
  /**
   * Override Thread run() method. Only check the RouterSecurity and stop()
   * Timer will be set to the limited time. Use setDuariionTime(int minutes) to set
   * the time limit. If the time is elasped over the setting time, the Timer object
   * will automatically kill this thread
   */ 
  public void run()   {
    // set timer. If time goes upto limited time, this thread will be killed automatically
    _timer = new TimerThread(this,_durationTime);
    // check security first	
    if(_security != null)    {
      // server,
      if(_role == true)  {
      	// if error occurs while procesing server login, security will immediately
      	// return and it will not directly invoke endConn() method.
				_security.processServerLogin(this);
      }	
    }	
    
    // succeful registration. Kill Timer
    if((_timer != null) && _timer.isAlive())  {
    	_timer.stop();
    	_timer = null;
    }			
  }		
  
  
  /**
   * Close the connection. Close socket,input stream and output stream
   */
  public synchronized void endConn()   {
    System.out.println("endConn");
    try {
      if(_in != null)  {
				_in.close();
				_in = null;
      }
    }  catch (IOException e2)  {
      System.out.println(e2.toString());
    }
    
   try {

	    if(_out != null)  {
					_out.close();
					_out = null;
    	}
    }  catch (IOException e3)  {
      System.out.println(e3.toString());
    }
    	
    	
    
    try  {  
      if(_client != null)  {	
				_client.close();
				_client = null;
      }	
    } catch (IOException e4)   {
      System.out.println(e4.toString());
    }
    
    if(isAlive())  {
    	stop();
    	try  {
    		join();
    	} catch (InterruptedException e1)  {
    		System.out.println(e1.toString());
    	}	
    }	
  }
  
}