package RouterLayer.Router;

import Abstract.*;
import BaseLayer.*;
import KQMLLayer.*;
import RouterLayer.util.*;

import java.net.*;
import java.lang.*;
import java.io.*;
import java.util.*;

/**
* Router server action class. 
* Router performs to route messages to the receiver agent and keep the messages to the 
* file until the receiver agent request to delete the messages.
* <UL>Router server features<BR>
* <LI>Route a message to the receiver and keep it to the receiver file<BR>
* <LI>Applets can send/receive messages using Router<BR>
* <LI>If the receiver agent is on-line, deliver the message promptly<BR>
* <LI>If the receiver agent is off-line, try to initiate the connection to the receiver
* if the receiver has ServerThread. If initialization is successful, send the message
* promptly. If initialization fails, keep it to the receiver file. Afterward, if 
* the receiver is connected to the Router, the Router sends received messages to the
* receiver<BR>
* <LI>Agent can register to the Router through on-line.<BR>
* <LI>RouterSecurity will check the password. Therefore, when an agent registers to the
* Router, it should send its password.<BR>
* <LI>The messages can be reserved to be sent later to the specific location in a specific
* time<BR>
* <LI>The agents can request the other agents address(Agent Name Service)<BR>
* <LI>The agents can request the current on-line agent list.<BR>
* </UL>
* The Router server will open two ServerSocket, one for accepting the registration and
* one for accepting the connection from the registered agents. The Router delegate its
* utilities(ANS,list-users,delete-message) to the RouterRecvThread. The Router does not
* have centralized MessageBuffer. Each RouterRecvThread has its own MessageBuffer and
* work invidually. For each router client, one unique RouterRecvThread will be created
* by accepting the conneciton. If a message is received from the connection the RouterRecvThread
* will parse the message to find out the receiver. Then, if the receiver is on-line,
* the message will be stored to the receiver RouterRecvThread's MessageBuffer after
* adding time stamp. While looping, the RouterRecvThread check its MessageBuffer. If the
* MessageBuffer is not empty, it will directly send the message through its connection
* and save it to its incoming file.(Note that by this approach, we can evade critical
* section when save/retrieve the message from the file).
* @see RouterLayer.Router.Connection.RouterRecvThread
*/

/*
* Copyright (c) 1996,1997,1998 <A HREF="http://cdr.stanford.edu/">Center for Design Research</A>, Stanford University.
* All rights reserved.<p>
*
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You can obtain the GNU General Public License at
* http://www.gnu.org/copyleft/gpl.html or by writing to the Free
* Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
* 02111-1307, USA.
* 
* This software is bound by the terms and conditions listed in the
* attached <a href="../LICENSE">LICENSE file</A>.
*  @author <A HREF="http://cdr.stanford.edu/html/people/jeon-bio.html>Heecheol Jeon</A>
*  @version Java(tm) 1.1
*/ 

public class RouterAction extends KQMLAgentAction  {

  /**
   * Incoming message file directory path. You should specify this data member for the
   * Router to save the incoming messages to the receiver file. The incoming file for
   * each registered agent will be created under the messageBoxPath directory
   */
  protected String _messageBoxPath;
  
  /**
   * Hashtable which maintains the reserved messages. The key field is time stamp and
   * the field value is Reservation object.
   * @see RouterLayer.Router.AgentAction.Reservation
   */
  protected Hashtable _reservedBook;
  
  /**
   * Hashtable which maintains the off-line agent list. An agent will be added to the 
   * off line list, if the receiver thread is time out or the Router can not initiate the
   * connection to the agent. Note that if an agent send disconnect-agent message to the
   * Router, the Router will never initiate the connection. Key is time and the field is 
   * a reservation object.
   */
  protected OffLineAgentTable _offLineList;
  
  protected int _offLineTrialTimes;
  protected int _offLineCheckPeriod;	
  
  protected int _reserveTrialTimes;
  protected int _reserveCheckPeriod;
  
  protected int _sleepPeriod;
  
  protected Vector _disconnectTable;
  
  /**
   * Set the ReservationTable
   * @param reservedBook ReservationTable from the RouterAction
   */
  public void setReservedBook(Hashtable reservedBook)  {
    _reservedBook = reservedBook;
  }	
  
  /**
  * Get reservation table
  * @return Reserved message table
  */
  public Hashtable getReservedBook()  {
    return _reservedBook;
  }
  
  /**
   * Set the Off line agent list
   * @param offlinelist Off line agent table from the RouterAction
   */
  public void setOffLineAgentList(OffLineAgentTable offlinelist)  {
    _offLineList = offlinelist;
  }	
  
  /**
   * Get reservation table
   * @return Reserved message table
   */
  public OffLineAgentTable getOffLineAgentList()  {
    return _offLineList;
  }
  
  
  
  /**
   * Default constructor. You should use setAddressFilePath, setMessageBoxPath, setPasswordFilePath
   * methods afterwards.
   */
  public RouterAction()  {
  }
  
  /**
   * Constructor 
	* @param id Router name
	* @param addressFilePath Address file path. Address file contains the addresses of Router and 
	*  RouterRegistrar at the minimum. Add on those, you can specify the addresses of other agents
	*  where the Router want to initiate the conneciton
	* @param messageBoxPath Message incoming file directory path
	* @param passwordFilePath Password file path
	* @param registryFilePath Registry file path
	* @param duration Maximum idle time
	* @param offLineTrialTimes Maximum trial times to voluntarily reconnect to stand alone agent
	* @param offLineCheckPeriod Off line agent time check period in minutes
	* @param reserveTrialTimes Maximum trial times to send reserved message 
	* @param reserveCheckPeriod Reseved message sending time check period in minutes
	* @param sleepPeriod Sleep period in minutes
   */
  public RouterAction(String id,
		      String addressFilePath,
		      String messageBoxPath,
		      String passwordFilePath,
		      String registryFilePath,
		      int durationTime,
		      int offLineTrialTimes,
		      int offLineCheckPeriod,
		      int reserveTrialTimes,
		      int reserveCheckPeriod,
		      int sleepPeriod)  {
    super(id);
    try  {
      
      // Read AddressTable file and make AddressTable
      BufferedReader input = new BufferedReader(new FileReader(new File(addressFilePath)));
      _addresses = new BAddressTable(input);
      input.close();
      
      _messageBoxPath = messageBoxPath;
      _security = new RouterSecurity(_addresses,passwordFilePath,registryFilePath);
      
      if(passwordFilePath != null)  {
	//want to check the password
	_security.setPasswordCheck(true);
      }
      
      // ConnectionTable
      _connections = new BConnectionTable();
      // no message queue for the router
      _queue = null;
      _endWith = '\004';
      _durationTime = durationTime;
      _reservedBook = new Hashtable();
      _offLineTrialTimes = offLineTrialTimes;
      _offLineCheckPeriod = offLineCheckPeriod;	
      
      _offLineList = new OffLineAgentTable(offLineTrialTimes,offLineCheckPeriod);
      _disconnectTable = new Vector();
      _reserveTrialTimes = reserveTrialTimes;
      _reserveCheckPeriod = reserveCheckPeriod;
      _sleepPeriod = sleepPeriod;
      
      
      // Create all the static connection
      Enumeration e = _addresses.elements();
      while(e.hasMoreElements())  {
	Address addr = (Address)e.nextElement();
	String type = addr.getType();
	// Only one Server thread is running.
	// My Agent Name should be the same as the Server ID
	// If you want multi-ServerSocket with Router, this part should be
	// modified
	if(type.equals("RouterServerThread"))  {
	  String serverid = addr.getID();
	  // error, will exit()
	  if(!id.equals(serverid))   {
	    processMessage("WrongServerAddress",null);
	  }
	  else  {
	    createServerThread(id,Thread.NORM_PRIORITY);
	  }
	}
	else
	  // receiver thread
	  if(type.equals("BRecvThread"))  {
	    if(addr.getPort() != -1)  {
	      try  {
		createReceiverThread(addr.getID(),Thread.NORM_PRIORITY);
	      } catch (Exception ee)  {
		System.out.println(ee.toString());
		// the receiver thread may fail if the server is not running
		// just ignore and if a message is to be sent, try to reconnect
	      }
	    }
	  }
	if(type.equals("RegistrarServerThread"))  {
	  RegistrarSecurity rs = new RegistrarSecurity((RouterSecurity)_security);
	  // send Router address to introduce the Router
	  RegistrarServerThread rth = new RegistrarServerThread(_addresses.getAddress(id),addr,null,null,rs);
	  rth.setPriority(Thread.NORM_PRIORITY);
	  rth.setDurationTime(10);  // 10 minutes to durate for the clients
	  rth.setEndWith(_endWith);
	  rth.start();
	}	
	else  {
	  String args[] = new String[1];
	  args[0] = type;
	  processMessage("UnsupportedType",args);
	}
      }
      System.out.println(id + " is running");
    } catch (Exception ee)  {
      // critical error, can't start
      String args[] = new String[1];
      args[0] = ee.toString();
      processMessage("InitializeFailed",args);
    }
  }
  
  /**
   * Override AgentAction createServerThread method
   * Create server thread using name of the agent
   * Override this method if your connection type is different.
   * @param id ServerThread name
   * @param priority Server thread priority
   * @return ServerThread, generated. Null if fails
   */
  public ServerThread createServerThread(String id,int priority)  {
    ServerThread server = null;
    try  {
      server =
	new RouterServerThread(_addresses.getAddress(id),_connections,_security,_messageBoxPath,_reservedBook,_offLineList,_disconnectTable);
      server.setPriority(priority);
      server.setDurationTime(_durationTime);
      server.setEndWith(_endWith);
      server.start();
    } catch (Exception e)  {
      System.out.println(e.toString());
    }
    
    return server;
  }
  
  
  /**
   * Override AgentAction createRecvThread method
   * Create client receiver thread using name of the agent
   * Override this method if your connection type is different.
   * Before you invoke this method, the Router name(the same as RouterAction thread name)
   * should be specified.
   * @param serverid Server agent name. This name is not my name. The name of the other
   * agent you want to connect
   * @param priority Receiver thread priority
   * @return ReceoverThread, generated
   */
  public ReceiverThread createReceiverThread(String serverid,int priority)  {
    ReceiverThread client = null;
    // Server address should be known
    Address addr = _addresses.getAddress(serverid);
    if(addr == null)  {
      System.out.println("Client Address for " + serverid + " is not specified");
      return null;
    }
    try  {
      client = new RouterRecvThread(addr,priority,_addresses.getAddress(getName()),_connections,new BMessageBuffer(),_security,_messageBoxPath,_reservedBook,_offLineList,_disconnectTable,getThreadGroup());
      
      client.setEndWith(_endWith);
      client.setDurationTime(_durationTime);
      client.setName(addr.getID());
      client.start();
      System.out.println("Client to " + serverid + " started");
    } catch (Exception e)  {
      System.out.println(e.toString());
    }
    
    return client;
  }
  
  /**
   * Get Router address
   * @return The Router address
   */
  public Address getAddress()  {
    return _addresses.getAddress(getName());
  }
  
  /**
   * Override processMessage of AgentAction.
   * @param command Command to be executed
   * @param obj String array arguments
   */
  public void processMessage(String command,Object obj)  {
    String[] args = (String[])obj;
    if(command.equals("WrongServerID"))  {
      System.out.println("Your first argument does not match with the AddressFile ServerID");
      System.exit(1);
    }
  }
  
  /**
   * Router does not implement Act(Object) method. A lot of Router action is delegated
   * to the RouterRecvThread
   * @param o Object to be instantiated and performed
   */
  public boolean Act(Object o) {
    // do nothing
    return true;
  }
  
  /**
   * check whether there's a reserved messages. If it's time to send
   * try to open connection and sent it to the receiver
   */
  public void run()   {
    while(true)   {
      try  {
	if(!_reservedBook.isEmpty())  {
	  sendReservedMessages();
	  sendOffLineAgentMessages();
	  sleep(_sleepPeriod);
	}
	else  {
	  // check every five seconds
	  sleep(_sleepPeriod);
	}
      } catch (Exception e)  {
	System.out.println(e.toString());
      }
    }
  }
  
  /**
   * Send reserved message if time meets
   */
  protected void sendReservedMessages()   {
    // get current time
    long time = System.currentTimeMillis();
    
    Enumeration e = _reservedBook.keys();
    while(e.hasMoreElements()) {
      Long reservedtime = (Long)e.nextElement();
      
      if(time > reservedtime.longValue())  {
	
	Reservation res = (Reservation)_reservedBook.get(reservedtime);
	String receiver = res.getReceiver();
	
	if(!_offLineList.contains(receiver))  {
	  RouterRecvThread rthread = (RouterRecvThread)_connections.getConnection(receiver);
	  // recv thread is running
	  if(rthread != null)   {
	    rthread.sendReservedMessages(res.getTimeString());
	    _reservedBook.remove(reservedtime);
	  }
	  else  {
	    
	    Address addr = res.getAddress();
	    
	    if((addr != null) && (addr.getPort() != -1))  {
	      try  {
		// reset the moving address target
		_addresses.addAddress(addr);
		RouterRecvThread thread =
		  (RouterRecvThread)createReceiverThread(addr.getID(),Thread.NORM_PRIORITY);
		thread.sendReservedMessages(res.getTimeString());
		_reservedBook.remove(reservedtime);
		
	      } catch (Exception ee)  {
		// reset time to try next time
		resetReservedTime(reservedtime.longValue(),res);
	      }
	    }
	    else  {
	      System.out.println("Can not send reserved message. Not a stand alone agent");
	      // reset time to try next time
	      resetReservedTime(reservedtime.longValue(),res);
	    }
	  }
	}	
	// agent is in the off line list
	else  {
	  resetReservedTime(reservedtime.longValue(),res);
	}	
      }
      else  {
	// do not need to check more. Other is definetely greater than
	// the current time.
	break;
      }
    }
  }
  
  /**
   * If the attempt to send reserved messages to the receiver, the Router will
   * reset time as one more hour from the original reserved time. This process will
   * be performed five times. If the receiver is still unreachable after five attempt,
   * the reserved messages will be canceled to be sent.
   * @param time Current setting time
   * @param res Reservation object to be delayed
   */
  protected void resetReservedTime(long time,Reservation res)  {
    // remove first. Then if it is not a maximum time, reset timer
    if(res.getTrialTimes() > _reserveTrialTimes)  {
      _reservedBook.remove(new Long(time));
      long newtime = time + _reserveCheckPeriod*1000*60; // unit in minutes
      res.incTimes();
      _reservedBook.put(new Long(newtime),res);
    }
  }
  
  
  /**
   * Send off line agent message if time meets
   */
  protected void sendOffLineAgentMessages()   {
    // get current time
    long time = System.currentTimeMillis();
    
    Enumeration e = _offLineList.elements();
    while(e.hasMoreElements()) {
      Reservation res = (Reservation)e.nextElement();
      long reservedtime = res.getTime();
      
      if(time > reservedtime)  {
	Address addr = res.getAddress();
	String agentid = addr.getID();
	
	try  {
	  // reset the moving address target
	  _addresses.addAddress(addr);
	  RouterRecvThread thread =
	    (RouterRecvThread)createReceiverThread(addr.getID(),Thread.NORM_PRIORITY);
	  _offLineList.remove(agentid);
	} catch (Exception ee)  {
	  // over maximum trial times. Send error message to all the message senders
	  // and unregister agent		
	  if(res.getTrialTimes() > _offLineTrialTimes)  {
	    unregisterAgent(addr);
	    _offLineList.remove(agentid);
	  }
	  else {		
	    time += _offLineTrialTimes*1000*60;
	    res.setTime(time);					
	    res.incTimes();
	  }	
	}
      }		
    }
  }	
  
  protected void unregisterAgent(Address addr)  {
    try  {
      RouterRecvThread client = new RouterRecvThread(_addresses.getAddress(getName()),addr,_connections,_messageBoxPath,_security,_offLineList);
      client.setEndWith(_endWith);
      client.setName(addr.getID());
      client.sendOffLineAgentErrorMessages();
      // note that it does not invoke the start() method
    }  catch (Exception e)  {
      System.out.println(e.toString() + " in send off line agent error messages");
    }
  }	
  
  public static void main(String args[])  {
    
    if(args.length != 1)  {
      System.out.println("Wrong arguments. Script file should be specified");
      System.exit(0);
    }
    
    try  {
      BufferedReader in = new BufferedReader(new FileReader(new File(args[0])));
      ReadOneLine input = new ReadOneLine(in);
      
      String name = input.getOneLine();
      String addrpath = input.getOneLine();
      String incomingpath = input.getOneLine();
      String passwordpath = null;
      String registryFilePath = null;
      String password = input.getOneLine();
      String registry = input.getOneLine();
      if(!password.equals("null"))  {
				passwordpath = password;
      }
      
      if(!registry.equals("null"))  {
      	registryFilePath = registry;
      }
      	
      int durationtime = Integer.valueOf(input.getOneLine()).intValue();
      int offlinemaxtimes = Integer.valueOf(input.getOneLine()).intValue();
      int offlinechecktime = Integer.valueOf(input.getOneLine()).intValue();
      int reservemaxtimes = Integer.valueOf(input.getOneLine()).intValue();
      int reservechecktime = Integer.valueOf(input.getOneLine()).intValue();
      int sleeptime = Integer.valueOf(input.getOneLine()).intValue();
      
      in.close();
      
      RouterAction action = new RouterAction(name,
					     addrpath,
					     incomingpath,
					     passwordpath,
					     registryFilePath,
					     durationtime,
					     offlinemaxtimes,
					     offlinechecktime,
					     reservemaxtimes,
					     reservechecktime,
					     sleeptime);
      action.start();
    }  catch (Exception e)  {
      System.out.println(e.toString());
    }
  }
}