package RouterLayer.Router;

import java.net.*;
import java.io.*;
import java.util.*;

import Abstract.*;
import BaseLayer.*;
import KQMLLayer.*;
import RouterLayer.util.*;
import RouterLayer.AgentClient.KQMLmail;

/**
* RouterRecvThread is a subclass of the BReceiverThread class. 
* RouterRecvThread action is:<BR>
* Run(while true)
*    <UL>Check message queue<BR>
*		<UL>If message queue is not empty<BR>
*			<UL>retrieve messages(not just one message. All the messages in the queue at once)
*				  send messages through the connection, save messages to the file, update idle time</UL>
*    Check socket buffer<BR>
*			<UL>if socket buffer is not empty<BR>
*				<UL>retrieve messages<BR>
*					<UL>for all the messages
*						parse the message to find out the receiver<BR>
*							<UL>if the receiver is the Router itself, perform the message<BR>
*									if the receiver is another agent<BR>
*										<UL>Check the ConnectionTable<BR>
*											<UL>if the receiver is connected<BR>
*														<UL>store the message to the receiver queue</UL>
*												  if the receiver is disconnected<BR>
*														<UL>Check the connection can be initiated<BR>
*							     							<UL>if the connection can be initiated<BR>
*																	<UL>Try to connect to the receiver<BR>
*																			<UL>if connection success<BR>
*    																			  <UL>store the message to the receiver queue</UL>
*																				  if connection failure<BR>
*									    										  <UL>save the message to the receiver file</UL></UL></UL>
*							     							if the connection can not be initiated<BR>
*																		<UL>save the message to the receiver file</UL></UL></UL></UL></UL></UL></UL></UL></UL></UL></UL>
* The RouterRecvThread will retain its own MessageBuffer, which does not related to the
* RouterAction. The Router utilities will be delegated to the RouterRecvThread.
* For message reservation, the RouterRecvThread will create Reservation object and
* add it to the ReservationTable, cretated and maintained by the RouterAction.
* For delete-message operation, the delete-messages can be buffered. 
* You can set the _maxDeleteSize to change the buffer size. Default is 1, this
* means that the message numbers will be updated every time for delete operation
* is performed. The message numbering should be identical to the RotuerClientAction.
* Therefore, if you set the delete-message buffer size, the RouterClientAction
* delete-message buffer size should be changed, too.
* @see RouterLayer.Router.RouterServerThread
* @see RouterLayer.AgentClient.RouterClientAgentAction
* @see RouterLayer.Router.RouterAction
*/

/* 
* Copyright (c) 1996,1997,1998 <A HREF="http://cdr.stanford.edu/">Center for Design Research</A>, Stanford University.
* All rights reserved.<p>
*
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You can obtain the GNU General Public License at
* http://www.gnu.org/copyleft/gpl.html or by writing to the Free
* Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
* 02111-1307, USA.
* 
* This software is bound by the terms and conditions listed in the
* attached <a href="../LICENSE">LICENSE file</A>.
*  @author <A HREF="http://cdr.stanford.edu/html/people/jeon-bio.html>Heecheol Jeon</A>
*  @version Java(tm) 1.1
*/ 

public class RouterRecvThread extends KQMLRecvThread {
  
  /**
   * The agent is not an registered user. The error message will be sent
   * when an agent want to ANS service, but the agent is not on the
   * registered user list
   */
  protected static final int NOTREGISTEREDUSER = 1000;
  
  /**
   * The receiver name is not specified in the KMQL message. Error message
   * will be sent
   */
  protected static final int NAMENOTSPECIFIED = 1002;	
  
  /**
   * File creation for the incoming file box failed
   */
  protected static final int CREATEFILEFAILED = 1003;	
  
  /**
   * Agent password is not passed during the Security process
   */
  protected static final int PASSWORDMISSING = 3;
  
  /**
   * Message method is not understandable
   */
  protected static final int NOTVALIDMESSAGEMETHOD = 7;
  
  /**
   * Receiver is not specified
   */
  protected static final int RECEIVERMISSING = 8;
  
  /**
   * Internal error
   */
  protected static final int INTERNALERROR = 100;
  
  /**
   * Unregister is not authorized
   */
  protected static final int UNREGISTERFAILED = 110;	
  
  /**
   * Parse error will be immediately returned
   */
  protected static final int PARSEERROR = 120;
  
  /**
   * Wrong format KQML message will be immediately returned
   */
  protected static final int WRONGKQMLFORMAT = 121;
  
  /**
    * Can not find email address for receiver 
    */
  protected final static int EMAILNOTSPECIFIED = 5005;	
    
	/**
	  * Message size is over MAX_MSG_LEN. Will be disconnected from the router
	  */    
	protected final static int MESSAGESIZETOOBIG = 123;
		  
  /**
   * Hashtable to record reserved messages. The key is time stamp and the field
   * value is reservation object
   */
  protected Hashtable _reservedBook;
  

  /**
   * Agent name array which contains the agent names those are off-line intentionally.
   * If a message receiver is in the off-line list, the RouterRecvThread will not send
   * the message to the receiver. The message will be saved to the file. If the receiver
   * is reconnect, the received messages will be sent.
   */
  protected OffLineAgentTable _offLineList;
  
  /**
   * Table, which contains the disconnected agent list. If the agent send 'disconnect-agent'
   * message to the router, the router will add the agent to this table. Until the agent
   * will reconnect, the router will not send message to the agent.
   */
  protected Vector _disconnectTable;
  
  /**
   * Message box file path
   * By giving messageBoxPath, I am assuming that the file extension as<BR>
   * incoming file - agentname.incoming<BR>
   * message file pointer file - agentname.fpt<BR>
   * reserved message file - agentname.reserved.timestamp. Time stamp is 12 digit numbers<BR>
   */
  protected String _messageBoxPath;
  
  /**
   * Vector to contains deleted message number. Delete operation will
   * be performed either when the receiver thread is disconnected
   * or number of delete-message is greater than _maxDeleteSize
   */
  protected int[] _deleteFPT;
  
  /**
   * Current message size in the MessageBuffer
   */
  protected int _currentFPTSize;
  
  /**
   * Current incoming file size
   */
  protected long _currentFileSize;
  
  /**
   * Current delete-message size
   */
  protected int _deleteFPTSize;
  
  /**
   * Maximun buffer size for deleted messages
   */
  protected int _maxDeleteSize;
  
  /**
   * Cache file pointer of the messages. 
   */
  protected RandomAccessFile _msgFPTStream;
  
  /**
   * Incoming file
   */
  protected RandomAccessFile _incomingFile;

  
  /**
   * Set the ReservationTable
   * @param reservedBook ReservationTable from the RouterAction
   */
  public void setReservedBook(Hashtable reservedBook)  {
    _reservedBook = reservedBook;
  }	
  
  /**
   * Set the off line agent name list. The off-line means that an agent
   * intentionally wants not to receive message before it reconnect to the Router.
   * @param offlinelist Off line agent name list. 
   */
  public void setOffLineList(OffLineAgentTable offlinelist)  {
    _offLineList = offlinelist;
  }
  
  /**
   * Set the disconnect table
   * @param disconnectlist Disconnected agent list
   */
  public void setDisconnectTable(Vector disconnectlist)  {
    _disconnectTable = disconnectlist;
  }
  
  /**
   * Default Constructor.
   */
  public RouterRecvThread()  {
    super();
  }
  
  /**
   * Constructor with the Agent name. If you use this constructor, you should
   * use set methods to set data members
   * @param name Agent name
   */	
  public RouterRecvThread(String name)  {
    super(name);
  }
  
  /**
   *  Constructor, to be called from the serverthread after accepting
   *  a connection. 
   * @param client_scoket The socket for the remote client;
   * @param priority The priority of the thread.
   * @param myaddress My address to be sent to the Router to check security
   * @param connections The connection table is passed as an argument.
   * @param queue The Message Queue. Normally null. It will be created by this if null
   * @param security The RouterSecurity
   * @param messageboxpath The Message incoming box directory
   * @param reservedBook The reservation table, passed from the RouterAction
   * @param offlinelist The off line agent name list.
   * @param disconnectlist The disconnected agent name list.
   */
  public RouterRecvThread(Socket client_socket,
			  int priority, 
			  Address myaddress,
			  ConnectionTable connections,	
			  MessageBuffer queue,
			  Security security,
			  String messageboxpath,
			  Hashtable reservedBook,
			  OffLineAgentTable offlinelist,
			  Vector disconnectlist) 
    throws ConnectionException    {
      super(client_socket,priority,myaddress,connections,queue,security);
      if(_queue == null)  _queue = new BMessageBuffer();
      _messageBoxPath = messageboxpath;
      _reservedBook = reservedBook;
      _offLineList = offlinelist;
      _disconnectTable = disconnectlist;
      _maxDeleteSize = 1;
      _role = true;
  }
  
  
  
  /**
   *  This constructor is called when an agent want to initialize a connection
   *  to the other agents, using Address
   * @param serveraddress The Router address
   * @param priority The priority of the thread.
   * @param myaddress My address to be sent to the Router to check security
   * @param connections The connection table is passed as an argument.
   * @param queue The Message Queue. Normally null. It will be created by this if null
   * @param security The RouterSecurity
   * @param messageboxpath The Message incoming box directory
   * @param reservedBook The reservation table, passed from the RouterAction
   * @param offlinelist The off line agent name list.
   * @param disconnectlist The disconnected agent name list.
   * @param tg The thread group to which this receiver thread will be owned to
   */
  public RouterRecvThread(Address serveraddress,
			  int priority, 
			  Address myaddress,
			  ConnectionTable connections,
			  MessageBuffer queue,
			  Security security,
			  String messageboxpath,
			  Hashtable reservedBook,
			  OffLineAgentTable offlinelist,
			  Vector disconnectlist,
			  ThreadGroup tg)  
    throws ConnectionException   {
      super(serveraddress,
	    priority,
	    myaddress,
	    connections,
	    queue,
	    security,
	    tg);
      if(_queue == null)  _queue = new BMessageBuffer();			
      _messageBoxPath = messageboxpath;
      _reservedBook = reservedBook;
      _offLineList = offlinelist;
      _disconnectTable = disconnectlist;
      _maxDeleteSize = 1;
      _role = false;
  }
  
  
  /**
   * Router receiver thread to send off-line agent's messages back to the senders with error.
   *	@param serveraddress Router address
   * @param myaddress Router address to be sent to the Router to check security
   * @param connections The connection table is passed as an argument.
   * @param security The RouterSecurity
   * @param messageboxpath The Message incoming box directory
   * @param offlinelist The off line agent name list.
   */
  public RouterRecvThread(Address serveraddress,
			  Address myaddress,
			  ConnectionTable connections,
			  String messageboxpath,
			  Security security,
			  OffLineAgentTable offlinelist)  {
    super(myaddress.getID());		
    _serverAddress = serveraddress;
    _myaddress = myaddress;
    _connections = connections;
    _messageBoxPath = messageboxpath;
    _offLineList = offlinelist;
    _role = false;
    try {
      setSecurity(security);
    }  catch (Exception e) {}
  }
  
  /**
   * Send queued messages to the agent. Update off-line/disconnect table
   */ 																		
  protected void initialize()  {
    openFiles();
    
    String name = getName();
    if(_offLineList.containsKey(name))  {
      _offLineList.remove(name);
    }		
    
    if(_disconnectTable.contains(name))  {
      _disconnectTable.removeElement(name);
    }	
  }
  
  /**
   * Send messages to the receiver by retriving file and open necessary files.
   * This method is invoked after security check has been passed. 
   */
  public synchronized void openFiles()  {
    try  {
      try  {
	
	//retrieve current message size		
	DataInputStream in = 
	  new DataInputStream(new FileInputStream(_messageBoxPath + getName() + ".fpt"));
	_currentFPTSize = in.readInt();
	
	File file = new File(_messageBoxPath + getName() + ".incoming");
	BufferedReader tmp = new BufferedReader(new FileReader(file));
	
	// this will be always '0'
	int msgstart = 0;
	
	char[] buffer = new char[MAX_MSG_SIZE];
	for(int msgno = 0; msgno < _currentFPTSize; msgno++)  {
	  int msgend = in.readInt();
	  int readlen = tmp.read(buffer,0,msgend - msgstart);
	  _out.write(buffer,0,readlen);
	  _out.flush();
	  msgstart = msgend;
	  _currentFileSize += readlen;
	}
	tmp.close();
	in.close();	
	
      }  catch (Exception e)  {
	System.out.println(e.toString());
	System.out.println("New file will be generated, error thrown");
      }		
      
      _deleteFPT = new int[_maxDeleteSize];
      _deleteFPTSize = 0;
      
      _incomingFile = new RandomAccessFile(_messageBoxPath + getName() + ".incoming","rw");
      // go to the end of the sending message
      _incomingFile.skipBytes((int)_currentFileSize);
      _msgFPTStream = new RandomAccessFile(_messageBoxPath + getName() + ".fpt","rw");		
      _msgFPTStream.skipBytes((int)_msgFPTStream.length());
    } catch (Exception e)  {
      System.out.println("Open files failed due to " + e.toString());
    }		
  }	
  
  
  /**
   * The run() method will performed the Security check at first. If the security
   * check has been passed, the Security will update the ConnectionTable. Then
   * the run() method will send the incoming messages and start to receive messages
   */ 
  public void run()   {
    if(_security != null)  {
      // server
      if(_role == true)  {
				_security.processServerLogin(this);
      }
    }			
    
    initialize();
    
    int maxMsgSize = MAX_MSG_SIZE - HEADER_LEN;
    char[] buffer = new char[maxMsgSize];
    int readStart = 0;
    
    while(true) {
      try {
	
	if(_in.ready()) {
		// message size is too big. End condition
		if(readStart >= (maxMsgSize - 1))  {
			// Message size is too big. End condition
			sendErrorMessage(MESSAGESIZETOOBIG,null);
			// clear queue
			receiveMessage();
			endConn();
		}	
		
		
	  // read full unblocked messages
	  int readLen = readBuffer(buffer,readStart,maxMsgSize - readStart);	

	  // if something is read 									  
	  if(readLen != -1)  {
	    int startPosition = 0;
	    int sum = 0;
	    while(true)  {
	      int unitLen = findEndLocation(buffer,startPosition,readStart + readLen);
	      if(unitLen != -1)  {
	      	// if unitLen == 0, it is bogus message, and ignore
	      	if(unitLen != 0)  {
						String msg = new String(buffer,startPosition,unitLen);
						// parse receiver and store to the receiver's queue
						// if the receiver is dead, create RouterRecvThread
						// if possible and store to the receiver's queue
						forwardMessage(msg);
					  resetTime();
					}	
					startPosition = startPosition + unitLen + 1;	
					sum = sum + unitLen + 1;
	      }	
	      else  {
					readStart = readStart + readLen - sum;

					// shift to the left
					for(int i=0; i< readStart; i++)  {
		  			buffer[i] = buffer[i + sum];
					}	
					break;
	      }	
	    }	
	  }		
	}
	else  {
	  if(!_queue.isEmpty())  {
	    // send received message to the connection and save to the
	    // incoming file
	    receiveMessage();
	  }
	  else  {	
	    // message check for every seconds
	    sleep(1000);
	    checkTime();
	  }	
	}		
      } catch (Exception ee)  {
	System.out.println("IOException " + getName());		
	endConn();
      }	
    }	
  }
  
  /**
   * Override BRecvThread's checkTime method to add the agent name to the off line agent list
   */
  protected void checkTime()  {
    if(_durationTime == -1) return;
    long ctimevalue = System.currentTimeMillis();
    if(((ctimevalue - _lastTime)/60000) > _durationTime)  {
      System.out.println("Timeout " + getName());   	
      _offLineList.addAgent(_security.getAddressTable().getAddress(getName()));
      endConn();
    }
  }		
  
  /**
   * New messages are received from the socket. The RouterRecvThread parse the message
   * to find out the receiver. If the receiver is the Router itself, it will provide
   * necessary services. If the receiver is another agent, it will pass the message
   * to other agent after adding time stamp. For Router service, refer to 
   * <A HREF="RouterLayer.Router.RouterAction">RouterAction</A> class.
   * @param msg String, received from the buffer
   */
  protected void forwardMessage(String msg)   {
    KQMLmessage kqml = null;
    try  {
      kqml = new KQMLmessage(msg);
    }  catch (ParseException p) {
      sendErrorMessage(PARSEERROR,msg);
      return;
    }
    
    String receiver = kqml.getValue("receiver");
    
    if(receiver == null)  {
      sendErrorMessage(RECEIVERMISSING,kqml.getSendString());
      return;
    }
    
    
    try {
      // if the receiver is the Router itself
      if(receiver.equals(_myaddress.getID()))  {
	processDefaultProtocol(kqml);
      }
      else  {
	// find time and add
	this.sendMessage(receiver,msg);
      }					
    } catch (Exception e)  {
      System.out.println(e.toString());
    }		
  }						
	
  /**
   * Processing default protocols : disconnect-agent, reserve-message, list-agent, request-address,
   * and unregister-agent. You need to extend this method if your router has extended capability
   * other than message routing(e.g. to support another internet protocol such as SMTP, FTP,etc.)
   * @param kqml Received KQML message
   */
  protected synchronized void processDefaultProtocol(KQMLmessage kqml)   {
    // currently, two performatives can be performed by this object
    String perf = kqml.getValue("performative");
    if(perf.equals("disconnect-agent"))  {
      // close all the opened files
      if(_disconnectTable.contains(getName()))  {
	_disconnectTable.addElement(getName());
      }	
      endConn();
    }
    else
      if(perf.equals("reserve-message"))  {
	reserveMessages(kqml);
      }	
      else 
	if(perf.equals("delete-message"))  {
	  String no = kqml.getValue("content");
	  _deleteFPT[_deleteFPTSize] = (Integer.valueOf(no)).intValue();
	  _deleteFPTSize++;
	  if(_deleteFPTSize == _maxDeleteSize)   {
	    deleteMessages();
	  }
	}
	else
	  if(perf.equals("list-agent"))  {
	    String reply = getMsgHeader();
	    reply = reply + "(registered-agent :sender ";
	    reply = reply + _myaddress.getID() + " :receiver " + getName();
	    reply = reply + " :content (";
	    Enumeration e = _security.getAddressTable().elements();
	    while(e.hasMoreElements())  {
	      Address addr = (Address)e.nextElement();
	      String id = addr.getID();
	      reply = reply + "(" + id + " " + addr.getHost() + " " + addr.getPort();
	      if(_connections.containsKey(id))  {
		reply = reply + " connected) ";
	      }	
	      else	{
		reply = reply + " disconnected) ";
	      }	
	    }
	    reply = reply + "))" + _endWith;
	    _queue.addElement(reply);
	  }		
	  else
	    if(perf.equals("request-address"))  {
	      String id = kqml.getValue("content");
	      if(id == null)  {
		sendErrorMessage(NAMENOTSPECIFIED,null);
	      }
	      else  {	
		Address addr = _security.getAddressTable().getAddress(id);
		if(addr == null)  {
		  sendErrorMessage(NOTREGISTEREDUSER,id);
		}
		else  {
		  String reply = getMsgHeader();
		  reply = reply + "(agent-address :sender ";
		  reply = reply + _myaddress.getID() + " :receiver " + getName();
		  reply = reply + " :name " + addr.getID();
		  reply = reply + " :host " + addr.getHost() + " :port ";
		  reply = reply + addr.getPort() + " :message-method " + addr.getType();
		  String desc = addr.getDescription();
		  if(desc != null)  {
		    reply = reply + " :description " + desc;
		  }
		  String replyWith = kqml.getValue("reply-with");
		  if((replyWith != null) && (!replyWith.equals("")))  {
		  	reply = reply + " :in-reply-to " + replyWith;
		  }
		  	
		  reply = reply + ")" + _endWith;
		  _queue.addElement(reply);
		}
	      }
	    }
	    else
	      if(perf.equals("unregister-agent"))  {
		String[] s = new String[2];
		s[0] = kqml.getValue("sender");
		s[1] = kqml.getValue("password");
		if(((RouterSecurity)_security).unregisterAgent(s))  {
		  endConn();
		}
		else {
		  sendErrorMessage(UNREGISTERFAILED,"");
		}
	      } 		 				
    
  }
  
  /**
   * Send message to the intended receiver. If the receiver is on-line, add message to the
   * receiver message buffer. If the receiver is off-line, try to connect to the receiver
   * If connected, add message to the receiver buffer. If fails, save message to the receiver
   * incoming file.
   * @param receiver The receiver agent
   * @param stampmsg The stamped message
   */
  protected synchronized void sendMessage(String receiver,String msg)  {
    
    String stampmsg = "N ";
    stampmsg = stampmsg + System.currentTimeMillis() + " " + msg + _endWith;
    
    ReceiverThread rthread = _connections.getConnection(receiver);
    
    // if receiver thread is running, add message to the receiver queue
    if(rthread != null)   {
      rthread.getMessageQueue().addMessage(stampmsg);
    }
    // receiver thread is dead. Create if possible
    else  {
      String[] arg = new String[1];
      arg[0] = receiver;
      if(!_security.isValidAgent(arg))   {
	sendErrorMessage(NOTREGISTEREDUSER,receiver);
      }
      else  {
	// if the agent is not in the disconnectTable, try to initiate message
	if(!_disconnectTable.contains(receiver))  {
	  Address serveraddr = _security.getAddressTable().getAddress(receiver);
	  if((serveraddr != null) && (serveraddr.getPort() != -1))  {
	    try  {
	      // if connection is successful, it will automatically remove itself from the
	      // offLineList.
	      RouterRecvThread client = new RouterRecvThread(serveraddr,getPriority(),_myaddress,
							     _connections,new BMessageBuffer(),_security,_messageBoxPath,_reservedBook,_offLineList,_disconnectTable,getThreadGroup());
	      client.setEndWith(_endWith);
	      client.setDurationTime(_durationTime);
	      client.start();		
	      client.getMessageQueue().addMessage(stampmsg);
	      return;
	    }  catch (Exception e)  {
	      // connection failure. create off-line agent and add it to the off-line list
	      if(!_offLineList.containsKey(receiver))  {
		_offLineList.addAgent(_security.getAddressTable().getAddress(receiver));
	      }		
	      System.out.println("Receiver thread for " + receiver + " failed");
	    }
	  }
	}		
	saveMessageToFile(receiver,stampmsg);
      }
    }		
    
  }
  
  /**
   * Message reservation on a specific time
   * Generate reserved message file. Extension is reserved time.
   * To evade confusion, delete messages first if exist.
   * The received KQML will look like<BR>
   * (reserve-message :sender A :receiver A :time 123456789012 
   * :content (2 5 7) :address (receiver-address :name B
   * :host cdr.stanford.edu :port 1234 :message-method RouterRecvThread 
   * :description (agent-info :password xyz))<BR>
   * The delete slot is optional but it would be better to set as 'y'
   * to delete message after you reserve<BR>
   * The address slot is optional. If you want to receive at a specific
   * location other than where you are working now, you will need to 
   * set the location. The description slot is optional. However, if
   * the Server side need password check, you should provide password
   * to the Router to connect safely.
   * @param kqml KQMLmessage, which contains the message reservation info.
   */
  protected synchronized void reserveMessages(KQMLmessage kqml)   {
    try  {
      String receiver;
      
      String time = kqml.getValue("time");
      String content = kqml.getValue("content");
      
      if((time == null) || (content == null)) {
	System.out.println("No reserved time/content. Reservation aborted");
	return;
      }		
			
      String address = kqml.getValue("address");
      Address newaddress = null;
      if(address != null)   {
	KQMLmessage addrkqml = new KQMLmessage(address);
	// do not check performative, 'receiver-address' preferable thouth
	receiver = addrkqml.getValue("name");
	String host = addrkqml.getValue("host");
	String port = addrkqml.getValue("port");
	String type = addrkqml.getValue("message-method");
	String description = addrkqml.getValue("description");
	newaddress = new Address(receiver,host,(Integer.valueOf(port)).intValue(),type,description);		
      }
      else  {
	// if the receiver address is not set, assume the sender will be a receiver
	receiver = kqml.getValue("sender");
      }		
      
      if(_incomingFile == null)  {
	_incomingFile = new RandomAccessFile(new File(_messageBoxPath + getName() + ".incoming"),"rw");
      }	
      
      BufferedOutputStream out =
	new BufferedOutputStream(new FileOutputStream(new File(_messageBoxPath + receiver + ".rsv." + time)));
      
      _msgFPTStream.seek(0);
      int msgsize = _msgFPTStream.readInt();
      
      int[] a = new int[msgsize + 1];
      // always starts from 0
      a[0] = 0;
      int i=1;			
      for(; i< msgsize + 1; i++)  {
	a[i] = _msgFPTStream.readInt();
      }	
      
      // reset to the end
      _msgFPTStream.seek((int)_msgFPTStream.length());
      
      byte[] buffer = new byte[MAX_MSG_SIZE];		
      
      Vector v = new takeOffList(content);
      Enumeration e = v.elements();
      while(e.hasMoreElements())  {
	int index = (Integer.valueOf((String)e.nextElement())).intValue();
	if(a[index] > (int)_currentFileSize)  {
	  System.out.println("Message requested to reserve does not exist");
	}
	else  {	
	  _incomingFile.seek(a[index]);			
	  
	  int readlen = -1;
	  readlen = _incomingFile.read(buffer,0,a[index + 1] - a[index]);
	  if(readlen != -1)  {
	    out.write(buffer,0,readlen);
	  }
	  else  {
	    System.out.println("File IO  Error while reserving messages");
	  }
	}
      }
      
      out.close();
      
      _incomingFile.seek(a[i-1]);
      
      Reservation res = new Reservation(receiver,newaddress,time);
      _reservedBook.put(new Long(time),res);
      
      v.removeAllElements();
										
    } catch (Exception e)  {
      System.out.println("Reserved Message Error " + e.toString());
    } 	
  }			
  
  /**
   * send reserved messages, invoked from the RouterAction
   * The reasoning is that when the RouterAction want to send reserved
   * message and the receiver thread is running, it does not need to
   * open a new connection to send. The messages will be saved to 
   * my message queue rather than sending directly, since if you
   * send the message directly, you will be in trouble to update the
   * incoming file, FPT file - the agent will received the messages
   * which is not registered to the Router and it will cause confusion.
   * No clean up after sending the reserved file. 
   * @param time  String, reserved time
   */
  public synchronized void sendReservedMessages(String time)   {
    try  {
      BufferedReader in = 
	new BufferedReader(new FileReader(new File(_messageBoxPath + getName() + ".rsv." + time)));
      char[] buffer = new char[MAX_MSG_SIZE];
      int readStart = 0;
      while(true)  {
	int readLen = in.read(buffer,0,MAX_MSG_SIZE);
	if(readLen == -1)  break;
	else  {
	  int startPosition = 0;
	  while(true)  {
	    int unitLen = findEndLocation(buffer,startPosition,readStart + readLen);
	    if(unitLen != -1)  {
	      // Note removing 'N' tag, include endWith char
	      String msg = new String(buffer,startPosition+1,unitLen - 1);
	      msg = "R" + msg + _endWith;
	      _queue.addMessage(msg);
	      startPosition = startPosition + unitLen + 1;	
	    }
	    else  {
	      readStart = readStart + readLen - startPosition;
	      // shift to the left
	      for(int i=0; i< readStart; i++)  {
		buffer[i] = buffer[i + startPosition];
	      }	
	      break;
	    }	
	  }	
	}
      }
      
      in.close();
    } catch (Exception e)  {
      System.out.println("Can not send reserved messages due to " + e.toString());
    }		
  }			
  
  
		
  /**
   * This method is invoked under the assumption that the receiver is off-line and
   * the Router can not connect to the receiver
   * @param receiver The receiver agent
   * @param msg Time stamped message
   */	
  protected synchronized void saveMessageToFile(String receiver,String msg)   {
    try  {
      
      RandomAccessFile fpt = new RandomAccessFile(_messageBoxPath + receiver + ".fpt","rw");
      int currentFileSize = 0, nofmsg = 0, currentIncomeSize = 0;
      
      currentFileSize = (int)fpt.length();
      
      // brand new created file.
      if(currentFileSize == 0) {
	fpt.writeInt(0);
	currentFileSize = (int)fpt.length();
      }
      else {
	nofmsg = fpt.readInt();
      }
      
      // current incoming file size
      fpt.seek(currentFileSize - 4);
      currentIncomeSize = fpt.readInt();
      
      RandomAccessFile out = 
	new RandomAccessFile(_messageBoxPath + receiver + ".incoming","rw");
      out.seek(currentIncomeSize);	
      out.writeBytes(msg);
      out.close();
      
      fpt.seek(currentFileSize);
      fpt.writeInt(currentIncomeSize + msg.length());
      
      fpt.seek(0);
      fpt.writeInt(++nofmsg);
      
      fpt.close();
		}  catch (IOException ee)  {
		  System.out.println("Message save to " + receiver + " Error " + ee.toString());
		}	
  }			
  
  /**
   * Message received to the queue. send message and save it to the file	
   * Time stamp is already attatched by other receiver threads
   * Should be synchronized, not to add messages to the queue
   * while writing. 
   */
  protected synchronized void receiveMessage()  {
    
    // IO error while sending messages
    boolean error = false;
    
    while(!_queue.isEmpty())  {
      String msg = (String)_queue.getMessage();
      saveMessageToIncomingFile(msg);
System.out.println("Message saved to file");		
			// try to send as many as possible. If connection is broken, thread will stop
			// automatically but all incoming messages are already in the file system.
    	// write to socket
    	if(!error)  {
    		try  {
     			writeMsg(msg);
    		}  catch (ConnectionException e)  {
System.out.println("Error. write msg");    			
    			error = true;
    		}	
    	}
    }
    
    if(error)  {
    	endConn();
    }		
    		
  }		
  
  protected void saveMessageToIncomingFile(String msg)  {
		try  {
			// save pointer
			_msgFPTStream.seek(0);
	
			// add number of messages
			_msgFPTStream.writeInt(++_currentFPTSize);
	
			// reset to the end
			_msgFPTStream.seek(_msgFPTStream.length());
			_currentFileSize += msg.length();
				
			// add new message pointer
			_msgFPTStream.writeInt((int)_currentFileSize);
	
			// save message to incoming file
			_incomingFile.writeBytes(msg);
		} catch (IOException e)  {
			System.out.println(e.toString());
			// critical error. Close socket and stop
			closeSocket();
		}		
			
	}	
						
	/**
	* Does not necessarily be synchronized. All the other threads
	* knows that this thread is running, therefore they will try to
	* add received messages to this thread to the message queue.
	* This method will be invoked when the agent wants to disconnect or
	* the _maxDeleteSize is reached.
	*/				
	protected void deleteMessages()   {
		try  {
			QSort qsort = new QSort();
//System.out.println("DeleteFPTSize is : " + _deleteFPTSize);	
			// sort needs size - 1		
			qsort.sort(_deleteFPT,0,_deleteFPTSize - 1);
//for(int k = 0; k < _deleteFPTSize ; k++)  {
//	System.out.print(_deleteFPT[k] + "\t");
//}	
			if(_msgFPTStream != null)  {
				_msgFPTStream.close();
				_msgFPTStream = null;
			}
			
			if(_msgFPTStream != null)  {
				_msgFPTStream.close();
			}
				
			DataInputStream in = 
			  new DataInputStream(new FileInputStream(_messageBoxPath + getName() + ".fpt"));
			  
			int nofmsg = in.readInt();
//System.out.println("Message size " + nofmsg);		
			int msgFPT[] = new int[nofmsg + 1];
			int i = 1;
			// message always start from 0
			msgFPT[0] = 0;
			for(; i< nofmsg + 1; i++)  {
				msgFPT[i] = in.readInt();
//System.out.println(msgFPT[i]);				
			}	
			in.close();

			/////////////////////////////////////////////  
			// reset current message number. Prevent error.  
			_currentFPTSize = nofmsg - _deleteFPTSize; 
			if(_currentFPTSize < 0)  {
				return;
			}	

			// reopen to write			
			DataOutputStream out = 
			  new DataOutputStream(new FileOutputStream(_messageBoxPath + getName() + ".fpt"));
			  
			 
			out.writeInt(_currentFPTSize);
//System.out.println("Remaining message size " + _currentFPTSize);		
			// set end as current incoming file size
			
			if(_incomingFile != null)  {
				_incomingFile.close();
			}	
			
			BufferedReader bin = 
			  new BufferedReader(new FileReader(new File(_messageBoxPath + getName() + ".incoming")));
			BufferedWriter bout =
			  new BufferedWriter(new FileWriter(new File(_messageBoxPath + getName() + ".tmpincoming")));  

			// maximum unit message size  
			char[] buffer = new char[MAX_MSG_SIZE];  
			int newMsgPoint = 0;

			// _deleteFPT[] index. Need to prevent the duplicated index	
			i = 0;
			
			for(int j = 0; j<_deleteFPTSize;)   {
				// _deleteFPT[i] is greater than the i(current index). Save.
				while((_deleteFPTSize > j) && (_deleteFPT[j] > i))	{
//System.out.println("deletefpt:" + _deleteFPT[j] + "I:" + i);				
					int nofchar = bin.read(buffer,0,msgFPT[i+1] - msgFPT[i]);
					if(nofchar != -1)  {
						// message writing
						bout.write(buffer,0,nofchar);
						
						newMsgPoint += nofchar;
						
						// FPT update
						out.writeInt(newMsgPoint);
						
						i++;
					}
					else {
						// critical error
						System.out.println("Message File Output Error");
						bin.close();
						bout.close();
						out.close();
					}				
				}

				while((j < _deleteFPTSize) && (_deleteFPT[j] == i) && (i < nofmsg)) {
//System.out.println("Same " + _deleteFPT[j] + " " + i);				
					bin.skip(msgFPT[i+1] - msgFPT[i]);
					i++;
					j++;
				}
				
				// remove the same elements in _deleteFPT, may happen
				while((j < _deleteFPTSize) && (i > _deleteFPT[j])) {
//System.out.println("Greater " + _deleteFPT[j] + " " + i);				
					j++;
				}	
			}				
			
			// if reset
			for(; i < nofmsg; i++)   {
					int nofchar = bin.read(buffer,0,msgFPT[i+1] - msgFPT[i]);
					if(nofchar != -1)  {
						// message writing
						bout.write(buffer,0,nofchar);

						newMsgPoint += nofchar;
						
						// FPT update
						out.writeInt(newMsgPoint);
						
					}
					else {
						// critical error
						System.out.println("Message File Output Error");
						bin.close();
						bout.close();
						out.close();
					}				
						
			}					
			
			/////////////////////////////////////////
			// Reset currentfile size
			_currentFileSize = newMsgPoint;

			out.close();
			bin.close();
			bout.close();
			
			// copy tmp file to the original incoming file
			BufferedReader ain = 
			  new BufferedReader(new FileReader(new File(_messageBoxPath + getName() + ".tmpincoming")));
			BufferedWriter aout =
			  new BufferedWriter(new FileWriter(new File(_messageBoxPath + getName() + ".incoming")));  
			
			int readlen = 0;
			// newMsgPoint is the file size
			while(readlen < newMsgPoint)  {
				int len = ain.read(buffer,0,MAX_MSG_SIZE);
				aout.write(buffer,0,len);
				readlen += len;
			}			
			
			ain.close();
			aout.close();
			
			////////////////////////////////////////////
			// reset delete message number
			_deleteFPTSize = 0;
			_incomingFile = new RandomAccessFile(_messageBoxPath + getName() + ".incoming","rw");
			_incomingFile.skipBytes((int)_incomingFile.length());
			_msgFPTStream = new RandomAccessFile(_messageBoxPath + getName() + ".fpt","rw");		
			_msgFPTStream.skipBytes((int)_msgFPTStream.length());
//System.out.println("Delete Success");			
		} catch (Exception e)  {
			System.out.println(e.toString());
		}
	}				
				
	public void sendOffLineAgentErrorMessages()  {
		try  {
			String routerid = _serverAddress.getID();
			String myid = _myaddress.getID();
			//retrieve current message size		
			DataInputStream in = 
				new DataInputStream(new FileInputStream(_messageBoxPath + getName() + ".fpt"));
			_currentFPTSize = in.readInt();
			File file = new File(_messageBoxPath + getName() + ".incoming");
			BufferedReader tmp = new BufferedReader(new FileReader(file));
		
			// this will be always '0'
			int msgstart = 0;
								
			char[] buffer = new char[MAX_MSG_SIZE];
			for(int msgno = 0; msgno < _currentFPTSize; msgno++)  {
				int msgend = in.readInt();
				int readlen = tmp.read(buffer,0,msgend - msgstart);
				// eliminate _endWith
				String msg = new String(buffer,0,readlen - 1);
				KQMLmail kqmlmail = new KQMLmail(msg,0);
				KQMLmessage kqml = kqmlmail.getKQMLmessage();
				String sender = kqml.getValue("sender");
				if((sender != null) &&
					 (!sender.equals(routerid)) &&
					 (!sender.equals(myid)))  	 {
					String error = "(message-delivery-error :sender ";
					error = error + routerid + " :receiver " + sender;
					error = error + " :content (" + kqml.getSendString() + ")";
					sendMessage(sender,error);
				}	
				msgstart = msgend;
			}	
			tmp.close();
			in.close();	

	 		String[] s = new String[2];
	 		s[0] = myid;
	 		String desc = _myaddress.getDescription();
	 		KQMLmessage pass = new KQMLmessage(desc);
	 		String password = pass.getValue("password");
	 		if(password == null)  {
		 		password = "";
	 		}
	 		
	 		s[1] = password;
			if(!((RouterSecurity)_security).unregisterAgent(s))  {
				System.out.println("Unregister for the agent " + myid + " failed");
			}		

		}  catch (Exception e)  {
			System.out.println(e.toString());
			System.out.println("send off line agent messages error");
		}		

		
	}		
		
						
	/**
	* The thread just writes on the connection for the remote side. 
  * @param writebuffer: the message as a normal ASCII string.
  * @exception Abstract.ConnectionException If IOError occurs, IOException will be thrown
  */
  public synchronized void writeMsg(String writebuffer)  
  throws ConnectionException  {
  	try  {
  		if(writebuffer.length() == 0)  {
  			return;
  		}
  			
  		if(writebuffer.charAt(writebuffer.length() - 1) != _endWith) {
    		_out.write(writebuffer,0,writebuffer.length());
    	}
    	else  {
    		_out.write(writebuffer,0,writebuffer.length() - 1);
    	}		
			_out.flush();
			_out.write(_endWith);
			_out.flush();
			resetTime();
		} catch (Exception e)  {
			throw new ConnectionException(e.toString());
		}	
  }

  
  
  /**
	* If there are messages to be deleted, deleteMessages() will be invoked
	* Send 'disconnect' message to the agent and Close all the connections.
	*/
  public synchronized void endConn()   {
			cleanUp();
	
    	// role is client. Send logout message
    	if(_role == false)  {
				processLogoutMessage();
			}	
			System.out.println(getName() + " disconnected");
			
			closeSocket();
 }

	/**
	* Clean up the files and remove connection from the connection table
	*/
	protected void cleanUp()  {
			// update message files first. 			
			if(_deleteFPTSize > 0)  {
				try  {
					deleteMessages();
				} catch (Exception eee)  {
					System.out.println(getName() + " can not delete messages due to " + eee.toString());
				}		
			}	
			
			// Save messages in the current queue.
			// If write socket error, this process is necessary	
			// _queue may be null in the case of RegistrarRecvThread	
			if(_queue != null)  {	
				try  {
					while(!_queue.isEmpty())  {
						String msg = (String)_queue.getMessage();
						saveMessageToIncomingFile(msg);
					}
				} catch (Exception er)  {
					System.out.println(er.toString());
				}		
			}	
			
    	if(_incomingFile != null)  {
    		try  {
    			_incomingFile.close();
    		} catch (IOException e1)  {
    			System.out.println("IncomingFile Close Fail: " + getName());
    		}		
    	}
    	
    	if(_msgFPTStream != null)  {
    		try  {
    			_msgFPTStream.close();
    		} catch (IOException e2)  {
    			System.out.println("FPTFile Close Fail: " + getName());
    		}		
    	}		
			
    	if(_connections != null)  {
//System.out.println("Disconnected " + getName());    		
    		if(_connections.isLiving(getName()))  {
    			_connections.removeConnection(getName());
    		}	
    	}	
	}	
	
	/**
	* Close socket and stop this thread
	*/
	protected void closeSocket()  {
			try  {
    		if(_in != null)  {
    			_in.close();
    		}
    	} catch (IOException e)  {
    		System.out.println(e.toString());
    	}		
    		
			try  {
    	
	   		if(_out != null)  {
  	  		_out.close();
    		}
    	} catch (IOException e2)  {
    		System.out.println(e2.toString());
    	}		

    		

			try  {	    
	    	if(_client != null)  {	
    			_client.close();
    		}	
    	} catch (IOException e5)  {
    		System.out.println(e5.toString());
    	}		
     	
	    stop();
  	  try  {
    		join();
    	} catch (InterruptedException e3)  {}	
	}
	
  /**
   * Send error message to the agent is error occurs.
   * @param code Error code
   * @param content Error content
   */
  public void sendErrorMessage(int code,String content)  {
    String errormsg = getMsgHeader();
    
    errormsg = errormsg + "(error :sender ";
    errormsg = errormsg + _myaddress.getID() + " :receiver " + getName();
    
    switch(code)  {
    case NOTREGISTEREDUSER:
      errormsg = errormsg + " :content (not-registered-user " + content;
      break;
    case PASSWORDMISSING:
      errormsg = errormsg + " :content (password-missing " + content;
      break;
    case NOTVALIDMESSAGEMETHOD:
      errormsg = errormsg + " :content (message-method-not-supported " + content;
      break;
    case INTERNALERROR:
    case CREATEFILEFAILED:
      errormsg = errormsg + " :content (sorry-internal-error " + content;
      break;
    case UNREGISTERFAILED:
      errormsg = errormsg + " :content (unregister-failed " + content;	
      break;
    case RECEIVERMISSING:
      errormsg = errormsg + " :content (receiver-missing " + content;	
      break;
		case EMAILNOTSPECIFIED:
			errormsg = errormsg + " :content (email address not spefied for " + content;	
      break;
		case MESSAGESIZETOOBIG:
			errormsg = errormsg + " :content (message size is too big. Will be disconnected";
			break;      
    case WRONGKQMLFORMAT:
      errormsg = errormsg + " :content (wrong-format-message";	
    }
    
    errormsg = errormsg + "))" + _endWith;
    _queue.addElement(errormsg);
  }		
  
  /**
   * Return message header, 'N 86472389	' where 86472389 represent current time
   */
  protected String getMsgHeader()   {
    String reply = "N ";
    reply = reply + System.currentTimeMillis() + " ";
    return reply;
  }		
  
}