package RouterLayer.Router;

import Abstract.*;
import RouterLayer.Router.*;
import RouterLayer.AgentClient.*;
import KQMLLayer.*;
import BaseLayer.*;

import java.lang.*;
import java.io.*;
import java.util.*;

/**
* Secure the connection. Check whether the connection,disconnection protocol
* is satisfied. Password check will be performed by setting passwordCheckBit
* using setPasswordCheck(true) method. Since there's no connection protocol
* upon which all the agent communities agree, we are just suggesting the protocol,
* which is neither sufficient nor theoretically correct. We hope this 
* problem to be resolved in the near future. This is our current connection protocol<BR>
* At first, upon request, Router will accept connection and send SMTP for the connection.
* Then, the agent will send 'reconnect-agent' protocol message and if the agent is
* acceptable, the Router will send 'connection-accepted' message to the agent.<BR>
* From beta version, Registry file will be retained to keep agent registry records.
* The registry will not be updated until the agent send 'unregister' message to router.
* For example,<BR>
* <UL>
* <LI>Agent send message to the Router<BR>
* (reconnect-agent :sender agent :receiver Router :password xxxx)<BR>
* <LI>Router will respond<BR>
* (connection-accepted :sender Router :receiver agent)<BR>
* or<BR>
* (error :sender Router :receiver agent :content (error-content))<BR>
* if error occurs.
* </UL>
*/

/*
* Copyright (c) 1996,1997,1998 <A HREF="http://cdr.stanford.edu/">Center for Design Research</A>, Stanford University.
* All rights reserved.<p>
*
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You can obtain the GNU General Public License at
* http://www.gnu.org/copyleft/gpl.html or by writing to the Free
* Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
* 02111-1307, USA.
* 
* This software is bound by the terms and conditions listed in the
* attached <a href="../LICENSE">LICENSE file</A>.
*  @author <A HREF="http://cdr.stanford.edu/html/people/jeon-bio.html>Heecheol Jeon</A>
*  @version Java(tm) 1.1
*/ 

public class RouterSecurity extends Security  {

  /**
   * secureAgent() method will return if 'whoiam' information is correct
   */
  public final int CLIENTOK = 0;
  
  /**
   * 'whoiam' message does not include the client name
   */
  public final int CLIENTNAMEMISSING = 1;
  
  /**
   * Client name is the same as the Router name. Connection will be denied
   */
  public final int CLIENTSAMENAME = 2;
  
  /**
   * Connection needs password check, but 'whoiam' message does not include password
   */
  public final int PASSWORDMISSING = 3;
  
  /**
   * The agent password is not matched with its pssword
   */
  public final int PASSWORDMISMATCH = 4;
  
  /**
   * Password file is corrupted
   */
  public final int PASSWORDFILECORRUPTED = 5;	
  
  /**
   * The agent is not registered to the Router
   */
  public final int NOCLIENTLIST = 6;
  
  /**
   * The message-method is not understandable to the Router
   */
  public final int NOTVALIDMESSAGEMETHOD = 7;
  
  public final int WRONGROUTERNAME = 8;
  /**
   * Hashtable which contains registered user list. Key is agent name and field value
   * is password
   */	
  Hashtable _userList;
  
  /**
   * Retain registry
   */
  String _registryFilePath;
  
  /**
   * Default constructor
   */
  public RouterSecurity()  {
    super();
  }
  
  /**
   * Constructor with AddressTable, password file path and registry file path. 
   * For beta version, password file should be passed even if you do not want to 
   * check password. In this case, passwor file will be used as quick reference
   * of registered agents. This constructor will set passwordCheckBit to be true.
   * If you do not want to check password, use next constructor. Registry file
   * can be null if you do not want to retain registry information.
   * @param addresses AddressTable
   * @param passfile Password file path
   * @param registryFile Registry file path
   */
  public RouterSecurity(AddressTable addresses,String passfile,String registryFile)  {
    super(addresses,passfile);
    _registryFilePath = registryFile;
    if(passfile != null)   {
      setUserList(passfile);
    }
  }
  
  /**
   * Constructor with AddressTable, password file path and registry file path. 
   * For beta version, password file should be passed even if you do not want to 
   * check password. In this case, passwor file will be used as quick reference
   * of registered agents. 
   * @param addresses AddressTable
   * @param passfile Password file path
   * @param registryFile Registry file path
	 * @param passwordCheckBit Password check bit set. If you do not want to password check,
	 * pass this parameter as false.  
   */
  public RouterSecurity(AddressTable addresses,String passfile,String registryFile,boolean passwordCheckBit)  {
    super(addresses,passfile);
    _passwordCheck = passwordCheckBit;
    _registryFilePath = registryFile;
    if(passfile != null)   {
      setUserList(passfile);
    }
  }
  
  /**
   * Constructor with AddressTable and password file path. 
   * For beta version, password file should be passed even if you do not want to 
   * check password. In this case, passwor file will be used as quick reference
   * of registered agents. 
   * @param addresses AddressTable
   * @param passfile Password file path
   */
  public RouterSecurity(AddressTable addresses,String passfile)  {
    this(addresses,passfile,null);
  }
  
  /**
   * Set password file path. User list will be read and updated.
   * @param passfile Password file path
   */
  public void setPasswordFilePath(String passfile)  {
    _passwordFilePath = passfile;
    setUserList(passfile);
  }
  
  /**
   * Set user list using password file path
   * @param passfile Password file path
   */
  protected void setUserList(String passfile)   {
    _userList = new Hashtable();		

    try  {
      BufferedReader in = 
				new BufferedReader(new FileReader(new File(_passwordFilePath)));
      
      String line = in.readLine();
      while(line != null)  {
				if((!line.startsWith("#")) && (!line.startsWith("\n"))) {
	  			int sindex = line.indexOf(',');
	  			String username = line.substring(0,sindex);
	  			String password = null;
	  			if(_passwordCheck) {
	    			password = line.substring(sindex + 1,line.length());				
	  			}	
	  			_userList.put(username,password);
				}	
				line = in.readLine();
      }
      in.close();
      
      if(_registryFilePath != null)  {
      	BufferedReader registry = 
					new BufferedReader(new FileReader(new File(_registryFilePath)));
			
				line = registry.readLine();
				while(line != null)  {
		    	StringTokenizer st = new StringTokenizer(line,","); 
 					int nofToken = st.countTokens();
					try  {
 						// should have at least four tokens - id, host, port, type
 						if(nofToken >= 4)  {
 							
   						String id = st.nextToken();
							if(_userList.containsKey(id))  {
     						String host = st.nextToken();
								int port = (Integer.valueOf(st.nextToken())).intValue();
     						String type = st.nextToken();
     						String desc = "";
     						if(st.hasMoreTokens())  {
	     						desc += st.nextToken();
     						}
     						Address addr = new Address(id,host,port,type,desc);
     						_addresses.addAddress(addr);
     					}		
     				}  	
     			}	catch (Exception e) {
     				System.out.println(e.toString());
     			}
     		
     			line = registry.readLine();
				}
			
				registry.close();
			}	
					      
    } catch (Exception e)  {
      System.out.println(e.toString());
      System.exit(1);
    }		
  }
  
  /**
   * Security check process from the RouterServerThread. The connection is accepted and
   * the security check is performed using this method. For error condition, the
   * connection may be denied or the RouterServer may give one more chance to correct
   * the wrong information such as agent name missing, password mismatch, etc. 
   * If the client agent satifies the connection protocol the conneciton will be 
   * added to the ConnectionTable and start to receive/send messages
   * For both registration and connection, the lowest level connection protocol is:<BR>
   * Router : 201 AMR "Router name"<BR>
   * for the successful connection.<P>
   * or
   * Router : 500 Internal Error<BR>
   * for any error conditions.<P>
   * Then, the Agent will request connection by sending "reconnect-agent" protocol.<BR>
   * For example,<BR>
   *	Agent :(reconnect-agent :sender agent :receiver Router :password xxxx)<BR>
   * If the host/port is changed,
   * Agent : (reconnect-agent :sender agent :receiver Rotuer :password xxxx :host new.host.com :port 4321)<BR>
   * The the Router will send<BR>
   * Router : (connection-accepted :sender Router :receiver agent)<BR>
   * @param con RouterRecvThread, will be started
   */
  public void processServerLogin(ReceiverThread con)   {
    
    // cross check will be performed. By sending host address and port
    // the client side will check. If correct the client will send 
    // register message to server
    Address address = con.getAddress();
    
    try  {
System.out.println("Server login");    	
      String loginString = "201 AMR ";
      loginString = loginString + address.getID();
      con.writeMsg(loginString);
System.out.println("Connection Requested");      
      String response = con.getOneMessage();	
      KQMLmessage kqml = new KQMLmessage(response);
      String perf = kqml.getValue("performative");
      perf = perf.toLowerCase();
      
      // host,port is optional. Sent only if the host/port is changed
      // If your system does not check the password, password needs not to be sent
      String agentid = kqml.getValue("sender");
      String myid = kqml.getValue("receiver");
      String host = kqml.getValue("host");
      String port = kqml.getValue("port");
      String password = kqml.getValue("password");
      String email = kqml.getValue("email");
      String description = null;
      
      if(perf.equals("reconnect-agent"))  {
	int checkresult = secureAgent(address.getID(),myid,agentid,password);
	if((checkresult == CLIENTOK) ||
	   (processErrorLogin(checkresult,con,myid,agentid))) {
	  Address agentaddr = _addresses.getAddress(agentid);
	  
	  if((password != null) || (email != null))  {
	    description = "(agent-info ";
	    if(password != null)  {
	      description = description + " :password " + password;
	    }
	    if(email != null)  {
	      description = description + " :email " + email;
	    }
	    description = description + ")";
	  }			
	  
	  if(agentaddr == null) {
	    agentaddr = new Address(agentid,host,(Integer.valueOf(port)).intValue(),"MessageRouter",description);
	    _addresses.addAddress(agentaddr); 	
	  }
	  else
	    if((host != null) && (!host.equals(agentaddr.getHost()))) {
	      // update email if necessary
	      String descrp = agentaddr.getDescription();
	      if(descrp != null)   {
		try  {
		  KQMLmessage tmpkqml = new KQMLmessage(descrp);
		  String tKQMLext = tmpkqml.getValue("KQML-extensions");
		  String tdescription = tmpkqml.getValue("description");
		  String tpassword = tmpkqml.getValue("password");
		  String temail = tmpkqml.getValue("email");
		  
		  description = "(agent-info";
		  if(tKQMLext != null)  {
		    description = description + " :KQML-extensions " + tKQMLext;
		  }
		  if(tdescription != null)  {
		    description = description + " :description " + tdescription;
		  }
		  // only e-mail changed
		  if(email != null)  {
		    description = description + " :email " + email;
		  }
		  if(tpassword != null)  {
		    description = description + " :password " + tpassword;		
		  }	
		  description = description + ")";
		  
		} catch (Exception e)  {}	
	      } 	 
	    }
	  agentaddr = new Address(agentid,host,(Integer.valueOf(port)).intValue(),agentaddr.getType(),description);
	  _addresses.addAddress(agentaddr); 	
	  
	  startConnection(con,agentid,myid);
	}		
	else  {
	  denyConnection(con,myid);
	}
      }
      else  {
	denyConnection(con,myid);
      }	
    }  catch (Exception e)  {
      System.out.println("Security " + e.toString());
      con.endConn();
    }
  }					
  
  /**
   * According to the error condition, the processErrorLogin will be performed.
   * If the agent name is missing or if the agent name is the same as the Router
   * name or if the password is missing(when passwordCheckBit is true) or 
   * if the password is mismatch(when passwordCheckBit is true), the Router will
   * send 'ask-one' message to the agent. If the agent can correct the problem,
   * the connection will be started. If not, the conneciton will be denied and
   * closed.
   * @param errorCode Error code which happens during connection process
   * @param con RouterRecvThread, which wants to be connected
   * @param myid The Router name
   */
  protected boolean processErrorLogin(int errorCode,ReceiverThread con,String myid,String agentid)   {		  
    try  {
      String error = "(tell :sender ";
      error = error + myid + " :receiver Unknown :content (";
      
      switch(errorCode)   {
      case CLIENTNAMEMISSING:
	error = error + "Client name is missing)";
	break;
      case CLIENTSAMENAME:
	error = error + "Same name connection is existing)";
	break;
      case PASSWORDMISSING:
	error = error + "Password is not specified)";
	break;
      case PASSWORDMISMATCH:
	error = error + "Password mismatch)";	
	break;
      case WRONGROUTERNAME:
	error = error + "Router name mismatching)";				
	break;
      case NOCLIENTLIST:
	error = error + "Not registered agent";
	if(agentid != null)  {
	  error = error + " <" + agentid + ">";
	}
	error = error + ")";
	break;					
      default:
	error = error + ")";	
      }		
      error = error + ")";
      // send error
      con.writeMsg(error);
System.out.println(error);      
      /*
      String msg = con.getOneMessage();
      KQMLmessage kqml = new KQMLmessage(msg);
      String perf = kqml.getValue("performative");
      String sender = kqml.getValue("sender");
      
      if(perf.equals("reply"))  {
				String content = kqml.getValue("content");
				KQMLmessage pass = new KQMLmessage(content);
				String password = pass.getValue("password");
				if(password != null)  {
	  			if(secureAgent(myid,sender,password) == CLIENTOK)   {
	    			return true;
	  			}
	  			else  {
	    			return false;
	  			}	
				}	
				else  {
	  			return false;
				}
      }
      else  {
				return false;
      }	
      */
    }		catch (Exception e)  {
      System.out.println(e.toString());
      con.endConn();
    }
    return false;
  }				
  
  /**
   * Start conneciton. All the conneciton protocol is satisfied. Send 'accepted' message
   * and add the conneciton to the ConnectionTable. This method set the ReceiverThread name
   * as the same as the agent name. 
   * @param con ReceiverThread to be started 
   * @param agentid The agent name
   * @param myid The Router name
   */
  protected void startConnection(ReceiverThread con,String agentid,String myid)  {
    System.out.println("Connection Request from " + agentid);				
    String access = "(connection-accepted :sender ";
    access = access + myid + " :receiver " + agentid + ")";
    try {
      con.writeMsg(access);
    }  catch (ConnectionException e) {
      con.endConn();
    }
    
    // override if the same name connection is living
    con.getConnectionTable().addConnection(agentid,con,true);
  } 	
  
  /**
   * Error occurs from the connection process. Send 'error' message to the agent
   * and disconnect the conneciton
   * @param con ReceiverThread to be disconnected
   * @param id The Router name
   */
  protected void denyConnection(ReceiverThread con,String id)  {
    String errormsg = "(error :sender ";
    errormsg = errormsg + id + " :content register-expected)";
    
    try {
      con.writeMsg(errormsg);
    }  catch (ConnectionException ee) {}
    con.endConn();
  }
  
  /**
   * Check the client send its name and the name is different from mine
   * Password check will be performed if passwordCheckBit is set
   * @param me My name, sent back from the client
   * @param client The agent name
   * @param password The agent Password
   */
  protected int secureAgent(String routerID,String me,String client,String password)  {
    if(client == null) return CLIENTNAMEMISSING;
    // client can not have the same name with me
    if(client.equals(me)) return CLIENTSAMENAME;
    
    if(!routerID.equals(me))  return WRONGROUTERNAME;
    if(_passwordCheck)   {
      if(password == null)  {
				return PASSWORDMISSING;
      }
      
      if(_userList.containsKey(client))   {
	String pass = (String)_userList.get(client);
	if(pass.equals(password))  {
	  return CLIENTOK;
	}
	else  {
	  return PASSWORDMISMATCH;
	}
      }
      else  {
	return NOCLIENTLIST;
      }
    }					
    
    return CLIENTOK;
  }	
  
  /**
   * Opposite to the processServerLogin. The agent(me) wants to connect to the
   * Router and it will send 'whoiam' message to the Router
   * @param c ReceiverThread, which will connect to the Router
   * @param serveraddress The Rotuer server address
   */
  public void processClientLogin(ReceiverThread c,Address serveraddress) 
  	throws ConnectionException   {
    if(_addresses == null)  {
      SystemOut("No AddressTable is defined");
      throw new ConnectionException("Address not defined");
    }
    
    // My address should be specified
    Address myaddress = c.getAddress();
    if(myaddress == null)   {
      SystemOut("Address for " + c.getAddress().getID() + " is not specified. Fail to client connect");
      throw new ConnectionException("Address should be specified");
    }
    
    if(serveraddress == null)   {
      SystemOut("Address for " + c.getAddress().getID() + " is not specified. Fail to client connect");
      throw new ConnectionException("Server address should be specified");
    }  		
System.out.println("Start to connect as " + c.getAddress().getID());    
    String initial = c.getOneMessage();
    if(!initial.startsWith("201 AMR"))  {
      throw new ConnectionException("Server response not matched to AMR protocol");
    }
    
    
    try  {
      String loginString = "(reconnect-agent :host ";
      loginString = loginString + myaddress.getHost() + " :port " + Integer.toString(myaddress.getPort()) + " :sender " + myaddress.getID();
      loginString = loginString + " :receiver " + serveraddress.getID();
      
      // if you need password check, your password for accessing to the server
      // and server password you will give to the server should be specified
      String description = myaddress.getDescription();
      if(description != null)  {
	KQMLmessage dkqml = new KQMLmessage(description);
	String password = dkqml.getValue("password");
	if(password != null)  {
	  loginString = loginString + " :password " + password;
	}	
	
	String email = dkqml.getValue("email");
	if((email != null) && (!email.equals(""))) {
	  loginString = loginString + " :email " + email;
	}	
      }
      loginString = loginString + ")";	
			
			String response = "";
System.out.println(loginString);			
     	c.writeMsg(loginString);  	
     	response = c.getOneMessage();
      		
      KQMLmessage respkqml = new KQMLmessage(response);
      String respperf = respkqml.getValue("performative");
      
      if(respperf.equals("connection-accepted"))  {		
	// success. The client side need not to override the current living ReceiverThread
	// When "I" want to initiate the conneciton, I should check the connection running
	// first and should kill it first.
	SystemOut("Connection established");
	
	c.getConnectionTable().addConnection(serveraddress.getID(),c,false);	
      }
      else   {
System.out.println(respkqml.getReadString());      	
      	throw new ConnectionException(respkqml.getReadString());
      }		
    } catch (Exception e)  {
      SystemOut(e.toString());    
      throw new ConnectionException("Security throws error " + e.toString());
    }		
  }
  
  /**
   * Send disconnect message to the connection.
   * @param c RouterRecvThread to be disconnected
   */
  public void processClientLogout(ReceiverThread c)  {
    String logout = "(disconnect-agent :sender ";
    logout = logout + c.getAddress().getID() + " :receiver " + c.getName() + ")";
    try {
      c.writeMsg(logout);
    }  catch (ConnectionException e) {}
    // waiting for disconneciton confirming message for server to
    // disconnect first
    //c.getOneMessage();
  }	
  
	
  /**
  * Implementation of agent validity checking. If the agent is registered agent 
  * and its password is correct,reutrn true
  * @param o Object to be checked. For RouterSecurity, it is a string array, agent name
  * and the password
  */
  public boolean isValidAgent(Object o) {
    if(_userList != null)   {
      String[] info = (String[])o;
      if(_userList.containsKey(info[0]))  {
				if((info.length == 2) && _passwordCheck)  {
			  String password = info[1];
	  		if(password == null)  {
		    	return false;
		  	}
	  
	  		if(password.equals((String)_userList.get(info[0])))  {
	    		return true;
	  		}
	  		else  {
		    	return false;
		  	}
			}
			else {
	  		return true;
			}	
    }
    else  {
			return false;
    }
   }
    
    return true;								
  }
  
  /**
   * Update user list. Invoked by RegistrarSecurity object
   * @param id The agent name
   * @param password The agent password 
   */
/*   
  public synchronized void updateUserList(String id,String password)  {
    try  {
      if(!_userList.containsKey(id))  {
	RandomAccessFile out = new RandomAccessFile(_passwordFilePath,"rw");
	out.seek(out.length());
	if(_passwordCheck)  {
	  out.writeBytes(id + "," + password + "\n");
	}
	else  {
	  out.writeBytes(id + ", \n");
	}
	
	out.close();
      }	
      _userList.put(id,password);
    } catch (Exception e)  {
      System.out.println(e.toString());
    }		
  }  	
*/
  /**
   * Update user list. Invoked by RegistrarSecurity object. Password file and Registry file
   * will be updated
   * @param id The agent name
   * @param password The agent password 
   */
  public synchronized void updateUserList(Address addr,String password)  {
    try  {
    	String id = addr.getID();
    	if(id == null)  {
    		return;
    	}
    		
    		
      if(!_userList.containsKey(id))  {
				RandomAccessFile out = new RandomAccessFile(_passwordFilePath,"rw");
				out.seek(out.length());
				if(_passwordCheck)  {
	  			out.writeBytes(id + "," + password + "\n");
				}
				else  {
	  			out.writeBytes(id + ", \n");
				}
	
				out.close();
				
				RandomAccessFile registry = new RandomAccessFile(_registryFilePath,"rw");
				registry.seek(registry.length());
				registry.writeBytes(addr.toString() + "\n");
System.out.println(addr.toString());	
				registry.close();
				
				_addresses.addAddress(addr);
      }	
      _userList.put(id,password);
      
    } catch (Exception e)  {
      System.out.println(e.toString());
    }		
  }  	
	
  
  /**
   * Unregister Agent. The password should be passed if password check is enforced.
   * The user will be removed from Password and Registry file
   * @param o String array,length two with agent name and its password
   */
  public synchronized boolean unregisterAgent(Object o)  {
    if(isValidAgent(o))  {
      try  {
				String[] arg = (String[])o;
				_addresses.removeAddress(arg[0]);
				_userList.remove(arg[0]);
				DataOutputStream out = new DataOutputStream(new FileOutputStream(new File(_passwordFilePath)));
	
				Enumeration e = _userList.keys();
				while(e.hasMoreElements())  {
	  			String id = (String)e.nextElement();
	  			out.writeBytes(id + "," + (String)_userList.get(id) + "\n");
				}
	
				out.close();	
				
				if(_registryFilePath != null)   {
					DataOutputStream registry = new DataOutputStream(new FileOutputStream(new File(_registryFilePath)));
	
					e = _addresses.elements();
					while(e.hasMoreElements())  {
	  				Address addr = (Address)e.nextElement();
	  				registry.writeBytes(addr.toString() + "\n");
					}
	
					registry.close();	
				}	
				
				
				return true;
      } catch (Exception e)  {
				System.out.println(e.toString());
      }		
    }
    
    return false;
  }		
  
  /**
  * Check whether the user is registered.
  * @param id Agent name
  * @return True if the agent is registered.
  */
  public boolean isRegisteredUser(String id)  {
    if(_userList.containsKey(id))  {
      return true;
    }
    
    return false;
  }		
  
  /**
   * System message output to the System.out. This method will be override if you want to use GUI
   * @param str String will be displayed to the System.out or GUI
   */
  public void SystemOut(String str)  {
    System.out.println(str);
  }	
}		